package mstparser;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ObjectInputStream;
import java.util.HashMap;

public class DecodingThread implements Runnable {

	private DependencyParser dp = null;
	private DependencyPipe2O pipe = null;
	private DependencyDecoder2O decoder = null;
	private Parameters params = null;
	private Parameters firstOrderParams = null;
	private BufferedReader in_g = null;
	private BufferedReader in_p = null;
	private ObjectInputStream in_s = null;
	private BufferedReader in_c = null;
	private BufferedWriter out = null;
	private Integer[] stats;	// 0: token; 1: candNum; 2:oracleCorr;
								// 0: cert
	private Double[] totalScore;
	private Integer[][] currID;	// 0: process; 1: output
	private boolean verbal;
	private OutputThread ot = null;
	
	private Thread t = null;
	
	public DecodingThread(DependencyParser dp,
			DependencyPipe2O pipe,
			DependencyDecoder2O decoder,
			Parameters params,
			BufferedReader in_g,
			BufferedReader in_p,
			ObjectInputStream in_s,
			BufferedReader in_c,
			BufferedWriter out,
			Integer[] stats,
			Double[] totalScore,
			Integer[][] currID,
			OutputThread ot,
			boolean verbal) {
		this.dp = dp;
		this.pipe = pipe;
		this.decoder = decoder;
		this.params = params;
		this.in_g = in_g;
		this.in_p = in_p;
		this.in_s = in_s;
		this.in_c = in_c;
		this.out = out;
		this.stats = stats;
		this.totalScore = totalScore;
		this.currID = currID;
		this.ot = ot;
		this.verbal = verbal;
		this.firstOrderParams = dp.getFirstOrderParams();
		
		t = new Thread(this, "decoding thread");
		t.start();
	}
	
	public void join() throws Exception {
		t.join();
	}
	
	@Override
	public void run() {
		try {
			//double totalTime = 0.0;
			
			while(true) {
				DependencyInstance gold = null;
				DependencyInstance pred = null;
				LocalFeatureData lfd = null;
				int currProcessID = 0;

				synchronized (currID[0]) {
					currProcessID = currID[0][0];
					currID[0][0]++;
					gold = pipe.createInstance(in_g);
					pred = pipe.createInstance(in_p);
					
					if (gold != null) {
	
						if (pred != null) {
							for (int i = 1; i < pred.dep.length; i++)
								if (pred.dep[i] < 0)
									pred.dep[i] = 0;
							pipe.buildChild(pred);
						}
						
						if (DependencyParser.testingMode <= 3) {
							int length = gold.sentence.length;
							
							lfd = new LocalFeatureData(length, pipe.types.length, true);
				    		if (DependencyParser.pruningMode == 0) {
				    			// currently the pred-file should still exist, lazy to change the code...
				    			lfd.loadCandidate(null, pipe, pred, null, true);
				    			if (DependencyParser.predHasCand) {
					    			in_p.readLine();
					    			for (int z = 1; z < gold.dep.length; z++)
					    				in_p.readLine();
					    			in_p.readLine();
				    			}
				    		}
				    		else if (DependencyParser.pruningMode == 1) {
								lfd.loadCandidate(null, pipe, pred, null, true);
							}
							else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
								in_p.readLine();	// score
								lfd.loadCandidate(in_p, pipe, pred, null, true);
								in_p.readLine();	// empty line
							}
				    		else if (DependencyParser.pruningMode == 4) {
				    			if (DependencyParser.predHasCand) {
					    			in_p.readLine();
					    			for (int z = 1; z < gold.dep.length; z++)
					    				in_p.readLine();
					    			in_p.readLine();
				    			}
				    			// load real candidate
				    			pipe.readConllSent(in_c, pipe.labeled);
				    			in_c.readLine();
				    			lfd.loadCandidate(in_c, pipe, pred, null, true);
				    			in_c.readLine();
				    		}
						}
						else if (DependencyParser.testingMode == 4) {
							lfd = new LocalFeatureData(gold.sentence.length, pipe.types.length, true);							
						}
						
						if (DependencyParser.pruningMode == 3) {
							gold.parserPredScore = new LocalFeatureData(gold.dep.length, pipe.possibleLab.size(), true);
							gold.parserPredScore.loadParserPredScore(in_s, true);
						}
					}
				}
				
				if (currProcessID >= DependencyParser.testSentences
						|| gold == null) {
					if (ot != null) {
						synchronized (ot.finishThreadNum) {
							ot.finishThreadNum[0]++;
						}
					}
					break;
				}

				if (!pipe.isDev(currProcessID))
					continue;
				
				pred.m = gold.m;
	    		pred.posA = gold.posA;
	    		pred.cm = gold.cm;
	    		pred.origWord = gold.origWord;
	    		if (DependencyParser.correctPos) {
		    		for (int z = 1; z < pred.dep.length; z++) {
		    			if (pred.sentence[z] != gold.sentence[z]
		    					|| pred.goldPos[z] != gold.pos[z]) {
		    				System.out.println("output parse bug 1");
		    			}
		    		}
	    		}
	    		else {
		    		for (int z = 1; z < pred.dep.length; z++) {
		    			if (pred.sentence[z] != gold.sentence[z]
		    					|| pred.pos[z] != gold.pos[z]) {
		    				System.out.println("output parse bug 1: " + gold.origWord[z] + "\t" + pred.sentence[z] + "\t" + gold.sentence[z]);
		    			}
		    		}
	    		}
	    		
				pipe.copyPredDepLab(gold, pred.dep, pred.labs);
	
				pred.parserPredScore = gold.parserPredScore;
				pred.parserPredDep = gold.parserPredDep;
				pred.parserPredLab = gold.parserPredLab;
	
				if (verbal) {
					if (DependencyParser.testingMode == 0 || DependencyParser.testingMode == 2
							|| DependencyParser.correctPos)
						System.out.print(currProcessID +" ");
					else if (currProcessID % 100 == 0)
						System.out.print(currProcessID + " ");
				}
	
				if (DependencyParser.testingMode <= 3) {
					
					GlobalFeatureData gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
					
					boolean done = false;

					if (DependencyParser.testingMode == 0 || DependencyParser.testingMode == 2) {
						Object[][] d = null;
			
						double time = 0.0;
						if (DependencyParser.correctPos)
							d = decoder.decodeSamplingWithPos(pred, lfd, gfd, params, true);
						else {
							long start = System.currentTimeMillis();
							//d = decoder.decodeSampling(pred, lfd, gfd, params, true);
							d = decoder.decodeSamplingRW(pred, lfd, gfd, params, firstOrderParams, true);
							//d = decoder.decodeNonProjective(pred, lfd, params, 1);
							time = System.currentTimeMillis() - start;
						}
						
						if (DependencyParser.testingMode == 0) {
							double goldScore = 0.0;
							if (DependencyParser.outputScore) {
								goldScore = pipe.getLocalScore(gold, lfd, params, -1) + pipe.getGlobalScore(gold, gfd, params);
							}
							synchronized (currID[1]) {
								ot.id2Pred.put(currProcessID, pred);
								ot.id2PredStr.put(currProcessID, (String)d[0][1]);
								totalScore[0] += (Double)d[0][2];
								totalScore[1] += (double)time;
								totalScore[2] += pred.pos.length - 1;
								if (DependencyParser.outputScore) {
									stats[0]++;
									double sampScore = (Double)d[0][2];
									if (sampScore > goldScore - 1e-4)
										stats[1]++;
									//System.out.println(sampScore + "\t" + goldScore);
								}
								if (DependencyParser.correctPos) {
									for (int i = 1; i < pred.pos.length; i++) {
										stats[0]++;
										if (pred.pos[i] == gold.pos[i])
											stats[1]++;
									}
								}
							}
						}
						else if (DependencyParser.testingMode == 2) {
							Object[][] mstd = null;
							if (DependencyParser.decodeType.equals("non-proj"))
								mstd = decoder.decodeNonProjective(pred, null, lfd, params, 1);
							else
								mstd = decoder.decodeProjective(pred, lfd, params, 1);
							double sampScore = (Double)d[0][2];
							double mstScore = params.getScore((FeatureVector)mstd[0][0]);
							//double mstScore = (Double)mstd[0][2];
							synchronized (currID[1]) {
								ot.id2Pred.put(currProcessID, null);
								ot.id2PredStr.put(currProcessID, 
										sampScore + "\t" + mstScore + "\t" + (Math.abs(sampScore - mstScore) < 1e-4 ? 1 : 0));
								stats[0]++;
								if (Math.abs(sampScore - mstScore) < 1e-4)
									stats[1]++;
							}
						}
					}
					else if (DependencyParser.testingMode == 1) {
						Object[][] d = null;
						if (DependencyParser.correctPos) {
							d = decoder.decodeSamplingWithPos(pred, lfd, null, params, true);
						}
						else {
							if (DependencyParser.decodeType.equals("non-proj"))
								d = decoder.decodeNonProjective(pred, null, lfd, params, 1);
							else
								d = decoder.decodeProjective(pred, lfd, params, 1);
						}
						synchronized (currID[1]) {
							ot.id2Pred.put(currProcessID, pred);
							ot.id2Gold.put(currProcessID, gold);
							ot.id2PredStr.put(currProcessID, (String)d[0][1]);
						}
					}
					else if (DependencyParser.testingMode == 3) {
						Object[][] d = null;
						if (DependencyParser.decodeType.equals("non-proj"))
							d = decoder.decodeNonProjective(pred, null, lfd, params, 1);
						else
							d = decoder.decodeProjective(pred, lfd, params, 1);
						while (!done) {
							synchronized (currID[1]) {
								while (!pipe.isDev(currID[1][0])) {
									currID[1][0]++;
								}
								if (currID[1][0] == currProcessID) {
									dp.outputConllSentence(out, pred, (String)d[0][1]);
									currID[1][0]++;
									done = true;
								}
							}
							Thread.sleep(100);
						}
					}
				}
				else if (DependencyParser.testingMode == 4) {
					// decode rerank

					// compute feature vector for top-k list
	    			FeatureVector[] fvList = new FeatureVector[DependencyParser.maxRerankK];
	    			HashMap<Integer, String> labChange = pipe.processChangeStr(in_p.readLine());
	    			HashMap<Integer, String> depChange = pipe.processChangeStr(in_p.readLine());
	    			int[][] depList = new int[DependencyParser.maxRerankK][pred.dep.length];
	    			int[][] labList = new int[DependencyParser.maxRerankK][pred.labs.length];
	    			
	    			pred.parserRerankScore = 1.0;
	    			fvList[0] = pipe.createFeatureVector(pred);
	    			int bestAcc = 0;

	    			for (int z = 0; z < pred.dep.length; z++) {
	    				depList[0][z] = pred.dep[z];
	    				labList[0][z] = pred.labs[z];
	    				if (z != 0 && depList[0][z] == gold.dep[z])
	    					bestAcc++;
	    			}

	    			int k = 1;
	    			String str = in_p.readLine();
	    			
	    			while (!str.isEmpty()) {
	    				if (k < DependencyParser.maxRerankK) {
		    				String[] data = str.split("\t");
		    				pred.parserRerankScore = Double.parseDouble(data[0]);
		    				for (int z = 1; z < data.length; z++) {
		    					int changeID = Integer.parseInt(data[z]);
		    					if (depChange.containsKey(changeID)) {
		    						String s = depChange.get(changeID);
		    						int p = s.indexOf(":");
		    						int mod = Integer.parseInt(s.substring(0, p));
		    						int head = Integer.parseInt(s.substring(p + 1));
		    						pred.dep[mod] = head;
		    					}
		    					else if (pipe.labeled && labChange.containsKey(z)) {
		    						String s = depChange.get(changeID);
		    						int p = s.indexOf(":");
		    						int mod = Integer.parseInt(s.substring(0, p));
		    						int lab = pipe.typeAlphabet.lookupIndex(s.substring(p + 1));
		    						pred.labs[mod] = lab;
		    					}
		    				}
		    				pipe.buildChild(pred);
		    				fvList[k] = pipe.createFeatureVector(pred);
		    				int acc = 0;
		    				
			    			for (int z = 0; z < pred.dep.length; z++) {
			    				depList[k][z] = pred.dep[z];
			    				labList[k][z] = pred.labs[z];
			    				if (z != 0 && depList[k][z] == gold.dep[z])
			    					acc++;
			    			}
		    				
			    			if (acc > bestAcc)
			    				bestAcc = acc;
			    			
		    				// recover
		    				decoder.copyDepLab(pred, depList[0], labList[0]);
		    				pipe.buildChild(pred);

		    				k++;
	    				}
	    				str = in_p.readLine();
	    			}
	    			
	    			// choose the one with best score
	    			Object[][] d = decoder.decodeReranking(pred, fvList, depList, labList, params);
					synchronized (currID[1]) {
						ot.id2Pred.put(currProcessID, pred);
						ot.id2PredStr.put(currProcessID, (String)d[0][1]);
						stats[0] += pred.dep.length - 1;
						stats[2] += bestAcc;
					}
				}
				
			}
			
			//System.out.println("total time: " + totalTime);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}

class OutputThread implements Runnable {

	public HashMap<Integer, DependencyInstance> id2Pred = null;
	public HashMap<Integer, DependencyInstance> id2Gold = null;
	public HashMap<Integer, String> id2PredStr = null;
	public Integer[] finishThreadNum = null;
	
	private Integer[] stats = null;
	private Parameters params = null;
	
	private BufferedWriter out = null;
	private Integer[] outputID = null;
	private DependencyParser dp = null;
	
	private Thread t;
	
	public OutputThread(BufferedWriter out, Integer[] outputID, DependencyParser dp, Parameters params, Integer[] stats) {
		this.out = out;
		this.outputID = outputID;
		this.dp = dp;
		this.params = params;
		this.stats = stats;
		
		id2Pred = new HashMap<Integer, DependencyInstance>();
		id2Gold = new HashMap<Integer, DependencyInstance>();
		id2PredStr = new HashMap<Integer, String>();
		finishThreadNum = new Integer[1];
		finishThreadNum[0] = 0;
		
		t = new Thread(this, "output thread");
		t.start();
	}
	
	public void join() throws Exception {
		t.join();
	}
	
	@Override
	public void run() {
		try {
			DependencyPipe2O pipe = dp.getPipe();
			boolean allFinish = false;
			int badcnt = 0;
			while (!allFinish || !id2Pred.isEmpty()) {
				boolean find = true;
				synchronized (finishThreadNum) {
					if (finishThreadNum[0] >= DependencyParser.devThread) {
						allFinish = true;
					}
				}
				synchronized (outputID) {
					
					while (!pipe.isDev(outputID[0])) {
						outputID[0]++;
					}
					if (id2Pred.containsKey(outputID[0].intValue())) {
						if (DependencyParser.testingMode == 0 || DependencyParser.testingMode == 4) {
							dp.outputConllSentence(out, id2Pred.get(outputID[0]), id2PredStr.get(outputID[0]));
						}
						else if (DependencyParser.testingMode == 1) {
							dp.generateCandidate(out, id2Pred.get(outputID[0]), id2Gold.get(outputID[0]), 
									id2PredStr.get(outputID[0]), params, stats);
						}
						else if (DependencyParser.testingMode == 2) {
							out.write(id2PredStr.get(outputID[0]) + "\n");
							out.flush();
						}
						id2Pred.remove(outputID[0]);
						id2Gold.remove(outputID[0]);
						id2PredStr.remove(outputID[0]);
						outputID[0]++;
					}
					else {
						find = false;
					}
				}
				if (allFinish && !id2Pred.isEmpty() && !find) {
					System.out.println("output thread bug");
					badcnt++;
					//System.exit(0);
				}
				else {
					badcnt = 0;
				}
				if (!find) {
					Thread.sleep(100);
				}
				if (badcnt > 1000)
					break;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
}

