package mstparser;

import static mstparser.FeatureVector.end;
import gnu.trove.map.hash.*;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

public class DependencyDecoder2O {
	
    DependencyPipe2O pipe;
    //double coolingRate = 0.95;
    //double initTemperature = 10.0;
    //int maxIter = 300;
    //int miniIter = 5;
    //int restart = 5;
    //String samplingMethod = "gibbs";
    boolean proj = true;
    public int updateTimes = 0;
    
    Random r = null;
    int seed = 0;
    //Random dr = new Random(2);
	
    public DependencyDecoder2O(DependencyPipe2O pipe) throws IOException {
    	this.pipe = pipe;
    	updateTimes = 0;
    }
    
    public void copyDepLab(DependencyInstance inst, int[] dep, int[] lab) {
    	if (inst.dep == null)
    		inst.dep = new int[dep.length];
    	if (inst.labs == null)
    		inst.labs = new int[lab.length];
    	for (int i = 0; i < inst.dep.length; i++) {
    		inst.dep[i] = dep[i];
    		inst.labs[i] = lab[i];
    	}
    }
    
    public void setProj() {
    	proj = DependencyParser.decodeType.equals("proj");
    }
    
    public void setSeed(int seed) {
    	this.seed = seed;
    	r = new Random(seed);
    }
    
	public boolean isDescendant(int head, int arg, int[] dep) {
		boolean ret = false;
		int step = 0;
		while (dep[head] != -1) {
			if (dep[head] == arg) {
				ret = true;
				break;
			}
			head = dep[head];
			step++;
			if (step > 1000) {
				System.out.println("loop bug");
				return false;
			}
		}
		return ret;
	}
	
	public double getInitTemp(DependencyInstance pred, 
			LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params) {
		int len = pred.dep.length;
    	double avgDiff = 0.0;
    	int cnt = 1;
    	double gScore = pipe.getGlobalScore(pred, gfd, params);
    	for (int i = 1; i < len; i++) {
    		double currScore = gScore + pipe.getLocalScore(pred, lfd, params, i);
			// build score
			int oldDep = pred.dep[i];
			int oldLab = pred.labs[i];
			for (int candH = 0; candH < len; candH++) {
				if (!lfd.cand[candH][i])
					continue;
				if (isDescendant(candH, i, pred.dep))		// loop
					continue;
				if (proj && !pipe.isProj(candH, i, pred.dep))
					continue;
				for (int candT = 0; candT < lfd.type; candT++) {
					if (!lfd.cand_dep[candH][i][candT])
						continue;
	        		int oldH = pred.dep[i];
					pred.dep[i] = candH;
					pred.labs[i] = candT;
	        		pipe.updateChildList(pred, candH, oldH, i);

					double score = pipe.getLocalScore(pred, lfd, params, i) + pipe.getGlobalScore(pred, gfd, params);
					//if (score < currScore) {
					//	avgDiff += score - currScore;
					//	cnt++;
					//}
					if (score - currScore < avgDiff) {
						avgDiff = score - currScore;
						cnt = 1;
					}
				}
			}
			
   			int oldH = pred.dep[i];
			pred.dep[i] = oldDep;
			pred.labs[i] = oldLab;
   			pipe.updateChildList(pred, oldDep, oldH, i);
    	}
    	if (Math.abs(avgDiff) < 1e-4) {
    		//System.out.println("init temp bug");
    		return 10.0;
    	}
    	else
    		return avgDiff / cnt / Math.log(0.8);
		
	}
	
	public String getParseString(DependencyInstance pred) {
   		// string
   		String spans = "";
   		for(int i = 1; i < pred.dep.length; i++) {
   			spans += pred.dep[i]+"|"+i+":"+pred.labs[i]+" ";
   		}
   		return spans.trim();
	}
	
    // decoding by sampling
    public Object[][] decodeSampling(DependencyInstance pred, 
    		LocalFeatureData lfd, GlobalFeatureData gfd,
    		Parameters params, 
    		boolean sampling) {
    	Random dr = new Random(seed);
    	
   	    int samplingMethod = 0;			// gibbs
   	    //int samplingMethod = 1;		// mh

   	    int len = pred.sentence.length;
    	
   		double initTemperature = 1.0;
   		if (sampling) {
   			initTemperature = getInitTemp(pred, lfd, gfd, params);
   			//if (DependencyParser.isSpmrl)
   			//	initTemperature /= 3;
   		}
   	    double coolingRate = 0.95;
   	    int maxIter = 300;
   	    int miniIter = 6;
   	    if (pipe.lang.equals("spmrl"))
   	    	miniIter = 30;
   	    int restart = 1;
   	    //if (pipe.lang.equals("arabic") || pipe.lang.equals("slovene") || DependencyParser.isSpmrl)
   	    //	restart = 1;
   	    if (!sampling)
   	    	restart = 1;
   	    
   	    //if (samplingMethod == 0)
   	    //	miniIter /= 6;
    	
    	Object[][] d = new Object[1][3];
    	double bestScore = -1e30;
    	int bestIter = 0;
    	long bestTime = 0;
    	int iterCount = 0;
    	// should not need to build parser prediction because they are already in the feature vector
    	
    	// begin sampling
    	pipe.copyPredDepLab(pred, pred.dep, pred.labs);
   	    for (int res = 0; res < restart; res++) {
   	    	// recover the starting
   	    	copyDepLab(pred, pred.parserPredDep, pred.parserPredLab);
   	    	pipe.buildChild(pred);
   	    	
   	    	double T = initTemperature;
   	    	//double T = 0.0;
   	    	double currScore = 0.0;
   			if (sampling) {
   				if (DependencyParser.pruningMode == 0) {
   					initSample(pred, lfd);
   				}
   				//currScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
	   	    	long start = System.currentTimeMillis();
   				for (int iter = 0; iter < maxIter; iter++) {
	   	    		boolean change = false;
	   	    		for (int z = 0; z < miniIter; z++) {
	   	    			
		   	      		int[] idx = new int[len];
		   	     		int id = getBottomUpOrder(pred, 0, idx, len - 1);
		   	     		if (id != 0) {
		   	     			System.out.println("build order bug");
		   	     		}
		   	     		
	   	        		for (int y = 1; y < len; y++) {
	   	        			int i = idx[y];
	   	        			//int i = y;
	   	        			
	   	        			// sample a dep
	   	        			//ArrayList<Double> score = new ArrayList<Double>();
	   	        			ArrayList<Integer> dep = new ArrayList<Integer>();
	   	        			ArrayList<Integer> lab = new ArrayList<Integer>();
	   	        			
	   	    				int oldDep = pred.dep[i];
	   	    				int oldLab = pred.labs[i];
	   	    				int sample = -1;
	
	   	    				if (samplingMethod == 0) {
	   	    					//System.out.println("not supported!");
	   	    					ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
	   	    					ArrayList<Double> score = new ArrayList<Double>();
	   	        				sample = gibbsSampling(pred, null, i, dep, lab, fvList, score, lfd, gfd, params, T, false, dr);
	   	        				
	   	        				if (dep.get(sample) != pred.dep[i] || lab.get(sample) != pred.labs[i]) {
   	        						pred.dep[i] = dep.get(sample); pred.labs[i] = lab.get(sample);
   	        						pipe.updateChildList(pred, dep.get(sample), oldDep, i);
   	        						change = true;
	   	        				}
	   	        			}
	   	        			else if (samplingMethod == 1) {
	   	        				sample = mhSampling(pred, i, dep, lab, lfd, dr);
	   	        				
	   	        				if (dep.get(sample) != oldDep || lab.get(sample) != oldLab) {
	   	        					
	   	        					//if (miniIter % 2 == 0) {
				   	    				double oldScore = pipe.getLocalScore(pred, lfd, params, i) + pipe.getGlobalScore(pred, gfd, params);
				   	    				
				   	        			pred.dep[i] = dep.get(sample);
				   	        			pred.labs[i] = lab.get(sample);
				   	        			pipe.updateChildList(pred, dep.get(sample), oldDep, i);
				   	        			//double newScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
				   	        			double newScore = pipe.getLocalScore(pred, lfd, params, i) + pipe.getGlobalScore(pred, gfd, params);
				   	        			
				   	        			// transition function for simulate annealing
				   	           			if (!accept(newScore, oldScore, T, dr)) {
				   	            			// not accept
					   	        			int oldH = pred.dep[i];
				   	            			pred.dep[i] = oldDep;
				   	            			pred.labs[i] = oldLab;
					   	        			pipe.updateChildList(pred, oldDep, oldH, i);
				   	            		}
				   	            		else {
				   	            			// accept
			   	                			change = true;
				   	                		//currScore = newScore;
				   	            		}
	   	        					//}
	   	        					//else {
//				   	    				double oldScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
//				   	    				
//				   	        			pred.dep[i] = dep.get(sample);
//				   	        			pred.labs[i] = lab.get(sample);
//				   	        			pipe.updateChildList(pred, dep.get(sample), oldDep, i);
//				   	        			
//				   	        			int oldDep2 = pred.dep[oldDep];
//				   	        			pred.dep[oldDep] = i;
//				   	        			pipe.updateChildList(pred, i, oldDep2, oldDep);
//				   	        			
//				   	        			//double newScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
//				   	        			double newScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
//				   	        			
//				   	        			// transition function for simulate annealing
//				   	           			if (!accept(newScore, oldScore, T, dr)) {
//				   	            			// not accept
//				   	           				pred.dep[oldDep] = oldDep2;
//				   	           				pipe.updateChildList(pred, oldDep2, i, oldDep);
//				   	           				
//					   	        			int oldH = pred.dep[i];
//				   	            			pred.dep[i] = oldDep;
//				   	            			pred.labs[i] = oldLab;
//					   	        			pipe.updateChildList(pred, oldDep, oldH, i);
//				   	            		}
//				   	            		else {
//				   	            			// accept
//			   	                			change = true;
//				   	                		//currScore = newScore;
//				   	            		}
//	   	        					}
	   	        				}
	   	        			}

	   	    				iterCount++;
	   	    				
		   	    			currScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
		   	    	    	if (currScore > bestScore) {
		   	    	    		bestScore = currScore;
		   	    	    		bestIter = iterCount;
		   	    	    		bestTime = System.currentTimeMillis() - start;
		   	    	    		// fv
		   	    	    		if (sampling) {
		   	    	    			d[0][0] = pipe.createFeatureVector(pred, lfd, gfd, -1);
		   	    	    		}
		   	    	    		d[0][1] = getParseString(pred);
		   	    	    		d[0][2] = new Double(currScore);
		   	    	    	}

	   	        		}
	   	        		
	   	    		}
	   	    		if (!change && T < 0.01) {
	   	    			break;
	   	    		}
	   	    		T = T * coolingRate;
	   	    	}
   			}
   			/*
   			currScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
   	    	if (currScore > bestScore) {
   	    		bestScore = currScore;
   	    		// fv
   	    		if (sampling) {
   	    			d[0][0] = pipe.createFeatureVector(pred, lfd, gfd, -1);
   	    		}
   	    		d[0][1] = getParseString(pred);
   	    		d[0][2] = new Double(currScore);
   	    	}
   	    	*/
   	    }
    	
   	    //System.out.println("best time: " + bestTime);
    	return d;
    }
    
    public Object[][] decodeSamplingWithPos(DependencyInstance pred, 
    		LocalFeatureData lfd, GlobalFeatureData gfd,
    		Parameters params, 
    		boolean sampling) {
    	//Random dr = new Random(4);
    	Random dr = new Random(seed);
    	
   	    int samplingMethod = 1;		//mh
   	    //int samplingMethod = 0;		//gibbs

   	    int len = pred.sentence.length;
    	
   		double initTemperature = 1.0;
   		if (sampling) {
   			initTemperature = getInitTemp(pred, lfd, gfd, params);
   			if (DependencyParser.isSpmrl)
   				initTemperature /= 3;
   		}
   	    double coolingRate = 0.95;
   	    int maxIter = 300;
   	    int miniIter = 60;
   	    if (DependencyParser.isSpmrl)
   	    	miniIter = 30;
   	    int restart = 4;
   	    if (pipe.lang.equals("arabic"))
   	    	restart = 2;
   	    else if (DependencyParser.isSpmrl)
   	    	restart = 1;
   	    if (!sampling)
   	    	restart = 1;
   	    
   	    if (samplingMethod == 0)
   	    	miniIter /= 6;
    	
 		int samePosAsPred = 0;
 		for (int i = 1; i < len; i++)
 			if (pred.currPosCandID[i] == 0)
 				samePosAsPred++;

 		Object[][] d = new Object[1][3];
    	double bestScore = -1e30;
    	
    	// begin sampling
    	pipe.copyPredDepLab(pred, pred.dep, pred.labs);
   	    for (int res = 0; res < restart; res++) {
   	    	// recover the starting
   	    	copyDepLab(pred, pred.parserPredDep, pred.parserPredLab);
   	    	for (int i = 1; i < len; i++)
   	    		samePosAsPred = updatePos(pred, i, 0, samePosAsPred);
   	    	if (samePosAsPred != len - 1) {
   	    		System.out.println("decode bug 0");
   	    		System.exit(0);
   	    	}
   	    	pipe.buildChild(pred);
   	    	
   	    	double T = initTemperature;
   	    	// double T = 0.0;
   	    	double currScore = 0.0;
   			if (sampling) {
   				if (DependencyParser.pruningMode == 0) {
   					initSample(pred, lfd);
   				}
   				// initial state, POS are initial, should be safe
   				//currScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
	   	    	for (int iter = 0; iter < maxIter; iter++) {
	   	    		boolean change = false;
	   	    		for (int z = 0; z < miniIter; z++) {
	   	        		for (int i = 1; i < len; i++) {
	   	        			// sample a dep
	   	        			//ArrayList<Double> score = new ArrayList<Double>();
	   	        			ArrayList<Integer> dep = new ArrayList<Integer>();
	   	        			ArrayList<Integer> lab = new ArrayList<Integer>();
	   	        			ArrayList<Integer> pos = new ArrayList<Integer>();
	   	        			
	   	    				int oldDep = pred.dep[i];
	   	    				int oldLab = pred.labs[i];
	   	    				int oldCandPID = pred.currPosCandID[i];
	   	    				int sample = -1;
	
	   	    				if (samplingMethod == 0) {
	   	    					//System.out.println("not supported!");
	   	    					ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
	   	    					ArrayList<Double> score = new ArrayList<Double>();
	   	        				sample = gibbsSamplingWithPos(pred, i, dep, lab, pos, samePosAsPred, fvList, score, 
	   	        						lfd, gfd, params, T, false, dr);
	   	        			
	   	        				if (dep.get(sample) != pred.dep[i] || lab.get(sample) != pred.labs[i]
	   	        						|| pos.get(sample) != oldCandPID) {
   	        						pred.dep[i] = dep.get(sample); pred.labs[i] = lab.get(sample);
   	        						pipe.updateChildList(pred, dep.get(sample), oldDep, i);
   	        						samePosAsPred = updatePos(pred, i, pos.get(sample), samePosAsPred);
   	        						change = true;
	   	        				}
	   	    				}
	   	        			else if (samplingMethod == 1) {
	   	        				sample = mhSamplingWithPos(pred, i, dep, lab, pos, samePosAsPred, lfd, dr);
	   	        				
	   	                		if (oldDep != dep.get(sample) || oldLab != lab.get(sample)
	   	                				|| oldCandPID != pos.get(sample)) {
	   	                			
			   	    				double oldScore = 0.0;
			   	    				if (samePosAsPred == len - 1) {
			   	    					oldScore = pipe.getLocalScore(pred, lfd, params, i) 
			   	        						+ pipe.getGlobalScore(pred, gfd, params);
			   	    				}
			   	    				else {
			   	        				FeatureVector fv = pipe.createFeatureVector(pred, lfd, i);
			   	        				oldScore = params.getScore(fv);
			   	    				}
			
			   	        			pred.dep[i] = dep.get(sample);
			   	        			pred.labs[i] = lab.get(sample);
			   	        			pipe.updateChildList(pred, dep.get(sample), oldDep, i);
			   	        			samePosAsPred = updatePos(pred, i, pos.get(sample), samePosAsPred);
			   	        			double newScore = 0.0;
			   	        			if (samePosAsPred == len - 1) {
			   	        				newScore = pipe.getLocalScore(pred, lfd, params, i) 
			   	        						+ pipe.getGlobalScore(pred, gfd, params);
			   	        			}
			   	        			else {
			   	        				FeatureVector fv = pipe.createFeatureVector(pred, lfd, i);
			   	        				newScore = params.getScore(fv);
			   	        			}

			   	        			// transition function for simulate annealing
			   	           			if (!accept(newScore, oldScore, T, dr)) {
			   	            			// not accept
				   	        			int oldH = pred.dep[i];
			   	            			pred.dep[i] = oldDep;
			   	            			pred.labs[i] = oldLab;
				   	        			pipe.updateChildList(pred, oldDep, oldH,i);
				   	        			samePosAsPred = updatePos(pred, i, oldCandPID, samePosAsPred);
			   	            		}
			   	            		else {
			   	            			// accept
		   	                			change = true;
			   	            		}

	   	                		}
	   	        			}
	   	           			
	   	        		}
	   	        		//if (DependencyParser.isSpmrl && !change)
	   	        		//	break;
	   	    		}	// end of mini iter
	   	    		
	   	    		
	   	    		
		   	    	// get currScore
	       			if (samePosAsPred == len - 1) {
	       				currScore = pipe.getLocalScore(pred, lfd, params, -1) 
	       						+ pipe.getGlobalScore(pred, gfd, params);
	       			}
	       			else {
	       				FeatureVector fv = pipe.createFeatureVector(pred, lfd, -1);
	       				currScore = params.getScore(fv);
	       			}
	       	    	if (currScore > bestScore) {
	       	    		bestScore = currScore;
	       	    		// fv
	       	    		if (sampling) {
	       	    			d[0][0] = pipe.createFeatureVector(pred, lfd, gfd, -1);
	       	    		}
	       	    		d[0][1] = getParseString(pred);
	       	    		d[0][2] = new Double(currScore);
	       	    	}

	   	    		if (!change) {
	   	    			//System.out.println("iter: " + iter);
	   	    			break;
	   	    		}
	   	    		T = T * coolingRate;
	   	    	
	   	    	}	// end of iter
	   	    	
   			}
   	    	if (currScore > bestScore) {
   	    		bestScore = currScore;
   	    		// fv
   	    		if (sampling) {
   	    			d[0][0] = pipe.createFeatureVector(pred, lfd, gfd, -1);
   	    		}
   	    		d[0][1] = getParseString(pred);
   	    		d[0][2] = new Double(currScore);
   	    	}
   	    }
    	
    	return d;
    }
    
    // decoding by sampling
    public Object[][] decodeSamplingRW(DependencyInstance pred, 
    		LocalFeatureData lfd, GlobalFeatureData gfd,
    		Parameters params, Parameters firstOrderParams,
    		boolean sampling) {
    	Random dr = new Random(seed);
    	
   	    int len = pred.sentence.length;
    	
   		double initTemperature = 1.0;
   		if (sampling) {
   			initTemperature = getInitTemp(pred, lfd, gfd, params);
   			//initTemperature /= 3;
   		}
   	    double coolingRate = 0.95;
   	    int maxIter = 300;
   	    int miniIter = 20;
   	    //if (pipe.lang.equals("spmrl") || pipe.lang.equals("arabic"))
   	    //	miniIter = 30;
   	    int restart = 4;
   	    //if (pipe.lang.equals("arabic") || pipe.lang.equals("slovene") || DependencyParser.isSpmrl)
   	    //	restart = 1;
   	    if (!sampling)
   	    	restart = 1;
    	
    	Object[][] d = new Object[1][3];
    	double bestScore = -1e30;
    	int iterCount = 0;
    	int bestIter = 0;
    	long bestTime = 0;
    	
    	// build score table
    	double[][] scoreTable = new double[len][len];
    	for (int m = 1; m < len; m++)
    		for (int h = 0; h < len; h++) {
    			if (!lfd.cand[h][m])
    				continue;
    			int small = h < m ? h : m;
    			int large = h > m ? h : m;
    			boolean attR = h < m ? true : false;
				lfd.fvs[h][m] = pipe.createFeatureVector(pred, small, large, attR, end);
				scoreTable[h][m] = params.getScore(lfd.fvs[h][m]);
				lfd.probs[h][m] = params.getScore(lfd.fvs[h][m]);
    		}
    	
    	// begin sampling
    	pipe.copyPredDepLab(pred, pred.dep, pred.labs);
   	    for (int res = 0; res < restart; res++) {
   	    	// recover the starting
   	    	copyDepLab(pred, pred.parserPredDep, pred.parserPredLab);
   	    	pipe.buildChild(pred);
   	    	
   	    	double T = initTemperature;
   	    	// double T = 0.0;
   	    	double currScore = 0.0;
   			if (sampling) {
   				if (DependencyParser.pruningMode == 0) {
   					initSample(pred, lfd);
   				}
   				currScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
   				//currScore = params.getScore(pipe.createFeatureVector(pred, lfd, gfd, -1));
   				
   				long start = System.currentTimeMillis();
	   	    	for (int iter = 0; iter < maxIter; iter++) {
	   	    		//System.out.print(".");
	   	    		boolean change = false;
	   	    		for (int z = 0; z < miniIter; z++) {
		   	      		int[] idx = new int[len];
		   	     		getBottomUpOrder(pred, 0, idx, len - 1);
		   	     		
	   	        		for (int y = 1; y < len; y++) {
	   	        			int id = idx[y];
	   	        			
		   	    			int[] sampleDep = new int[len];
		   	    			boolean[] toBeSampled = new boolean[len];
		   	    			toBeSampled[id] = true;
		   	    			//int numChange = 1;
		   	    			/*
		   	    			int numChange = dr.nextInt(3) + 1;
		   	    			if (numChange >= len - 1) {
		   	    				for (int i = 1; i < len; i++)
		   	    					toBeSampled[i] = true;
		   	    			}
		   	    			else {
			   	    			int pointer = 0;
			   	    			for (int i = 0; i < numChange; i++) {
			   	    				int jump = dr.nextInt(len);
			   	    				pointer = (pointer + jump) % len + 1;
			   	    				if (pointer >= len)
			   	    					pointer = 1;
			   	    				while (toBeSampled[pointer]) {
			   	    					pointer++;
			   	    					if (pointer >= len)
			   	    						pointer = 1;
			   	    				}
			   	    				if (toBeSampled[pointer])
			   	    					System.out.println("build to be sampled 1: " + pointer);
			   	    				toBeSampled[pointer] = true;
			   	    			}
		   	    			}
		   	    			*/
		   	    			ArrayList<Integer> vio = new ArrayList<Integer>();
		   	    			for (int h = 0; h < len; h++) {
		   	    				if (!lfd.cand[h][id])
		   	    					continue;
		   	    				if (isDescendant(h, id, pred.dep))
		   	    					//toBeSampled[h] = true;
		   	    					vio.add(h);
		   	    			}
		   	    			if (vio.size() > 0 && dr.nextInt(2) == 1) {
			   	    			for (int i = 0; i < 2; i++) {
			   	    				toBeSampled[vio.get(dr.nextInt(vio.size()))] = true;
			   	    			}
		   	    			}
		   	    			
		   	    			//toBeSampled[id] = true;
		   	    			//int h = pred.dep[id];
		   	    			//if (r.nextDouble() > 0.5)
		   	    			//toBeSampled[h] = true;
		   	    			//for (int i = 0; i < pred.child[id].length; i++)
		   	    			//	toBeSampled[pred.child[id][i]] = true;
		   	    			//System.out.print("#");
		   	    			//randomWalkSampling(pred, null, lfd, params, dr, toBeSampled, sampleDep);
		   	    			randomWalkSampling(pred, null, lfd, scoreTable, dr, toBeSampled, 1.0, sampleDep);
		   	    			/*
		   	    			for (int i = 1; i < pred.dep.length; i++) {
		   	    				sampleDep[i] = pred.dep[i];
		   	    			}
		   	    			ArrayList<Double> score = new ArrayList<Double>();
		   	    			ArrayList<Integer> depList = new ArrayList<Integer>();
		   	    			for (int h = 0; h < pred.dep.length; h++) {
		   	    				if (!lfd.cand[h][id])
		   	    					continue;
		   	    				if (isDescendant(h, id, pred.dep))
		   	    					continue;
		   	    				depList.add(h);
		   	    				score.add(scoreTable[h][id]);
		   	    			}
		   	    			int sample = samplePoint(score, T, dr);
		   	    			sampleDep[id] = depList.get(sample);
		   	    			*/
		   	    			//System.out.print("@");
		   	    			//for (int i = 0; i < pred.dep.length; i++) {
		   	    			//	System.out.println(i + "\t" + pred.dep[i]);
		   	    			//}
		   	    			//System.exit(0);
		   	    			
		   	    			int[] oldDep = new int[len];
		   	    			for (int i = 1; i < len; i++)
		   	    				oldDep[i] = pred.dep[i];
					   	    
		   	    			double oldProposalScore = 0.0;
		   	    			double newProposalScore = 0.0;
		   	    			
		   	    			for (int i = 1; i < len; i++) {
		   	    				//if (!toBeSampled[i] && pred.dep[i] != sampleDep[i]) {
		   	    				//	System.out.println("decode sampling rw bug 1");
		   	    				//}
		   	    				if (toBeSampled[i]) {
		   	    					oldProposalScore += scoreTable[pred.dep[i]][i];
		   	    					newProposalScore += scoreTable[sampleDep[i]][i];
		   	    				}
					   	        pred.dep[i] = sampleDep[i];
		   	    			}
	   	        			pipe.buildChild(pred);
	   	        			double newScore = pipe.getLocalScore(pred, lfd, params, -1) + pipe.getGlobalScore(pred, gfd, params);
	   	        			//double newScore = params.getScore(pipe.createFeatureVector(pred, lfd, gfd, -1));
					   	        			
	  	        			// transition function for simulate annealing
	   	           			//if (!accept(newScore, currScore, T, dr)) {
	   	           			if (!accept(newScore/* + oldProposalScore*/, 
	   	           					currScore/* + newProposalScore*/, T, dr)) {
	   	            			// not accept
	   		   	    			for (int i = 1; i < len; i++) {
	   					   	        pred.dep[i] = oldDep[i];
	   		   	    			}
	   	   	        			pipe.buildChild(pred);
	   	            		}
	  	            		else {
	   	            			// accept
	  	            			currScore = newScore;
	   		   	    			for (int i = 1; i < len; i++) {
	   		   	    				if (oldDep[i] != pred.dep[i]) {
	   		   	    					change = true;
	   		   	    					break;
	   		   	    				}
	   		   	    			}
	  	            		}
		   	           	
	   	           			iterCount++;
	   		   	    		if (currScore > bestScore) {
	   		   	    			bestScore = currScore;
	   		   	    			bestIter = iterCount;
	   		   	    			bestTime = System.currentTimeMillis() - start;
	   		   	    			//change = true;
	   		   	    			// fv
	   		   	    			if (sampling) {
	   		   	    				//d[0][0] = pipe.createFeatureVector(pred, lfd, gfd, -1);
	   		   	    			}
	   		   	    			d[0][1] = getParseString(pred);
	   		   	    			d[0][2] = new Double(currScore);
	   		   	    		}

	   	        		}
	   	           			
   	        		}
	   	        		
	   	    		if (T < 0.01 && !change) {
	   	    			//System.out.println("iter: " + iter);
	   	    			break;
	   	    		}
	   	    		T = T * coolingRate;
	   	    	}
   			}
   	    }
    	
   	    //System.out.println("best time: " + bestTime);
    	return d;
    }
    
    public void cycleErase(int[] dep, int i) {
    	while (dep[i] != -1) {
    		int next = dep[i];
    		dep[i] = -1;
    		i = next;
    	}
    }
    
    public void cycleErase(int[] dep, int i, boolean[] toBeSampled) {
    	while (dep[i] != -1) {
    		if (toBeSampled[i]) {
    			int next = dep[i];
    			dep[i] = -1;
        		i = next;
    		}
    		else {
    			i = dep[i];
    		}
    	}
    }
    
    public boolean sameCollapsedNode(int mod, int head, DependencyInstance pred, boolean[] toBeSampled) {
    	// check if mod and head are in the same collapsed node, avoid self loop
    	// toBeSampled[mod] should always be true
    	int curr = head;
    	while (curr != -1 && curr != mod && !toBeSampled[curr]) {
    		curr = pred.dep[curr];
    	}
    	return curr == mod;
    }
    
    public void randomWalkSampling(DependencyInstance pred, DependencyInstance gold, 
    		LocalFeatureData lfd, 
    		//Parameters params,
    		double[][] scoreTable,
    		Random r,
    		boolean[] toBeSampled, double T, int[] dep) {
    	// basically, the subtrees which is not to be sampled will be collapsed into one node
    	int len = pred.dep.length;
    	//double[][] scoreTable = new double[len][len];
    	//boolean[][] visited = new boolean[len][len];
    	
    	boolean[] inTree = new boolean[len];
    	inTree[0] = true;
    	for (int i = 0; i < len; i++) {
    		if (!toBeSampled[i]) {
    			dep[i] = pred.dep[i];
    		}
    		else {
    			dep[i] = -1;
    		}
    	}
    	
    	for (int i = 1; i < len; i++) {
    		int curr = i;
    		//int loopTime = 0;
    		while (!inTree[curr]) {
    			//System.out.println(curr);
    			if (!toBeSampled[curr]) {
    				curr = dep[curr];
    			}
    			else {
    				// sample new head
        			ArrayList<Double> score = new ArrayList<Double>();
        			ArrayList<Integer> depList = new ArrayList<Integer>();
        			
        			for (int candH = 0; candH < len; candH++) {
        				if (!lfd.cand[candH][curr])
        					continue;
        				if (sameCollapsedNode(curr, candH, pred, toBeSampled))
        					continue;
        				depList.add(candH);
        				//if (!visited[candH][curr]) {
        				//	int small = candH < curr ? candH : curr;
        				//	int large = candH > curr ? candH : curr;
        				//	boolean attR = candH < curr ? true : false;
       					//	if (lfd.fvs[candH][curr] == null) {
       					//		lfd.fvs[candH][curr] = pipe.createFeatureVector(pred, small, large, attR, end);
       					//	}
       					//	scoreTable[candH][curr] = params.getScore(lfd.fvs[candH][curr]);
       					//	visited[candH][curr] = true;
        				//}
       					score.add(scoreTable[candH][curr]);
        			}
        			
        			int sample = samplePoint(score, T, r);
        			dep[curr] = depList.get(sample);
        			curr = dep[curr];
    			}
    			
    			if (dep[curr] != -1 && toBeSampled[curr] && !inTree[curr]) {
    				//cycleErase(dep, curr);
    				cycleErase(dep, curr, toBeSampled);
    				//loopTime++;
    				//if (loopTime % 100 == 0)
    				//	System.out.print("loop:" + loopTime + " ");
    			}
    		}
    		curr = i;
    		while (!inTree[curr]) {
    			inTree[curr] = true;
    			curr = dep[curr]; 
    		}
    	}
    }
    
    public int gibbsSampling(DependencyInstance pred, DependencyInstance gold, int i, 
    		ArrayList<Integer> dep, ArrayList<Integer> lab, 
    		ArrayList<FeatureVector> fvList, ArrayList<Double> scoreList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, 
    		Parameters params, double temp, boolean isTrain, Random r) {
    	// in the training, params is updates, so cannot use old lfd directly
    	
    	//System.out.println("begin sampling: " + lfd.cand[pred.dep[i]][i]);
    	int len = pred.dep.length;
    	
		int oldDep = pred.dep[i];
		int oldLab = pred.labs[i];

		// build score
		for (int candH = 0; candH < len; candH++) {
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			int oldH = pred.dep[i];
			pred.dep[i] = candH;
			pipe.updateChildList(pred, candH, oldH, i);
			//pipe.buildChild(pred);
			for (int candT = 1; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				dep.add(candH);
				lab.add(candT);
				pred.labs[i] = candT;
				
				double s = 0.0;
				if (isTrain) {
					FeatureVector fv = null;
					fv = pipe.createFeatureVector(pred, lfd, gfd, i);
					s = params.getScore(fv);
					// cost augmented
					//s = params.getScore(fv) + (candH == gold.dep[i] ? 0.0 : 1.0);
					fvList.add(fv);
				}
				else {
					s = pipe.getLocalScore(pred, lfd, params, i) + pipe.getGlobalScore(pred, gfd, params);
				}
				scoreList.add(s);
			}
		}
		
		int oldH = pred.dep[i];
		pred.dep[i] = oldDep;
		pred.labs[i] = oldLab;
		pipe.updateChildList(pred, oldDep, oldH, i);
		//pipe.buildChild(pred);
		if (scoreList.size() == 0) {
			System.out.println("aaa");
		}
		
		// sample
		int sample = samplePoint(scoreList, temp, r);
		
		if (sample == -1) {
			for (int j = 0; j < scoreList.size(); j++)
				System.out.print(scoreList.get(j) + " ");
			System.out.println();
			System.out.println("bad sampling: " + lfd.cand[pred.dep[i]][i]);
			System.out.println(temp);
		}
    	
		return sample;
    }
    
    public int gibbsSamplingWithPos(DependencyInstance pred, int i, 
    		ArrayList<Integer> dep, ArrayList<Integer> lab, 
    		ArrayList<Integer> pos, int samePosAsPred,
    		ArrayList<FeatureVector> fvList, ArrayList<Double> scoreList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, 
    		Parameters params, double temp, boolean isTrain, Random r) {
    	// in the training, params is updates, so cannot use old lfd directly
    	
    	//System.out.println("begin sampling: " + lfd.cand[pred.dep[i]][i]);
    	int len = pred.dep.length;
    	
		int oldDep = pred.dep[i];
		int oldLab = pred.labs[i];
		int oldPosCandID = pred.currPosCandID[i];
		int oldSame = samePosAsPred;

		// build score
		for (int candH = 0; candH < len; candH++) {
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			int oldH = pred.dep[i];
			pred.dep[i] = candH;
			pipe.updateChildList(pred, candH, oldH, i);
			//pipe.buildChild(pred);
			for (int candT = 1; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				pred.labs[i] = candT;
				
				for (int candPID = 0; candPID < pred.candPos[i].length; candPID++) {
					dep.add(candH);
					lab.add(candT);
					pos.add(candPID);
					samePosAsPred = updatePos(pred, i, candPID, samePosAsPred);
					
					double s = 0.0;
					if (isTrain) {
						FeatureVector fv = null;
						if (samePosAsPred == len - 1) {
							fv = pipe.createFeatureVector(pred, lfd, gfd, i);
						}
						else {
							fv = pipe.createFeatureVector(pred, lfd, i);
						}
							
						s = params.getScore(fv);
						fvList.add(fv);
					}
					else {
						if (samePosAsPred == len - 1) {
							s = pipe.getLocalScore(pred, lfd, params, i) + pipe.getGlobalScore(pred, gfd, params);
						}
						else {
							FeatureVector fv = pipe.createFeatureVector(pred, lfd, i);
							s = params.getScore(fv);
						}
					}
					scoreList.add(s);
				}
			}
		}
		
		int oldH = pred.dep[i];
		pred.dep[i] = oldDep;
		pred.labs[i] = oldLab;
		pipe.updateChildList(pred, oldDep, oldH, i);
		samePosAsPred = updatePos(pred, i, oldPosCandID, samePosAsPred);
		if (samePosAsPred != oldSame) {
			System.out.println("gibbs sampling with pos bug 1");
		}
		//pipe.buildChild(pred);
		if (scoreList.size() == 0) {
			System.out.println("aaa");
		}
		
		// sample
		int sample = samplePoint(scoreList, temp, r);
		
		if (sample == -1) {
			for (int j = 0; j < scoreList.size(); j++)
				System.out.print(scoreList.get(j) + " ");
			System.out.println();
			System.out.println("bad sampling: " + lfd.cand[pred.dep[i]][i]);
			System.out.println(temp);
		}
    	
		return sample;
    }
    
    public int updatePos(DependencyInstance inst, int arg, int candPID, int samePosAsPred) {
    	if (inst.currPosCandID[arg] == 0)
    		samePosAsPred--;
    	inst.pos[arg] = inst.candPos[arg][candPID];
    	inst.posA[arg] = inst.candCpos[arg][candPID];
    	inst.specialPos[arg] = inst.candSpecialPos[arg][candPID];
    	inst.currPosCandID[arg] = candPID;
    	if (inst.currPosCandID[arg] == 0)
    		samePosAsPred++;
    	return samePosAsPred;
    }
    
    public int mhSampling(DependencyInstance pred, int i, ArrayList<Integer> dep, ArrayList<Integer> lab,
    		LocalFeatureData lfd, Random r) {
    	int len = pred.dep.length;
    	// uniform sampling
		for (int candH = 0; candH < len; candH++) {
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			for (int candT = 1; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				dep.add(candH);
				lab.add(candT);
			}
		}
		int sample = r.nextInt(dep.size());
		return sample;
    }
    
    public int mhSamplingWithPos(DependencyInstance pred, int i, 
    		ArrayList<Integer> dep, ArrayList<Integer> lab,
    		ArrayList<Integer> pos, int samePosAsPred,
    		LocalFeatureData lfd, Random r) {
    	int len = pred.dep.length;
    	// uniform sampling
		for (int candH = 0; candH < len; candH++) {
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			for (int candT = 1; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				for (int candPID = 0; candPID < pred.candPos[i].length; candPID++) {
					int z = 1;
					if (candPID == 0) {
						// 6 times?
						z = 6;
					}
					for (int y = 0; y < z; y++) {
						dep.add(candH);
						lab.add(candT);
						pos.add(candPID);
					}
				}
			}
		}
		int sample = r.nextInt(dep.size());
		return sample;
    }
    
    public int compare(int goldDep, int goldLab, int oldDep, int oldLab, int newDep, int newLab) {
    	int oldScore = (oldDep == goldDep ? 1 : 0) + (oldDep == goldDep && oldLab == goldLab ? 1 : 0);
    	int newScore = (newDep == goldDep ? 1 : 0) + (newDep == goldDep && newLab == goldLab ? 1 : 0);
    	return newScore - oldScore;
    }
    
    public int compare(int goldDep, int goldLab, int goldPos, 
    		int oldDep, int oldLab, int oldPos,
    		int newDep, int newLab, int newPos) {
    	int oldScore = (oldPos == goldPos ? 1 : 0) 
    			+ (oldDep == goldDep ? 1 : 0) 
    			+ (oldDep == goldDep && oldLab == goldLab ? 1 : 0);
    	int newScore = (newPos == goldPos ? 1 : 0) 
    			+ (newDep == goldDep ? 1 : 0) 
    			+ (newDep == goldDep && newLab == goldLab ? 1 : 0);
    	return newScore - oldScore;
    }
    
    public int compareMIRA(int goldDep, String goldLab, int oldDep, String oldLab, int newDep, String newLab) {
    	int oldScore = (oldDep == goldDep ? 1 : 0) + (oldLab.equals(goldLab) ? 1 : 0);
    	int newScore = (newDep == goldDep ? 1 : 0) + (newLab.equals(goldLab) ? 1 : 0);
    	return newScore - oldScore;
    }
    
    public double numError(int[] goldDep, int[] goldLab, int[] predDep, int[] predLab) {
    	double e = 0.0;
    	for (int i = 0; i < goldDep.length; i++) {
    		if (goldDep[i] != predDep[i])
    			e += 1.0;
    		else if (goldLab[i] != predLab[i])
    			e += 0.5;
    	}
    	return e;
    }
    
    public double numError(int[] goldDep, int[] goldLab, int[] goldPos,
    		int[] predDep, int[] predLab, int[] predPos) {
    	double e = 0.0;
    	for (int i = 0; i < goldDep.length; i++) {
    		if (goldDep[i] != predDep[i])
    			e += 2.0;
    		else if (goldLab[i] != predLab[i])
    			e += 1.0;
    		if (goldPos[i] != predPos[i])
    			e += 1.0;
    	}
    	return e;
    }
    
    public int findOldSample(int oldDep, int oldLab, ArrayList<Integer> dep, ArrayList<Integer> lab) {
    	for (int i = 0; i < dep.size(); i++) {
    		if (oldDep == dep.get(i) && oldLab == lab.get(i)) {
    			return i;
    		}
    	}
    	System.out.println("can't find old sample");
    	return -1;
    }
    
    public int findOldSample(int oldDep, int oldLab, int oldPID,
    		ArrayList<Integer> dep, ArrayList<Integer> lab, ArrayList<Integer> pos) {
    	for (int i = 0; i < dep.size(); i++) {
    		if (oldDep == dep.get(i) && oldLab == lab.get(i) && oldPID == pos.get(i)) {
    			return i;
    		}
    	}
    	System.out.println("can't find old pos sample");
    	return -1;
    }
    
    public void debugFVList(ArrayList<FeatureVector> fvList) {
    	for (int i = 0; i < fvList.size(); i++) {
    		FeatureVector fv = fvList.get(i);
    		int count = 0;
    		for (FeatureVector curr = fv; curr.next != null; curr = curr.next) {
    			count++;
    		}
    		System.out.print(count + "\t");
    	}
    	System.out.println();
    	//System.exit(0);
    }
    
    public void sampleRank(DependencyInstance gold, DependencyInstance pred, 
		   LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params, int iters) {
    	// sample rank training
    	int len = pred.sentence.length;
		int maxIter = 1;
   	    String samplingMethod = "gibbs";
   	    //if (DependencyParser.pruningMode == 0)
   	    //	String samplingMethod = "MH";
		double T = 1.0;
		//double T = 0.0;
		
    	for (int iter = 0; iter < maxIter; iter++) {
    		
     		int[] idx = new int[len];
     		int id = getBottomUpOrder(pred, 0, idx, len - 1);
     		if (id != 0) {
     			System.out.println("build order bug");
     		}

     		for (int z = 1; z < len; z++) {
     			int i = idx[z];
     			
    			ArrayList<Integer> dep = new ArrayList<Integer>();
    			ArrayList<Integer> lab = new ArrayList<Integer>();
    			ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
    			ArrayList<Double> scoreList = new ArrayList<Double>();

				int sample = -1;

				if (samplingMethod.equals("gibbs")) {
					sample = gibbsSampling(pred, gold, i, dep, lab, 
							fvList, scoreList, lfd, gfd, params, T, true, r);
    			}
    			else if (samplingMethod.equals("MH")) {
    				System.out.println("not supported!");
    				sample = mhSampling(pred, i, dep, lab, lfd, r);
    			}
				
				//if (updateTimes < 20)
				//	debugFVList(fvList);
				
				if (sample == -1)
					continue;
				
				int oldDep = pred.dep[i];
				int oldLab = pred.labs[i];

				int newDep = dep.get(sample);
				int newLab = lab.get(sample);
				
				int oldSample = findOldSample(oldDep, oldLab, dep, lab);
				
				if (oldSample == -1) {
					System.exit(0);
				}
				
				if (oldDep != newDep || oldLab != newLab) {
					
					FeatureVector oldFV = null;
					double oldScore = 0.0;
					FeatureVector newFV = null;
					double newScore = 0.0;
					
					if (scoreList.size() != 0) {
						oldFV = fvList.get(oldSample);
						oldScore = scoreList.get(oldSample);
						//oldScore = params.getScore(oldFV);
						
						int oldH = pred.dep[i];
						pred.dep[i] = newDep; pred.labs[i] = newLab;
						pipe.updateChildList(pred, newDep, oldH, i);
						//pipe.buildChild(pred);
						newFV = fvList.get(sample);
						newScore = scoreList.get(sample);
						//newScore = params.getScore(newFV);
					}
					else {
						oldFV = pipe.createFeatureVector(pred);
						oldScore = params.getScore(oldFV);
						
						int oldH = pred.dep[i];
						pred.dep[i] = newDep; pred.labs[i] = newLab;
						pipe.updateChildList(pred, newDep, oldH, i);
						//pipe.buildChild(pred);
						newFV = pipe.createFeatureVector(pred);
						newScore = params.getScore(newFV);
					}
					
					int comp = compare(gold.dep[i], gold.labs[i], oldDep, oldLab, newDep, newLab);
						
					//if (!DependencyParser.updateGold) {
						if (comp > 0 && newScore - oldScore < comp) {
							// update toward new tree
							double upd = (double)updateTimes;
							
			    			if (DependencyParser.learningMode == 0) {
			    				// perceptron
			    				params.updatePerceptron(newFV, oldFV, 1.0, upd);
			    			}
			    			else if (DependencyParser.learningMode == 1) {
			    				// mira
			    				params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, comp, upd);
			    			}
			    			//updateTimes++;
						}
						if (comp < 0 && oldScore - newScore < -comp) {
							// update toward old tree
							double upd = (double)updateTimes;
							
			    			if (DependencyParser.learningMode == 0) {
			    				// perceptron
			    				params.updatePerceptron(oldFV, newFV, 1.0, upd);
			    			}
			    			else if (DependencyParser.learningMode == 1) {
			    				// mira
			    				params.updateParamsMIRA(oldFV, newFV, oldScore, newScore, -comp, upd);
			    			}
			    			//updateTimes++;
						}
		    			//updateTimes++;
						
					//}
					//else {
						/*
						if (comp != 0) {
							FeatureVector[] targetFV = new FeatureVector[2];
							double[] targetScore = new double[2];
							FeatureVector[] currFV = new FeatureVector[2];
							double[] currScore = new double[2];
							double[] e = new double[2];
							if (comp > 0) {
								targetFV[0] = newFV;
								targetScore[0] = newScore;
								currFV[0] = oldFV;
								currScore[0] = oldScore;
								e[0] = comp;
							}
							else if (comp < 0) {
								targetFV[0] = oldFV;
								targetScore[0] = oldScore;
								currFV[0] = newFV;
								currScore[0] = newScore;
								e[0] = -comp;
							}
							targetFV[1] = gold.fv;
							targetScore[1] = params.getScore(gold.fv);
							if (scoreList.size() != 0) {
								currFV[1] = pipe.createFeatureVector(pred, lfd, gfd, -1);
								currScore[1] = params.getScore(currFV[1]);
							}
							else  {
								currFV[1] = newFV;
								currScore[1] = newScore;
							}
							e[1] = numError(gold.dep, gold.labs, pred.dep, pred.labs);
							upd = (double)updateTimes;
							
							// update separately or together?
							params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, e, upd);
							//if (targetScore[0] - currScore[0] < e[0])
							//	params.updateParamsMIRA(targetFV[0], currFV[0], targetScore[0], currScore[0], e[0], upd);
							//if (targetScore[1] - currScore[1] < e[1])
							//	params.updateParamsMIRA(targetFV[1], currFV[1], targetScore[1], currScore[1], e[1], upd);
						}
						*/
						/*
						else {
							// update to the gold
							FeatureVector targetFV = gold.fv;
							double targetScore = params.getScore(targetFV);
							FeatureVector currFV = null;
							double currScore = 0.0;
							if (scoreList.size() != 0) {
								currFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
								currScore = params.getScore(currFV);
							}
							else  {
								currFV = newFV;
								currScore = newScore;
							}
							double e = numError(gold.dep, gold.labs, pred.dep, pred.labs);
							if (targetScore - currScore < e)
								params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, e, upd);
						}
						*/
					//}
					
	    			// accept new proposal?
					if (samplingMethod.equals("gibbs")) {
	    				// always accept
						if (pred.dep[i] != newDep) {
							System.out.println("sample rank bug 1");
						}
					}
					else if (samplingMethod.equals("MH")) {
	    				// accept ratio
	    				if (accept(newScore, oldScore, T, r)) {
	    					// accept
	    				}
	    				else {
	    					// reject
	    					int oldH = pred.dep[i];
	    					pred.dep[i] = oldDep;
	    					pred.labs[i] = oldLab;
	    					pipe.updateChildList(pred, oldDep, oldH, i);
	    				}
					}
				}

				//updateTimes++;
				
    		}
    	}
    	
    	//updateTimes++;

    	if (DependencyParser.updateGold) {
			FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
			double oldScore = params.getScore(oldFV);
			
			FeatureVector newFV = gold.fv;
			double newScore = params.getScore(newFV);
			
			double err = numError(gold.dep, gold.labs, pred.dep, pred.labs);
			//double err = 2.0;
			if (newScore - oldScore < err) {
				double upd = (double)updateTimes;
				
    			if (DependencyParser.learningMode == 0) {
    				// perceptron
    				params.updatePerceptron(newFV, oldFV, 1.0, upd);
    			}
    			else if (DependencyParser.learningMode == 1) {
    				// mira
    				params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, err, upd);
    			}
			}
			//updateTimes++;
    	}
    	
    	updateTimes++;
   }
   
    public void sampleRankWithPos(DependencyInstance gold, DependencyInstance pred, 
 		   LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params, int iters) {
     	// sample rank training
     	int len = pred.sentence.length;
 		int maxIter = 1;
    	String samplingMethod = "gibbs";
 		double T = 1.0;
 		// double T = 0.0;
 		
 		int samePosAsPred = 0;
 		for (int i = 1; i < len; i++)
 			if (pred.currPosCandID[i] == 0)
 				samePosAsPred++;
     	
     	for (int iter = 0; iter < maxIter; iter++) {
     		for (int i = 1; i < len; i++) {
     			ArrayList<Integer> dep = new ArrayList<Integer>();
     			ArrayList<Integer> lab = new ArrayList<Integer>();
     			ArrayList<Integer> pos = new ArrayList<Integer>();
     			ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
     			ArrayList<Double> scoreList = new ArrayList<Double>();

 				int sample = -1;

 				if (samplingMethod.equals("gibbs")) {
 	  				sample = gibbsSamplingWithPos(pred, i, dep, lab, pos, samePosAsPred, 
    						fvList, scoreList, lfd, gfd, params, T, true, r);
     			}
     			else if (samplingMethod.equals("MH")) {
     				System.out.println("not supported!");
     				sample = mhSampling(pred, i, dep, lab, lfd, r);
     			}
 				
 				//if (updateTimes < 20)
 				//	debugFVList(fvList);
 				
 				if (sample == -1)
 					continue;
 				
 				int oldDep = pred.dep[i];
 				int oldLab = pred.labs[i];
 				int oldCandPID = pred.currPosCandID[i];

 				int newDep = dep.get(sample);
 				int newLab = lab.get(sample);
 				int newCandPID = pos.get(sample);
 				
 				int oldSample = findOldSample(oldDep, oldLab, oldCandPID, 
 						dep, lab, pos);
 				
 				if (oldSample == -1) {
 					System.exit(0);
 				}
 				
 				FeatureVector oldFV = null;
 				double oldScore = 0.0;
 				FeatureVector newFV = null;
 				double newScore = 0.0;
 				
 				if (scoreList.size() != 0) {
 					oldFV = fvList.get(oldSample);
 					oldScore = scoreList.get(oldSample);
 					
 					pred.dep[i] = newDep; pred.labs[i] = newLab;
 					samePosAsPred = updatePos(pred, i, newCandPID, samePosAsPred);
 					pipe.updateChildList(pred, newDep, oldDep, i);
 					//pipe.buildChild(pred);
 					newFV = fvList.get(sample);
 					newScore = scoreList.get(sample);
 				}
 				else {
 					System.out.println("should not go here 0");
 					System.exit(0);
 				}
 				
 				if (oldDep != newDep || oldLab != newLab || oldCandPID != newCandPID) {
 					
 					int comp = compare(gold.dep[i], gold.labs[i], gold.pos[i],
 							oldDep, oldLab, pred.candPos[i][oldCandPID],
 							newDep, newLab, pred.candPos[i][newCandPID]);
 						
 					//if (!DependencyParser.updateGold) {
 						if (comp > 0 && newScore - oldScore < comp) {
 							// update toward new tree
 							double upd = (double)updateTimes;
 							
 			    			if (DependencyParser.learningMode == 0) {
 			    				// perceptron
 			    				params.updatePerceptron(newFV, oldFV, 1.0, upd);
 			    			}
 			    			else if (DependencyParser.learningMode == 1) {
 			    				// mira
 			    				params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, comp, upd);
 			    			}
 			    			//updateTimes++;
 						}
 						if (comp < 0 && oldScore - newScore < -comp) {
 							// update toward old tree
 							double upd = (double)updateTimes;
 							
 			    			if (DependencyParser.learningMode == 0) {
 			    				// perceptron
 			    				params.updatePerceptron(oldFV, newFV, 1.0, upd);
 			    			}
 			    			else if (DependencyParser.learningMode == 1) {
 			    				// mira
 			    				params.updateParamsMIRA(oldFV, newFV, oldScore, newScore, -comp, upd);
 			    			}
 			    			//updateTimes++;
 						}
 		    			//updateTimes++;
  					
 	    			// accept new proposal?
 					if (samplingMethod.equals("gibbs")) {
 	    				// always accept
 						if (pred.dep[i] != newDep) {
 							System.out.println("sample rank bug 1");
 						}
 					}
 					else if (samplingMethod.equals("MH")) {
 	    				// accept ratio
 						System.out.println("should not go here 1");
 					}
 				}

 				//updateTimes++;
 				
     		}
     	}
     	
     	if (DependencyParser.updateGold) {
 			FeatureVector oldFV = null;
 			if (samePosAsPred == len - 1)
 				oldFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
 			else
 				oldFV = pipe.createFeatureVector(pred, lfd, -1);
 			double oldScore = params.getScore(oldFV);
 			
 			FeatureVector newFV = gold.fv;
 			double newScore = params.getScore(newFV);
 			
 			//double err = numError(gold.dep, gold.labs, gold.pos,
 			//		pred.dep, pred.labs, pred.pos);
 			double err = 1.0;
 			if (newScore - oldScore < err) {
 				double upd = (double)updateTimes;
 				
     			if (DependencyParser.learningMode == 0) {
     				// perceptron
     				params.updatePerceptron(newFV, oldFV, 1.0, upd);
     			}
     			else if (DependencyParser.learningMode == 1) {
     				// mira
     				params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, err, upd);
     			}
 			}
     	}
     	
     	updateTimes++;
    }
    
    public void sampleRankRW(DependencyInstance gold, DependencyInstance pred, 
 		   LocalFeatureData lfd, GlobalFeatureData gfd, 
 		   Parameters params, Parameters firstOrderParams, boolean restart, int iters) {
     	// sample rank training
     	int len = pred.sentence.length;
    	
     	// build score table
    	double[][] scoreTable = new double[len][len];
    	for (int m = 1; m < len; m++)
    		for (int h = 0; h < len; h++) {
    			if (!lfd.cand[h][m])
    				continue;
    			int small = h < m ? h : m;
    			int large = h > m ? h : m;
    			boolean attR = h < m ? true : false;
				lfd.fvs[h][m] = pipe.createFeatureVector(pred, small, large, attR, end);
				scoreTable[h][m] = params.getScore(lfd.fvs[h][m]);
				lfd.probs[h][m] = params.getScore(lfd.fvs[h][m]);
    		}
    	
    	if (restart) {
 			int[] sampleDep = new int[len];
 			boolean[] toBeSampled = new boolean[len];
 			for (int i = 0; i < len; i++) {
 				if (r.nextDouble() > 0.5)
 					toBeSampled[i] = true;
 			}
 			randomWalkSampling(pred, gold, lfd, scoreTable, r, toBeSampled, 1.0, sampleDep);
 			for (int i = 1; i < len; i++)
 				pred.dep[i] = sampleDep[i];
 			pipe.buildChild(pred);
    	}


    	double T = 1.0;
    	for (int i = 0; i < iters; i++)
    		T *= 0.99;
 		
 		int[] idx = new int[len];
 		getBottomUpOrder(pred, 0, idx, len - 1);
 		
    	for (int z = 1; z < len; z++) {

 			int[] sampleDep = new int[len];
 			boolean[] toBeSampled = new boolean[len];
 			//int numChange = 1;
 			/*
 			int numChange = r.nextInt(3) + 1;
 			int[] multipleChange = new int[numChange];
 			
   			if (numChange >= len - 1) {
   				for (int i = 1; i < len; i++) {
   					toBeSampled[i] = true;
   					multipleChange[i - 1] = i;
   				}
   			}
   			else {
	    			int pointer = 0;
	    			for (int i = 0; i < numChange; i++) {
	    				int jump = r.nextInt(len);
	    				pointer = (pointer + jump) % len + 1;
	    				if (pointer >= len)
	    					pointer = 1;
	    				while (toBeSampled[pointer]) {
	    					pointer++;
	    					if (pointer >= len)
	    						pointer = 1;
	    				}
	    				if (toBeSampled[pointer])
	    					System.out.println("build to be sampled 1: " + pointer);
	    				toBeSampled[pointer] = true;
	    				multipleChange[i] = pointer;
	    			}
   			}
   			*/
 			int id = idx[z];
 			toBeSampled[id] = true;
 			int changeNum = 1;

   			ArrayList<Integer> vio = new ArrayList<Integer>();
   			for (int h = 0; h < len; h++) {
   				if (!lfd.cand[h][id])
   					continue;
   				if (isDescendant(h, id, pred.dep))
   					//toBeSampled[h] = true;
   					vio.add(h);
   			}
   			if (vio.size() > 0 && r.nextInt(2) == 1) {
   				for (int i = 0; i < 1; i++) {
   					int j = vio.get(r.nextInt(vio.size()));
   					if (!toBeSampled[j])
   						changeNum++;
   					toBeSampled[j] = true;
   				}
   			}
 			
   			int[] multipleChange = new int[changeNum];
   			int p = 0;
   			for (int i = 0; i < toBeSampled.length; i++)
   				if (toBeSampled[i]) {
   					multipleChange[p] = i;
   					p++;
   				}
    		//for (int h = 0; h < len; h++) {
    		//	if (!lfd.cand[h][i])
    		//		continue;
			//	scoreTable[h][i] = params.getScore(lfd.fvs[h][i]);
    		//}
   			for (int i = 0; i < multipleChange.length; i++) {
   				int m = multipleChange[i];
   				for (int h = 0; h < len; h++) {
   					if (!lfd.cand[h][m])
   						continue;
   					scoreTable[h][m] = params.getScore(lfd.fvs[h][m]);
   				}
   			}

    		randomWalkSampling(pred, gold, lfd, scoreTable, r, toBeSampled, 1.0, sampleDep);

 			//int i = pointer;

 			//int oldDep = pred.dep[i];
 			//int oldLab = pred.labs[i];
    		double oldProposalScore = 0.0;
    		double oldAcc = 0.0;
    		
    		int[] oldDep = new int[len];
    		for (int i = 1; i < len; i++) {
    			oldDep[i] = pred.dep[i];
    			if (toBeSampled[i])
    				oldProposalScore += scoreTable[pred.dep[i]][i];
    			if (pred.dep[i] == gold.dep[i])
    				oldAcc += 2.0;
    		}

    		FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, multipleChange);
    		double oldScore = params.getScore(oldFV);
    		
    		double newProposalScore = 0.0;
    		double newAcc = 0.0;
    		for (int i = 1; i < len; i++) {
    			pred.dep[i] = sampleDep[i];
    			if (toBeSampled[i])
    				newProposalScore += scoreTable[pred.dep[i]][i];
    			if (pred.dep[i] == gold.dep[i])
    				newAcc += 2.0;
    		}
    		pipe.buildChild(pred);
    		FeatureVector newFV = pipe.createFeatureVector(pred, lfd, gfd, multipleChange);
    		double newScore = params.getScore(newFV);
    		
    		double comp = newAcc - oldAcc;
    		
 			//int newDep = sampleDep[i];
 			//int newLab = oldLab;
    		

 			//if (oldDep != newDep || oldLab != newLab) {
 				
 				//FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, i);
 				//double oldScore = params.getScore(oldFV);
 				
 				//pred.dep[i] = newDep; pred.labs[i] = newLab;
 				//pipe.updateChildList(pred, newDep, oldDep, i);

 				//FeatureVector newFV = pipe.createFeatureVector(pred, lfd, gfd, i);
 				//double newScore = params.getScore(newFV);

 				//int comp = compare(gold.dep[i], gold.labs[i], oldDep, oldLab, newDep, newLab);

 				if (comp > 0 && newScore - oldScore < comp) {
 	 				//System.out.println("update 1");

 					// update toward new tree
 					double upd = (double)updateTimes;

 					if (DependencyParser.learningMode == 0) {
 						// perceptron
 						params.updatePerceptron(newFV, oldFV, 1.0, upd);
 					}
 					else if (DependencyParser.learningMode == 1) {
 						// mira
 						params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, comp, upd);
 					}
 					//updateTimes++;
 				}
 				if (comp < 0 && oldScore - newScore < -comp) {
 	 				//System.out.println("update 2");
 					// update toward old tree
 					double upd = (double)updateTimes;

 					if (DependencyParser.learningMode == 0) {
 						// perceptron
 						params.updatePerceptron(oldFV, newFV, 1.0, upd);
 					}
 					else if (DependencyParser.learningMode == 1) {
 						// mira
 						params.updateParamsMIRA(oldFV, newFV, oldScore, newScore, -comp, upd);
 					}
 					//updateTimes++;
 				}
 				//updateTimes++;

 				// accept ratio
 				if (accept(newScore/* + oldProposalScore*/, oldScore/* + newProposalScore*/, T, r)) {
 					// accept
 				}
 				else {
 					// reject
 					//int oldH = pred.dep[i];
 					//pred.dep[i] = oldDep;
 					//pred.labs[i] = oldLab;
 					//pipe.updateChildList(pred, oldDep, oldH, i);
 					for (int i = 1; i < len; i++) {
 						pred.dep[i] = oldDep[i];
 					}
 					pipe.buildChild(pred);
 				}

 			//}
 		}
 
     	if (DependencyParser.updateGold) {
 			FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
 			double oldScore = params.getScore(oldFV);
 			
 			FeatureVector newFV = gold.fv;
 			double newScore = params.getScore(newFV);
 			
 			double err = numError(gold.dep, gold.labs, pred.dep, pred.labs);
 			//double err = 2.0;
 			if (newScore - oldScore < err) {
 				double upd = (double)updateTimes;
 				
     			if (DependencyParser.learningMode == 0) {
     				// perceptron
     				params.updatePerceptron(newFV, oldFV, 1.0, upd);
     			}
     			else if (DependencyParser.learningMode == 1) {
     				// mira
     				params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, err, upd);
     			}
 			}
 			//updateTimes++;
     	}
     	
     	updateTimes++;
    }
    
    public void initSample(DependencyInstance pred, LocalFeatureData lfd) {
    	Random r = new Random(seed);
    	int len = pred.dep.length;
    	for (int i = 1; i < len; i++) {
    		pred.dep[i] = 0;
    		pred.labs[i] = 1;
    	}
    	for (int i = 1; i < len; i++) {
    		ArrayList<Integer> dep = new ArrayList<Integer>();
    		ArrayList<Integer> lab = new ArrayList<Integer>();
    		int sample = mhSampling(pred, i, dep, lab, lfd, r);
    		pred.dep[i] = dep.get(sample);
    		pred.labs[i] = lab.get(sample);
    	}
    	pipe.buildChild(pred);
    }
   
   
   public boolean accept(double newScore, double oldScore, double temperature, Random r) {
    	if (newScore >= oldScore)
    		return true;
    	double accRatio = Math.exp((newScore - oldScore) / temperature);
    	if (r.nextDouble() < accRatio)
    		return true;
    	else
    		return false;
    }
    
	public static double logsumexp(double num1, double num2) {
        double max_exp = 0.0;
        double sum = 0.0;
        if (num2 > num1) {
                max_exp = num2;
                sum = 1.0 + Math.exp(num1 - max_exp);
        }
        else {
                max_exp = num1;
                sum = 1.0 + Math.exp(num2 - max_exp);
        }
        double ret = Math.log(sum) + max_exp;
        return ret;
	}

    public int samplePoint(ArrayList<Double> score, double temp, Random r) {
    	if (Math.abs(temp) < 1e-4) {
    		// select the best one
    		double maxScore = -1e30;
    		int ret = -1;
    		for (int i = 0; i < score.size(); i++)
    			if (score.get(i) > maxScore) {
    				maxScore = score.get(i);
    				ret = i;
    			}
    		return ret;
    	}
    	
    	double sumScore = -1e30;
    	for (int i = 0; i < score.size(); i++) {
    		score.set(i, score.get(i) * 1.0 / temp);
    		sumScore = logsumexp(sumScore, score.get(i));
    	}
    	double[] prob = new double[score.size()];
    	prob[0] = Math.exp(score.get(0) - sumScore);
    	for (int i = 1; i < score.size(); i++)
    		prob[i] = prob[i - 1] + Math.exp(score.get(i) - sumScore);
    	if (Math.abs(prob[prob.length - 1] - 1.0) > 1e-4) {
    		System.out.println("sample point bug");
    		System.exit(0);
    	}
    	double p = r.nextDouble();
    	int ret = 0;
    	for (; ret < prob.length; ret++) {
    		if (p < prob[ret])
    			break;
    	}
    	return ret;
    }
    
    // re-ranking
    public int trainReranking(DependencyInstance gold, DependencyInstance pred, 
    		FeatureVector[] fvList, int target, int oldSample,
    		int[][] depList, int[][] labList, Parameters params, int iters) {
    	// decoding
    	ArrayList<Double> score = new ArrayList<Double>();
    	for (int i = 0; i < fvList.length; i++) {
    		if (fvList[i] == null)
    			continue;
    		double s = params.getScore(fvList[i]);
    		score.add(s);
    	}
    	
    	// update parameter
    	int best = -1;
    	if (DependencyParser.learningMode == 2) {
    		//top-k sample rank 
    		best = samplePoint(score, 1.0, r);
    		double upd = (double)updateTimes;
    		double newError = numError(gold.dep, gold.labs, depList[best], labList[best]);
    		double oldError = numError(gold.dep, gold.labs, depList[oldSample], labList[oldSample]);
    		double comp = oldError - newError;
    		
    		double newScore = score.get(best);
    		double oldScore = score.get(oldSample);
    		
    		if (comp > 0 && newScore - oldScore < comp) {
    			// update towards new
    			params.updateParamsMIRA(fvList[best], fvList[oldSample], newScore, oldScore, comp, upd);
    		}
    		else if (comp < 0 && oldScore - newScore < -comp) {
    			// update towards old
    			params.updateParamsMIRA(fvList[oldSample], fvList[best], oldScore, newScore, -comp, upd);
    		}
    		
    		if (DependencyParser.updateGold) {
    			double error = newError;
    			newScore = params.getScore(gold.fv);
    			oldScore = score.get(best);
    			if (newScore - oldScore < error) 
    				params.updateParamsMIRA(gold.fv, fvList[best], newScore, oldScore, error, upd);
    		}
    	}
    	else if (DependencyParser.learningMode == 3) {
    		// top-k exact decoding mira update
    		double maxScore = -1e30;
    		for (int i = 0; i < score.size(); i++) {
    			double augScore = score.get(i) + numError(gold.dep, gold.labs, depList[i], labList[i]);
    			if (augScore > maxScore) {
    				best = i;
    				maxScore = augScore;
    			}
    		}
    		double upd = (double)updateTimes;
    		double error = numError(depList[target], labList[target], depList[best], labList[target]);
    		double targetScore = score.get(target);
    		double currScore = score.get(best);
    		if (targetScore - currScore < error)
    			params.updateParamsMIRA(fvList[target], fvList[best], targetScore, currScore, error, upd);
    	}
    	updateTimes++;
    	
    	// update prediction
    	for (int i = 0; i < pred.dep.length; i++) {
    		pred.dep[i] = depList[best][i];
    		pred.labs[i] = labList[best][i];
    	}
    	
    	return best;
    }
    
    public Object[][] decodeReranking(DependencyInstance pred, 
    		FeatureVector[] fvList, int[][] depList, int[][] labList, Parameters params) {
    	// decoding
    	double maxScore = -1e30;
    	int maxID = -1;
    	for (int i = 0; i < fvList.length; i++) {
    		if (fvList[i] == null)
    			continue;
    		double score = params.getScore(fvList[i]);
    		//System.out.println(score);
    		//for (int z = 0; z < depList[i].length; z++)
    		//	System.out.print(depList[i][z] + " ");
    		if (score > maxScore) {
    			maxScore = score;
    			maxID = i;
    		}
    	}
    	//System.out.println(maxID);
    	//System.exit(0);
    	
    	for (int i = 0; i < pred.dep.length; i++) {
    		pred.dep[i] = depList[maxID][i];
    		pred.labs[i] = labList[maxID][i];
    	}

    	Object[][] d = new Object[1][3];
		d[0][0] = fvList[maxID];
		d[0][1] = getParseString(pred);
		d[0][2] = new Double(maxScore);
		
    	return d;
    }

    public int getBottomUpOrder(DependencyInstance pred, int arg, int[] idx, int id) {
    	for (int i = 0; i < pred.child[arg].length; i++) {
    		idx[id] = pred.child[arg][i];
    		id--;
    	}
    	for (int i = 0; i < pred.child[arg].length; i++) {
    		id = getBottomUpOrder(pred, pred.child[arg][i], idx, id);
    	}
    	return id;
    }
    
    // pseudo max
    public void pseudoMaxTraining(DependencyInstance gold, DependencyInstance pred, 
 		   LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params, int iters) {
     	// pseudo max training
    	
    	// copy gold dep to pred first
    	copyDepLab(pred, gold.dep, gold.labs);
    	pipe.buildChild(pred);
    	
     	int len = pred.sentence.length;
 		int maxIter = 1;
 		
     	for (int iter = 0; iter < maxIter; iter++) {
     		//ArrayList<FeatureVector> targetFV = new ArrayList<FeatureVector>();
     		//ArrayList<Double> targetScore = new ArrayList<Double>();
     		//ArrayList<FeatureVector> currFV = new ArrayList<FeatureVector>();
     		//ArrayList<Double> currScore = new ArrayList<Double>();
 			//ArrayList<Double> error = new ArrayList<Double>();
     		
     		int[] idx = new int[len];
     		int id = getBottomUpOrder(pred, 0, idx, len - 1);
     		if (id != 0) {
     			System.out.println("build order bug");
     		}
     		/*
     		for (int i = 0; i < len; i++)
     			idx[i] = i;
     		for (int i = 0; i < len; i++) {
     			int i1 = r.nextInt(len - 1) + 1;
     			int i2 = r.nextInt(len - 1) + 1;
     			int tmp = idx[i1];
     			idx[i1] = idx[i2];
     			idx[i2] = tmp;
     		}*/
     		/*
 			ArrayList<Integer> dep = new ArrayList<Integer>();
 			ArrayList<Integer> mod = new ArrayList<Integer>();
 			ArrayList<Integer> lab = new ArrayList<Integer>();
 			ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
 			ArrayList<Double> scoreList = new ArrayList<Double>();

			int sample = findMaxLossDep(pred, gold, mod, dep, lab, fvList, scoreList, lfd, gfd, params);
			int m = mod.get(sample);
			int h = dep.get(sample);
			
			if (pred.dep[m] != h) {
				FeatureVector currFV = fvList.get(sample);
				double currScore = scoreList.get(sample);
					
				FeatureVector targetFV = gold.fv;
				double targetScore = params.getScore(targetFV);

				double upd = (double)updateTimes;
				params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, 1.0, upd);
				//params.updateParamsSubgradient(targetFV, currFV, targetScore, currScore, 1.0, upd);
			}
			updateTimes++;
			*/
     		
     		for (int z = 1; z < len; z++) {
     			
     			int i = idx[z];
     			
     			ArrayList<Integer> dep = new ArrayList<Integer>();
     			ArrayList<Integer> lab = new ArrayList<Integer>();
     			ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
     			ArrayList<Double> scoreList = new ArrayList<Double>();

 				int sample = findMaxLossDep(pred, gold, i, dep, lab, fvList, scoreList, lfd, gfd, params);
 				
 				if (sample != -1) {
 					
 	 				int newDep = dep.get(sample);
 	 				int newLab = lab.get(sample);
 	 				
 	 				if (pred.dep[i] != newDep || pred.labs[i] != newLab) {
 	 					// not gold, update
 	 					
 	 					
 	 	 				FeatureVector currFV = fvList.get(sample);
 	 	 				double currScore = scoreList.get(sample) - 1.0;
 	 	 				
 	 	 				FeatureVector targetFV = pipe.createFeatureVector(pred, lfd, gfd, i);
 	 					double targetScore = params.getScore(targetFV);
 	 	 				
 	 					//int comp = compare(gold.dep[i], gold.labs[i], newDep, newLab, gold.dep[i], gold.labs[i]);
 	 					
 	 					
 	 	 				//currFV.add(fvList.get(sample));
 	 	 				//currScore.add(scoreList.get(sample));
 	 	 				
 	 	 				//FeatureVector fv = pipe.createFeatureVector(pred, lfd, gfd, i);
 	 	 				//targetFV.add(fv);
 	 	 				//targetScore.add(params.getScore(fv));
 	 	 				
 	 					//error.add(1.0);
						
 	 					// update toward new tree
 						double upd = (double)updateTimes;
 	 							
 	    				// mira
 						params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, 1.0, upd);
 				     	updateTimes++;
 						
 						int oldH = pred.dep[i];
 						pred.dep[i] = newDep;
 						pipe.updateChildList(pred, newDep, oldH, i);
 						pred.labs[i] = newLab;
 						
 	 				}
 	 				updateTimes++;
 				}
 				
 				/*
 				FeatureVector fv = pipe.createFeatureVector(pred, lfd, gfd, i);
 				double s = params.getScore(fv);
 				for (int z = 0; z < dep.size(); i++) {
 					currFV.add(fvList.get(z));
 					currScore.add(scoreList.get(z));

 					targetFV.add(fv);
 					targetScore.add(s);

 					error.add(1.0);
 				}
 				*/
 				
 				/*
				FeatureVector fv = pipe.createLocalFeatureVector(pred, lfd, FeatureVector.end, i);
 				for (int j = i + 1; j < i + 2; j++) {
	 				dep.clear();
	 				fvList.clear();
	 				scoreList.clear();
	 				ArrayList<Integer> sibDep = new ArrayList<Integer>();
	 				
	 				sample = findMaxLossTwoDep(pred, gold, i, j, dep, sibDep, fvList, scoreList, lfd, gfd, params);
	 				
	 				if (sample != -1) {
	 					FeatureVector currFV = fvList.get(sample);
	 					double currScore = scoreList.get(sample) - 2.0;
	
	 					FeatureVector sibfv = pipe.createLocalFeatureVector(pred, lfd, fv, j);
	 					FeatureVector targetFV = sibfv;
	 					double targetScore = params.getScore(sibfv);
	
	 					double error = 2.0;

	 					double upd = (double)updateTimes;
 	    				// mira
 						params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, error, upd);
 				     	updateTimes++;
	 				}
 				}*/
 				
 				
 				/*
 	     		FeatureVector[] tfv = new FeatureVector[targetFV.size()];
 	     		double[] ts = new double[targetScore.size()];
 	     		FeatureVector[] cfv = new FeatureVector[currFV.size()];
 	     		double[] cs = new double[currScore.size()];
 	     		double[] e = new double[error.size()];
 	     		
 	     		for (int z = 0; z < tfv.length; z++) {
 	     			tfv[z] = targetFV.get(z);
 	     			ts[z] = targetScore.get(z);
 	     			cfv[z] = currFV.get(z);
 	     			cs[z] = currScore.get(z);
 	     			e[z] = error.get(z);
 	     		}

 	     		if (tfv.length > 0) {
 		     		double upd = (double)updateTimes;
 		     		//System.out.println("begin");
 					params.updateParamsMIRA(tfv, cfv, ts, cs, e, upd);
 		     		//System.out.println("end");
 			     	updateTimes++;
 	     		}
 	     		*/
 				
     		}
     		
			FeatureVector currFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
			double currScore = params.getScore(currFV);
				
			FeatureVector targetFV = gold.fv;
			double targetScore = params.getScore(targetFV);
			
			double error = numError(gold.dep, gold.labs, pred.dep, pred.labs);
     		
			double upd = (double)updateTimes;
			params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, error, upd);
	     	updateTimes++;
			// mira
     		/*
     		FeatureVector[] tfv = new FeatureVector[targetFV.size()];
     		double[] ts = new double[targetScore.size()];
     		FeatureVector[] cfv = new FeatureVector[currFV.size()];
     		double[] cs = new double[currScore.size()];
     		double[] e = new double[error.size()];
     		
     		for (int i = 0; i < tfv.length; i++) {
     			tfv[i] = targetFV.get(i);
     			ts[i] = targetScore.get(i);
     			cfv[i] = currFV.get(i);
     			cs[i] = currScore.get(i);
     			e[i] = error.get(i);
     		}

     		if (tfv.length > 0) {
	     		double upd = (double)updateTimes;
	     		//System.out.println("begin");
				params.updateParamsMIRA(tfv, cfv, ts, cs, e, upd);
	     		//System.out.println("end");
		     	updateTimes++;
     		}
     		*/
     	}
     	//updateTimes++;
     	
     	// check
     	//for (int i = 1; i < pred.dep.length; i++)
     	//	if (pred.dep[i] != gold.dep[i] || pred.labs[i] != gold.labs[i]) {
     	//		System.out.println("pseudo max bug 1");
     	//		System.exit(0);
     	//	}
     	
     	//updateTimes++;
    }
    
    public int findMaxLossDep(DependencyInstance pred, DependencyInstance gold, int i, 
    		ArrayList<Integer> dep, ArrayList<Integer> lab, 
    		ArrayList<FeatureVector> fvList, ArrayList<Double> scoreList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params) {
    	
    	int len = pred.dep.length;
    	
		int oldDep = pred.dep[i];
		int oldLab = pred.labs[i];

		// build score
		for (int candH = 0; candH < len; candH++) {
			//if (candH == oldDep)
			//	continue;
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			int oldH = pred.dep[i];
			pred.dep[i] = candH;
			pipe.updateChildList(pred, candH, oldH, i);
			for (int candT = 1; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				dep.add(candH);
				lab.add(candT);
				pred.labs[i] = candT;
				
				//double loss = (double)compare(gold.dep[i], gold.labs[i], pred.dep[i], pred.labs[i], gold.dep[i], gold.labs[i]);

				FeatureVector fv = pipe.createFeatureVector(pred, lfd, gfd, i);
				double s = params.getScore(fv) + (candH == gold.dep[i] ? 0.0 : 1.0);
				fvList.add(fv);
				scoreList.add(s);
			}
		}
		
		int oldH = pred.dep[i];
		pred.dep[i] = oldDep;
		pred.labs[i] = oldLab;
		pipe.updateChildList(pred, oldDep, oldH, i);
		//pipe.buildChild(pred);
		if (scoreList.size() == 0) {
			//System.out.println("aaa");
			return -1;
		}
		
		// sample
		int sample = 0;
		double maxScore = scoreList.get(0);
		for (int z = 1; z < scoreList.size(); z++) {
			if (scoreList.get(z) > maxScore) {
				maxScore = scoreList.get(z);
				sample = z;
			}
		}
    	
		return sample;
    }
    
    public int findMaxLossDep(DependencyInstance pred, DependencyInstance gold, 
    		ArrayList<Integer> mod, ArrayList<Integer> dep, ArrayList<Integer> lab, 
    		ArrayList<FeatureVector> fvList, ArrayList<Double> scoreList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params) {
    	
    	int len = pred.dep.length;
    	
    	FeatureVector allFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
    	double allScore = params.getScore(allFV);
    	
    	double ms = -1e40;
    	
    	for (int i = 1; i < len; i++) {
			int oldDep = pred.dep[i];
			int oldLab = pred.labs[i];
	
	    	FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, i);
	    	double oldScore = params.getScore(oldFV);

	    	// build score
			for (int candH = 0; candH < len; candH++) {
				//if (candH == oldDep)
				//	continue;
				if (!lfd.cand[candH][i])
					continue;
				if (isDescendant(candH, i, pred.dep))		// loop
					continue;
				if (proj && !pipe.isProj(candH, i, pred.dep))
					continue;
				int oldH = pred.dep[i];
				pred.dep[i] = candH;
				pipe.updateChildList(pred, candH, oldH, i);
				for (int candT = 1; candT < lfd.type; candT++) {
					if (!lfd.cand_dep[candH][i][candT])
						continue;
					//mod.add(i);
					//dep.add(candH);
					//lab.add(candT);
					pred.labs[i] = candT;
					
					//double loss = (double)compare(gold.dep[i], gold.labs[i], pred.dep[i], pred.labs[i], gold.dep[i], gold.labs[i]);
	
					FeatureVector fv = pipe.createFeatureVector(pred, lfd, gfd, i);
					double s = allScore - oldScore + params.getScore(fv) + (candH == gold.dep[i] ? 0.0 : 1.0);
					
					//FeatureVector tmpfv = pipe.createFeatureVector(pred, lfd, gfd, -1);
					//double tmps = params.getScore(tmpfv) + (candH == gold.dep[i] ? 0.0 : 1.0);
					//if (Math.abs(s - tmps) > 1e-4) {
					//	System.out.println("score mismatch");
					//	System.exit(0);
					//}
					//fvList.add(fv);
					//scoreList.add(s);
					if (s > ms) {
						ms = s;
						mod.clear(); mod.add(i);
						dep.clear(); dep.add(candH);
						lab.clear(); lab.add(candT);
						fvList.clear(); fvList.add(pipe.createFeatureVector(pred, lfd, gfd, -1));
						scoreList.clear(); scoreList.add(params.getScore(fvList.get(0)));
					}
				}
			}
			
			int oldH = pred.dep[i];
			pred.dep[i] = oldDep;
			pred.labs[i] = oldLab;
			pipe.updateChildList(pred, oldDep, oldH, i);
    	}
    	
		//pipe.buildChild(pred);
		if (scoreList.size() == 0) {
			//System.out.println("aaa");
			return -1;
		}
		
		// sample
		int sample = 0;
		double maxScore = scoreList.get(0);
		for (int z = 1; z < scoreList.size(); z++) {
			if (scoreList.get(z) > maxScore) {
				maxScore = scoreList.get(z);
				sample = z;
			}
		}
    	
		return sample;
    }
    
    public int findMaxLossTwoDep(DependencyInstance pred, DependencyInstance gold, int i, int j,
    		ArrayList<Integer> dep, ArrayList<Integer> sibDep, ArrayList<FeatureVector> fvList, ArrayList<Double> scoreList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params) {
    	
    	int len = pred.dep.length;
    	
    	if (i == len - 1)
    		return -1;
    	
		int oldDep = pred.dep[i];
		int oldSibDep = pred.dep[j];

		// build score
		for (int candH = 0; candH < len; candH++) {
			if (candH == oldDep)
				continue;
			if (!lfd.cand[candH][i])
				continue;
			if (isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			int oldH = pred.dep[i];
			pred.dep[i] = candH;
			pipe.updateChildList(pred, candH, oldH, i);
			
			FeatureVector fv = pipe.createLocalFeatureVector(pred, lfd, FeatureVector.end, i);
			
			for (int candSibH = 0; candSibH < len; candSibH++) {
				if (candSibH == oldSibDep)
					continue;
				if (!lfd.cand[candSibH][j])
					continue;
				if (isDescendant(candSibH, j, pred.dep))		// loop
					continue;
				if (proj && !pipe.isProj(candSibH, j, pred.dep))
					continue;

				dep.add(candH);
				sibDep.add(candSibH);
				int oldSibH = pred.dep[j];
				pred.dep[j] = candSibH;
				pipe.updateChildList(pred, candSibH, oldSibH, j);
				
				//double loss = (double)compare(gold.dep[i + 1], gold.labs[i + 1], pred.dep[i + 1], pred.labs[i + 1], gold.dep[i + 1], gold.labs[i + 1]);

				FeatureVector sibfv = pipe.createLocalFeatureVector(pred, lfd, fv, j);
				double s = params.getScore(sibfv) + 2.0;
				fvList.add(sibfv);
				scoreList.add(s);
			}
			
			int oldSibH = pred.dep[j];
			pred.dep[j] = oldSibDep;
			pipe.updateChildList(pred, oldSibDep, oldSibH, j);

		}
		
		int oldH = pred.dep[i];
		pred.dep[i] = oldDep;
		pipe.updateChildList(pred, oldDep, oldH, i);

		if (scoreList.size() == 0) {
			//System.out.println("aaa");
			return -1;
		}
		
		// sample
		int sample = 0;
		double maxScore = scoreList.get(0);
		for (int z = 1; z < scoreList.size(); z++) {
			if (scoreList.get(z) > maxScore) {
				maxScore = scoreList.get(z);
				sample = z;
			}
		}
    	
		return sample;
    }
    
    // pseudo max
    public void pseudoMaxTrainingWithSample(DependencyInstance gold, DependencyInstance pred, 
 		   LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params, int iters) {
     	// pseudo max training
    	
     	int len = pred.sentence.length;
 		int maxIter = 1;
 		
     	for (int iter = 0; iter < maxIter; iter++) {
     		for (int i = 1; i < len; i++) {
     			ArrayList<Integer> dep = new ArrayList<Integer>();
     			ArrayList<Integer> lab = new ArrayList<Integer>();
     			ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>();
     			ArrayList<Double> scoreList = new ArrayList<Double>();
 				
				int sample = gibbsSampling(pred, gold, i, dep, lab, 
						fvList, scoreList, lfd, gfd, params, 1.0, true, r);

				if (sample == -1)
 					continue;
 				
 				int newDep = dep.get(sample);
 				int newLab = lab.get(sample);
 				
 				if (pred.dep[i] != newDep || pred.labs[i] != newLab) {
 					// not gold, update
 					
 					int oldH = pred.dep[i];
					pred.dep[i] = newDep; pred.labs[i] = newLab;
					pipe.updateChildList(pred, newDep, oldH, i);

					FeatureVector currFV = fvList.get(sample);
 	 				double currScore = scoreList.get(sample);
 	 				
 	 				FeatureVector targetFV = pipe.createFeatureVector(gold, lfd, gfd, i);
 	 				double targetScore = params.getScore(targetFV);
 	 				
 					double comp = numError(gold.dep, gold.labs, pred.dep, pred.labs);
 						
					// update toward new tree
					double upd = (double)updateTimes;
 							
    				// mira
					params.updateParamsMIRA(targetFV, currFV, targetScore, currScore, comp, upd);
			     	updateTimes++;
 				}
 				
     		}
     	}
     	
     	//updateTimes++;
    }
    
    // dynamic programming
 
    protected int[][] getTypes(LocalFeatureData lfd, int len) {
    	int[][] static_types = new int[len][len];
    	for(int i = 0; i < len; i++) {
    		for(int j = 0; j < len; j++) {
    			if(i == j) {
    				static_types[i][j] = 0; 
    				continue; 
    			}
				if (!lfd.cand[i][j])
					continue;
    			int wh = -1;
    			//int wh = 0;
    			double best = Double.NEGATIVE_INFINITY;
    			for(int t = 0; t < pipe.types.length; t++) {
    				if (!lfd.cand_dep[i][j][t])
    					continue;
    				double score = 0.0;
    				if(i < j) {
    					//if (lfd.cand_lab[i][t][0][1] && lfd.cand_lab[j][t][0][0])
    					score = lfd.probs_lab[i][t][0][1] + lfd.probs_lab[j][t][0][0] + lfd.probs_dep[i][j][t];
    				}
    				else {
       					//if (lfd.cand_lab[i][t][1][1] && lfd.cand_lab[j][t][1][0])
       					score = lfd.probs_lab[i][t][1][1] + lfd.probs_lab[j][t][1][0] + lfd.probs_dep[i][j][t];
    				}

    				if(score > best) { 
    					wh = t; 
    					best = score; 
    				}
    			}
    			static_types[i][j] = wh;
    		}
    	}
    	return static_types;
    }

    // same as decode, except return K best
    public Object[][] decodeSecondOrderProjective(DependencyInstance inst,
    		LocalFeatureData lfd, int K) {

    	int[] toks = inst.sentence;
    	FeatureVector[][] fvs = lfd.fvs;
    	double[][] probs = lfd.probs;
    	FeatureVector[][][][] nt_fvs = lfd.fvs_lab;
    	double[][][][] nt_probs = lfd.probs_lab;
    	FeatureVector[][][] fvs_trips = lfd.fvs_trips;
    	double[][][] probs_trips = lfd.probs_trips;
       	FeatureVector[][][] fvs_sibs = lfd.fvs_sibs;
    	double[][][] probs_sibs = lfd.probs_sibs;

    	int[][] static_types = null;
    	if(pipe.labeled) {
    		static_types = getTypes(lfd, toks.length);
    	}

    	KBestParseForest2O pf = new KBestParseForest2O(0, toks.length-1, inst, K);

    	for(int s = 0; s < toks.length; s++) {
    		pf.add(s,-1,0,0.0,new FeatureVector(-1,-1.0,null));
    		pf.add(s,-1,1,0.0,new FeatureVector(-1,-1.0,null));
    	}

    	for(int j = 1; j < toks.length; j++) {
    		for(int s = 0; s < toks.length && s+j < toks.length; s++) {
    			int t = s+j;

    			FeatureVector prodFV_st = fvs[s][t];
    			FeatureVector prodFV_ts = fvs[t][s];				
    			double prodProb_st = probs[s][t];
    			double prodProb_ts = probs[t][s];
    			
    			int type1 = pipe.labeled ? static_types[s][t] : 0;
    			int type2 = pipe.labeled ? static_types[t][s] : 0;
    			type1 = type1 == -1 ? 0 : type1;
    			type2 = type2 == -1 ? 0 : type2;

    			FeatureVector nt_fv_st = lfd.fvs_dep[s][t][type1];
    			FeatureVector nt_fv_ts = lfd.fvs_dep[t][s][type2];
    			double nt_prob_st = lfd.probs_dep[s][t][type1];
    			double nt_prob_ts = lfd.probs_dep[t][s][type2];

    			FeatureVector nt_fv_s_01 = nt_fvs[s][type1][0][1];
    			FeatureVector nt_fv_s_10 = nt_fvs[s][type2][1][0];
    			FeatureVector nt_fv_t_00 = nt_fvs[t][type1][0][0];
    			FeatureVector nt_fv_t_11 = nt_fvs[t][type2][1][1];
    			double nt_prob_s_01 = nt_probs[s][type1][0][1];
    			double nt_prob_s_10 = nt_probs[s][type2][1][0];
    			double nt_prob_t_00 = nt_probs[t][type1][0][0];
    			double nt_prob_t_11 = nt_probs[t][type2][1][1];

    			if(true) {
    				// case when r == s
    				ParseForestItem[] b1 = pf.getItems(s,s,0,0);
    				ParseForestItem[] c1 = pf.getItems(s+1,t,1,0);
    				if(!(b1 == null || c1 == null)
    						&& lfd.cand_dep[s][t][type1]
    						&& lfd.cand_trips[s][s][t]
    						&& lfd.cand_sibs[s][t][0]) {

    					FeatureVector prodFV_sst = pf.cat(fvs_trips[s][s][t],fvs_sibs[s][t][0]);
    					double prodProb_sst = probs_trips[s][s][t]+probs_sibs[s][t][0];

    					int[][] pairs = pf.getKBestPairs(b1,c1);

    					for(int k = 0; k < pairs.length; k++) {

    						if(pairs[k][0] == -1 || pairs[k][1] == -1)
    							break;

    						int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    						double bc = b1[comp1].prob+c1[comp2].prob;

    						// create sibling pair
    						// create parent pair: s->t and s->(start,t)
    						bc += prodProb_st + prodProb_sst;

    						FeatureVector fv_fin = pf.cat(prodFV_st,prodFV_sst);
    						if(pipe.labeled) {
    							bc += nt_prob_s_01 + nt_prob_t_00 + nt_prob_st;
    							fv_fin = FeatureVector.cat(nt_fv_s_01,FeatureVector.cat(nt_fv_t_00,fv_fin));
    							fv_fin = fv_fin.concat(nt_fv_st);
    						}

    						pf.add(s,s,t,type1,0,1,bc,fv_fin,b1[comp1],c1[comp2]);

    					}
    				}

    				// case when r == t
    				b1 = pf.getItems(s,t-1,0,0);
    				c1 = pf.getItems(t,t,1,0);
    				if(!(b1 == null || c1 == null)
    						&& lfd.cand_dep[t][s][type2]
    						&& lfd.cand_trips[t][t][s]
    						&& lfd.cand_sibs[t][s][0]) {

    					FeatureVector prodFV_stt = pf.cat(fvs_trips[t][t][s],fvs_sibs[t][s][0]);
    					double prodProb_stt = probs_trips[t][t][s]+probs_sibs[t][s][0];

    					int[][] pairs = pf.getKBestPairs(b1,c1);

    					for(int k = 0; k < pairs.length; k++) {

    						if(pairs[k][0] == -1 || pairs[k][1] == -1)
    							break;

    						int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    						double bc = b1[comp1].prob+c1[comp2].prob;

    						// create sibling pair
    						// create parent pair: s->t and s->(start,t)
    						bc += prodProb_ts + prodProb_stt;

    						FeatureVector fv_fin = pf.cat(prodFV_ts,prodFV_stt);
    						if(pipe.labeled) {
    							bc += nt_prob_t_11 + nt_prob_s_10 + nt_prob_ts;
    							fv_fin = FeatureVector.cat(nt_fv_t_11,FeatureVector.cat(nt_fv_s_10,fv_fin));
    							fv_fin = fv_fin.concat(nt_fv_ts);
    						}

    						pf.add(s,t,t,type2,1,1,bc,fv_fin,b1[comp1],c1[comp2]);
    					}
    				}
    			}

    			for(int r = s; r < t; r++) {

    				// First case - create sibling
    				ParseForestItem[] b1 = pf.getItems(s,r,0,0);
    				ParseForestItem[] c1 = pf.getItems(r+1,t,1,0);

    				if(!(b1 == null || c1 == null)) {

    					int[][] pairs = pf.getKBestPairs(b1,c1);

    					for(int k = 0; k < pairs.length; k++) {

    						if(pairs[k][0] == -1 || pairs[k][1] == -1)
    							break;

    						int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    						double bc = b1[comp1].prob+c1[comp2].prob;

    						pf.add(s,r,t,-1,0,2,bc,new FeatureVector(-1,-1.0,null),b1[comp1],c1[comp2]);
    						pf.add(s,r,t,-1,1,2,bc,new FeatureVector(-1,-1.0,null),b1[comp1],c1[comp2]);
    					}
    				}
    			}

    			for(int r = s+1; r < t; r++) {
    				// s -> (r,t)
    				ParseForestItem[] b1 = pf.getItems(s,r,0,1);
    				ParseForestItem[] c1 = pf.getItems(r,t,0,2);

    				if(!(b1 == null || c1 == null)
    						&& lfd.cand_dep[s][t][type1]
    						&& lfd.cand_trips[s][r][t]
    						&& lfd.cand_sibs[r][t][1]) {

    					int[][] pairs = pf.getKBestPairs(b1,c1);

    					for(int k = 0; k < pairs.length; k++) {

    						if(pairs[k][0] == -1 || pairs[k][1] == -1)
    							break;

    						int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    						double bc = b1[comp1].prob+c1[comp2].prob;

    						bc += prodProb_st + probs_trips[s][r][t] + probs_sibs[r][t][1];
    						FeatureVector fv_fin = pf.cat(prodFV_st,pf.cat(fvs_trips[s][r][t],fvs_sibs[r][t][1]));

    						if(pipe.labeled) {
    							bc += nt_prob_s_01 + nt_prob_t_00 + nt_prob_st;
    							fv_fin = FeatureVector.cat(nt_fv_s_01,FeatureVector.cat(nt_fv_t_00,fv_fin));
    							fv_fin = fv_fin.concat(nt_fv_st);
    						}

    						pf.add(s,r,t,type1,0,1,bc,fv_fin,b1[comp1],c1[comp2]);
    					}
    				}

    				// t -> (r,s)
    				b1 = pf.getItems(s,r,1,2);
    				c1 = pf.getItems(r,t,1,1);

    				if(!(b1 == null || c1 == null)
    						&& lfd.cand_dep[t][s][type2]
    						&& lfd.cand_trips[t][r][s]
    						&& lfd.cand_sibs[r][s][1]) {

    					int[][] pairs = pf.getKBestPairs(b1,c1);

    					for(int k = 0; k < pairs.length; k++) {

    						if(pairs[k][0] == -1 || pairs[k][1] == -1)
    							break;

    						int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    						double bc = b1[comp1].prob+c1[comp2].prob;

    						bc += prodProb_ts + probs_trips[t][r][s] + probs_sibs[r][s][1];

    						FeatureVector fv_fin = pf.cat(prodFV_ts,pf.cat(fvs_trips[t][r][s],fvs_sibs[r][s][1]));
    						if(pipe.labeled) {
    							bc += nt_prob_t_11 + nt_prob_s_10 + nt_prob_ts;
    							fv_fin = FeatureVector.cat(nt_fv_t_11,FeatureVector.cat(nt_fv_s_10,fv_fin));
    							fv_fin = fv_fin.concat(nt_fv_ts);
    						}

    						pf.add(s,r,t,type2,1,1,bc,fv_fin,b1[comp1],c1[comp2]);
    					}
    				}

    			}


    			// Finish off pieces incom + comp -> comp
    			for(int r = s; r <= t; r++) {

    				if(r != s) {
    					ParseForestItem[] b1 = pf.getItems(s,r,0,1);
    					ParseForestItem[] c1 = pf.getItems(r,t,0,0);

    					if(!(b1 == null || c1 == null)) {
    						//continue;

    						int[][] pairs = pf.getKBestPairs(b1,c1);
    						for(int k = 0; k < pairs.length; k++) {

    							if(pairs[k][0] == -1 || pairs[k][1] == -1)
    								break;

    							int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    							double bc = b1[comp1].prob+c1[comp2].prob;

    							if(!pf.add(s,r,t,-1,0,0,bc,new FeatureVector(-1,-1.0,null),b1[comp1],c1[comp2]))
    								break;
    						}
    					}
    				}

    				if(r != t) {
    					ParseForestItem[] b1 = pf.getItems(s,r,1,0);
    					ParseForestItem[] c1 = pf.getItems(r,t,1,1);

    					if(!(b1 == null || c1 == null)) {
    						//continue;

    						int[][] pairs = pf.getKBestPairs(b1,c1);
    						for(int k = 0; k < pairs.length; k++) {

    							if(pairs[k][0] == -1 || pairs[k][1] == -1)
    								break;

    							int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    							double bc = b1[comp1].prob+c1[comp2].prob;

    							if(!pf.add(s,r,t,-1,1,0,bc,new FeatureVector(-1,-1.0,null),b1[comp1],c1[comp2]))
    								break;
    						}
    					}
    				}

    			}

    		}
    	}

    	return pf.getBestParses();
    }

    // 1st order nonproj
    public void fillTable(DependencyInstance inst, LocalFeatureData lfd, Parameters params) {
    	int len = inst.dep.length;
    	for (int h = 0; h < len; h++)
    		for (int m = 1; m < len; m++) {
    			if (m == h)
    				continue;
    			//if (lfd.fvs[h][m] == null) {
    				lfd.fvs[h][m] = pipe.createFeatureVector(inst, h < m ? h : m, h > m ? h : m, h < m ? true : false, FeatureVector.end);
    				lfd.probs[h][m] = params.getScore(lfd.fvs[h][m]);
    			//}
    		}
    }
 
    public Object[][] decodeProjective(DependencyInstance inst,
    		LocalFeatureData lfd, Parameters params, int K) {
    	fillTable(inst, lfd, params);

    	int[] toks = inst.sentence;

    	FeatureVector[][] fvs = lfd.fvs;
    	double[][] probs = lfd.probs;
    	FeatureVector[][][][] nt_fvs = lfd.fvs_lab;
    	double[][][][] nt_probs = lfd.probs_lab;

    	int[][] static_types = null;
    	if(pipe.labeled) {
    		static_types = getTypes(lfd, toks.length);
    	}

    	KBestParseForest2O pf = new KBestParseForest2O(0, toks.length-1, inst, K);

    	for(int s = 0; s < toks.length; s++) {
    		pf.add(s,-1,0,0.0,new FeatureVector(-1,-1.0,null));
    		pf.add(s,-1,1,0.0,new FeatureVector(-1,-1.0,null));
    	}

    	for(int j = 1; j < toks.length; j++) {
    		for(int s = 0; s < toks.length && s+j < toks.length; s++) {
    			int t = s+j;

    			FeatureVector prodFV_st = fvs[s][t];
    			FeatureVector prodFV_ts = fvs[t][s];				
    			double prodProb_st = probs[s][t];
    			double prodProb_ts = probs[t][s];

    			int type1 = pipe.labeled ? static_types[s][t] : DependencyPipe2O.NOTYPE;
    			int type2 = pipe.labeled ? static_types[t][s] : DependencyPipe2O.NOTYPE;

    			FeatureVector nt_fv_s_01 = nt_fvs[s][type1][0][1];
    			FeatureVector nt_fv_s_10 = nt_fvs[s][type2][1][0];
    			FeatureVector nt_fv_t_00 = nt_fvs[t][type1][0][0];
    			FeatureVector nt_fv_t_11 = nt_fvs[t][type2][1][1];
    			double nt_prob_s_01 = nt_probs[s][type1][0][1];
    			double nt_prob_s_10 = nt_probs[s][type2][1][0];
    			double nt_prob_t_00 = nt_probs[t][type1][0][0];
    			double nt_prob_t_11 = nt_probs[t][type2][1][1];

    			for(int r = s; r <= t; r++) {

    				/** first is direction, second is complete*/
    				/** _s means s is the parent*/
    				if(r != t) {
    					ParseForestItem[] b1 = pf.getItems(s,r,0,0);
    					ParseForestItem[] c1 = pf.getItems(r+1,t,1,0);

    					if(b1 != null && c1 != null) {
    						int[][] pairs = pf.getKBestPairs(b1,c1);
    						for(int k = 0; k < pairs.length; k++) {

    							if(pairs[k][0] == -1 || pairs[k][1] == -1)
    								break;

    							int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    							double bc = b1[comp1].prob+c1[comp2].prob;

    							double prob_fin = bc+prodProb_st;
    							FeatureVector fv_fin = prodFV_st;
    							if(pipe.labeled) {
    								fv_fin = FeatureVector.cat(nt_fv_s_01,FeatureVector.cat(nt_fv_t_00,fv_fin));
    								prob_fin += nt_prob_s_01+nt_prob_t_00;
    							}
    							pf.add(s,r,t,type1,0,1,prob_fin,fv_fin,b1[comp1],c1[comp2]);

    							prob_fin = bc+prodProb_ts;
    							fv_fin = prodFV_ts;
    							if(pipe.labeled) {
    								fv_fin = FeatureVector.cat(nt_fv_t_11,FeatureVector.cat(nt_fv_s_10,fv_fin));
    								prob_fin += nt_prob_t_11+nt_prob_s_10;
    							}
    							pf.add(s,r,t,type2,1,1,prob_fin,fv_fin,b1[comp1],c1[comp2]);

    						}
    					}						
    				}					
    			}


    			for(int r = s; r <= t; r++) {

    				if(r != s) {
    					ParseForestItem[] b1 = pf.getItems(s,r,0,1);
    					ParseForestItem[] c1 = pf.getItems(r,t,0,0);
    					if(b1 != null && c1 != null) {
    						int[][] pairs = pf.getKBestPairs(b1,c1);
    						for(int k = 0; k < pairs.length; k++) {

    							if(pairs[k][0] == -1 || pairs[k][1] == -1)
    								break;

    							int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    							double bc = b1[comp1].prob+c1[comp2].prob;

    							if(!pf.add(s,r,t,-1,0,0,bc,
    									new FeatureVector(-1,-1.0,null),
    									b1[comp1],c1[comp2]))
    								break;
    						}
    					}
    				}

    				if(r != t) {
    					ParseForestItem[] b1 = pf.getItems(s,r,1,0);
    					ParseForestItem[] c1 = pf.getItems(r,t,1,1);
    					if(b1 != null && c1 != null) {
    						int[][] pairs = pf.getKBestPairs(b1,c1);
    						for(int k = 0; k < pairs.length; k++) {

    							if(pairs[k][0] == -1 || pairs[k][1] == -1)
    								break;

    							int comp1 = pairs[k][0]; int comp2 = pairs[k][1];

    							double bc = b1[comp1].prob+c1[comp2].prob;

    							if(!pf.add(s,r,t,-1,1,0,bc,
    									new FeatureVector(-1,-1.0,null),b1[comp1],c1[comp2]))
    								break;
    						}
    					}
    				}
    			}
    		}

    	}

    	return pf.getBestParses();
    }

    public Object[][] decodeNonProjective(DependencyInstance inst, 
    		DependencyInstance gold, LocalFeatureData lfd, Parameters params, int K) {

    	fillTable(inst, lfd, params);
      int[] pos = inst.pos;

      int numWords = inst.dep.length;

      int[][] oldI = new int[numWords][numWords];
      int[][] oldO = new int[numWords][numWords];
      double[][] scoreMatrix = new double[numWords][numWords];
      double[][] orig_scoreMatrix = new double[numWords][numWords];
      boolean[] curr_nodes = new boolean[numWords];
      TIntIntHashMap[] reps = new TIntIntHashMap[numWords];

      int[][] static_types = null;
      if (pipe.labeled) {
        static_types = getTypes(lfd, pos.length);
      }
      
      double[][] probs = lfd.probs;

      for (int i = 0; i < numWords; i++) {
        curr_nodes[i] = true;
        reps[i] = new TIntIntHashMap();
        reps[i].put(i, 0);
        for (int j = 0; j < numWords; j++) {
          // score of edge (i,j) i --> j
          scoreMatrix[i][j] = probs[i][j] + (gold != null && i != gold.dep[j] ? 1.0 : 0.0)
                  + (pipe.labeled ? lfd.probs_lab[i][static_types[i][j]][i < j ? 0 : 1][1]
                          + lfd.probs_lab[j][static_types[i][j]][i < j ? 0 : 1][0] : 0.0);
          orig_scoreMatrix[i][j] = probs[i][j] + (gold != null && i != gold.dep[j] ? 1.0 : 0.0)
                  + (pipe.labeled ? lfd.probs_lab[i][static_types[i][j]][i < j ? 0 : 1][1]
                          + lfd.probs_lab[j][static_types[i][j]][i < j ? 0 : 1][0] : 0.0);
          oldI[i][j] = i;
          oldO[i][j] = j;
          if (i == j || j == 0)
            continue; // no self loops of i --> 0
        }
      }

      TIntIntHashMap final_edges = chuLiuEdmonds(scoreMatrix, curr_nodes, oldI, oldO, false,
              new TIntIntHashMap(), reps);
      int[] par = new int[numWords];
      int[] ns = final_edges.keys();
      for (int i = 0; i < ns.length; i++) {
        int ch = ns[i];
        int pr = final_edges.get(ns[i]);
        par[ch] = pr;
      }

      int[] n_par = getKChanges(par, orig_scoreMatrix, Math.min(K, par.length));
      int new_k = 1;
      for (int i = 0; i < n_par.length; i++)
        if (n_par[i] > -1)
          new_k++;

      // Create Feature Vectors;
      int[][] fin_par = new int[new_k][numWords];
      FeatureVector[][] fin_fv = new FeatureVector[new_k][numWords];
      fin_par[0] = par;
      int c = 1;
      for (int i = 0; i < n_par.length; i++) {
        if (n_par[i] > -1) {
          int[] t_par = new int[par.length];
          for (int j = 0; j < t_par.length; j++)
            t_par[j] = par[j];
          t_par[i] = n_par[i];
          fin_par[c] = t_par;
          c++;
        }
      }
      for (int k = 0; k < fin_par.length; k++) {
        for (int i = 0; i < fin_par[k].length; i++) {
          int ch = i;
          int pr = fin_par[k][i];
          if (pr != -1) {
            fin_fv[k][ch] = lfd.fvs[pr][ch];
            if (pipe.labeled) {
              fin_fv[k][ch] = fin_fv[k][ch].concat(lfd.fvs_lab[ch][static_types[pr][ch]][ch < pr ? 1 : 0][0]);
              fin_fv[k][ch] = fin_fv[k][ch].concat(lfd.fvs_lab[pr][static_types[pr][ch]][ch < pr ? 1 : 0][1]);
            }
          } else
            fin_fv[k][ch] = FeatureVector.end;
        }
      }

      FeatureVector[] fin = new FeatureVector[new_k];
      String[] result = new String[new_k];
      for (int k = 0; k < fin.length; k++) {
        fin[k] = FeatureVector.end;
        for (int i = 1; i < fin_fv[k].length; i++)
          fin[k] = fin[k].concat(fin_fv[k][i]);
        result[k] = "";
        for (int i = 1; i < par.length; i++)
          result[k] += fin_par[k][i] + "|" + i
                  + (pipe.labeled ? ":" + static_types[fin_par[k][i]][i] : ":0") + " ";
      }

      // create d.
      Object[][] d = new Object[new_k][3];

      for (int k = 0; k < new_k; k++) {
        d[k][0] = fin[k];
        d[k][1] = result[k].trim();
        d[k][2] = params.getScore(fin[k]);
      }

      return d;
    }

    private int[] getKChanges(int[] par, double[][] scoreMatrix, int K) {
      int[] result = new int[par.length];
      int[] n_par = new int[par.length];
      double[] n_score = new double[par.length];
      for (int i = 0; i < par.length; i++) {
        result[i] = -1;
        n_par[i] = -1;
        n_score[i] = Double.NEGATIVE_INFINITY;
      }

      boolean[][] isChild = calcChilds(par);

      for (int i = 1; i < n_par.length; i++) {
        double max = Double.NEGATIVE_INFINITY;
        int wh = -1;
        for (int j = 0; j < n_par.length; j++) {
          if (i == j || par[i] == j || isChild[i][j])
            continue;
          if (scoreMatrix[j][i] > max) {
            max = scoreMatrix[j][i];
            wh = j;
          }
        }
        n_par[i] = wh;
        n_score[i] = max;
      }

      for (int k = 0; k < K; k++) {
        double max = Double.NEGATIVE_INFINITY;
        int wh = -1;
        int whI = -1;
        for (int i = 0; i < n_par.length; i++) {
          if (n_par[i] == -1)
            continue;
          double score = scoreMatrix[n_par[i]][i];
          if (score > max) {
            max = score;
            whI = i;
            wh = n_par[i];
          }
        }

        if (max == Double.NEGATIVE_INFINITY)
          break;
        result[whI] = wh;
        n_par[whI] = -1;
      }

      return result;
    }

    private boolean[][] calcChilds(int[] par) {
      boolean[][] isChild = new boolean[par.length][par.length];
      for (int i = 1; i < par.length; i++) {
        int l = par[i];
        while (l != -1) {
          isChild[l][i] = true;
          l = par[l];
        }
      }
      return isChild;
    }

    private static TIntIntHashMap chuLiuEdmonds(double[][] scoreMatrix, boolean[] curr_nodes,
            int[][] oldI, int[][] oldO, boolean print, TIntIntHashMap final_edges,
            TIntIntHashMap[] reps) {

      // need to construct for each node list of nodes they represent (here only!)

      int[] par = new int[curr_nodes.length];
      int numWords = curr_nodes.length;

      // create best graph
      par[0] = -1;
      for (int i = 1; i < par.length; i++) {
        // only interested in current nodes
        if (!curr_nodes[i])
          continue;
        double maxScore = scoreMatrix[0][i];
        par[i] = 0;
        for (int j = 0; j < par.length; j++) {
          if (j == i)
            continue;
          if (!curr_nodes[j])
            continue;
          double newScore = scoreMatrix[j][i];
          if (newScore > maxScore) {
            maxScore = newScore;
            par[i] = j;
          }
        }
      }

      if (print) {
        System.out.println("After init");
        for (int i = 0; i < par.length; i++) {
          if (curr_nodes[i])
            System.out.print(par[i] + "|" + i + " ");
        }
        System.out.println();
      }

      // Find a cycle
      ArrayList cycles = new ArrayList();
      boolean[] added = new boolean[numWords];
      for (int i = 0; i < numWords && cycles.size() == 0; i++) {
        // if I have already considered this or
        // This is not a valid node (i.e. has been contracted)
        if (added[i] || !curr_nodes[i])
          continue;
        added[i] = true;
        TIntIntHashMap cycle = new TIntIntHashMap();
        cycle.put(i, 0);
        int l = i;
        while (true) {
          if (par[l] == -1) {
            added[l] = true;
            break;
          }
          if (cycle.contains(par[l])) {
            cycle = new TIntIntHashMap();
            int lorg = par[l];
            cycle.put(lorg, par[lorg]);
            added[lorg] = true;
            int l1 = par[lorg];
            while (l1 != lorg) {
              cycle.put(l1, par[l1]);
              added[l1] = true;
              l1 = par[l1];

            }
            cycles.add(cycle);
            break;
          }
          cycle.put(l, 0);
          l = par[l];
          if (added[l] && !cycle.contains(l))
            break;
          added[l] = true;
        }
      }

      // get all edges and return them
      if (cycles.size() == 0) {
        // System.out.println("TREE:");
        for (int i = 0; i < par.length; i++) {
          if (!curr_nodes[i])
            continue;
          if (par[i] != -1) {
            int pr = oldI[par[i]][i];
            int ch = oldO[par[i]][i];
            final_edges.put(ch, pr);
            // System.out.print(pr+"|"+ch + " ");
          } else
            final_edges.put(0, -1);
        }
        // System.out.println();
        return final_edges;
      }

      int max_cyc = 0;
      int wh_cyc = 0;
      for (int i = 0; i < cycles.size(); i++) {
        TIntIntHashMap cycle = (TIntIntHashMap) cycles.get(i);
        if (cycle.size() > max_cyc) {
          max_cyc = cycle.size();
          wh_cyc = i;
        }
      }

      TIntIntHashMap cycle = (TIntIntHashMap) cycles.get(wh_cyc);
      int[] cyc_nodes = cycle.keys();
      int rep = cyc_nodes[0];

      if (print) {
        System.out.println("Found Cycle");
        for (int i = 0; i < cyc_nodes.length; i++)
          System.out.print(cyc_nodes[i] + " ");
        System.out.println();
      }

      double cyc_weight = 0.0;
      for (int j = 0; j < cyc_nodes.length; j++) {
        cyc_weight += scoreMatrix[par[cyc_nodes[j]]][cyc_nodes[j]];
      }

      for (int i = 0; i < numWords; i++) {

        if (!curr_nodes[i] || cycle.contains(i))
          continue;

        double max1 = Double.NEGATIVE_INFINITY;
        int wh1 = -1;
        double max2 = Double.NEGATIVE_INFINITY;
        int wh2 = -1;

        for (int j = 0; j < cyc_nodes.length; j++) {
          int j1 = cyc_nodes[j];

          if (scoreMatrix[j1][i] > max1) {
            max1 = scoreMatrix[j1][i];
            wh1 = j1;// oldI[j1][i];
          }

          // cycle weight + new edge - removal of old
          double scr = cyc_weight + scoreMatrix[i][j1] - scoreMatrix[par[j1]][j1];
          if (scr > max2) {
            max2 = scr;
            wh2 = j1;// oldO[i][j1];
          }
        }

        scoreMatrix[rep][i] = max1;
        oldI[rep][i] = oldI[wh1][i];// wh1;
        oldO[rep][i] = oldO[wh1][i];// oldO[wh1][i];
        scoreMatrix[i][rep] = max2;
        oldO[i][rep] = oldO[i][wh2];// wh2;
        oldI[i][rep] = oldI[i][wh2];// oldI[i][wh2];

      }

      TIntIntHashMap[] rep_cons = new TIntIntHashMap[cyc_nodes.length];
      for (int i = 0; i < cyc_nodes.length; i++) {
        rep_cons[i] = new TIntIntHashMap();
        int[] keys = reps[cyc_nodes[i]].keys();
        Arrays.sort(keys);
        if (print)
          System.out.print(cyc_nodes[i] + ": ");
        for (int j = 0; j < keys.length; j++) {
          rep_cons[i].put(keys[j], 0);
          if (print)
            System.out.print(keys[j] + " ");
        }
        if (print)
          System.out.println();
      }

      // don't consider not representative nodes
      // these nodes have been folded
      for (int i = 1; i < cyc_nodes.length; i++) {
        curr_nodes[cyc_nodes[i]] = false;
        int[] keys = reps[cyc_nodes[i]].keys();
        for (int j = 0; j < keys.length; j++)
          reps[rep].put(keys[j], 0);
      }

      chuLiuEdmonds(scoreMatrix, curr_nodes, oldI, oldO, print, final_edges, reps);

      // check each node in cycle, if one of its representatives
      // is a key in the final_edges, it is the one.
      int wh = -1;
      boolean found = false;
      for (int i = 0; i < rep_cons.length && !found; i++) {
        int[] keys = rep_cons[i].keys();
        for (int j = 0; j < keys.length && !found; j++) {
          if (final_edges.contains(keys[j])) {
            wh = cyc_nodes[i];
            found = true;
          }
        }
      }

      int l = par[wh];
      while (l != wh) {
        int ch = oldO[par[l]][l];
        int pr = oldI[par[l]][l];
        final_edges.put(ch, pr);
        l = par[l];
      }

      if (print) {
        int[] keys = final_edges.keys();
        Arrays.sort(keys);
        for (int i = 0; i < keys.length; i++)
          System.out.print(final_edges.get(keys[i]) + "|" + keys[i] + " ");
        System.out.println();
      }

      return final_edges;

    }
}
