package mstparser;

import java.io.*;

public class DependencyEvaluator {
	
	public static double[] evaluateByScript(String gold, String pred) {
		double[] ret = new double[2];
		try {
			Runtime rt = Runtime.getRuntime();
			String cmd = "perl eval.pl -g " + gold + " -s " + pred + " -o result.txt";
			System.out.println(cmd);
			Process pr = rt.exec(cmd);
			if (pr.waitFor() == 0) {
				System.out.println("eval succeed");
				BufferedReader br = new BufferedReader(new FileReader("result.txt"));
				String str = br.readLine();
				ret[1] = Double.parseDouble(str.substring(str.indexOf("= ") + 2, str.indexOf(" %")));
				str = br.readLine();
				ret[0] = Double.parseDouble(str.substring(str.indexOf("= ") + 2, str.indexOf(" %")));
				br.close();
			}
			else {
				System.out.println("eval failed");
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return ret;
	}
	
	public static boolean sameSent(DependencyInstance gold, DependencyInstance pred,
			Alphabet posAlphabet) {
		if (gold.dep.length != pred.dep.length)
			return false;
		for (int i = 1; i < gold.dep.length; i++) {
			if (!gold.origWord[i].equals(pred.origWord[i]) 
					|| (!DependencyParser.correctPos && gold.pos[i] != posAlphabet.lookupIndex("UNK") && pred.pos[i] != posAlphabet.lookupIndex("UNK") && gold.pos[i] != pred.pos[i]))
				return false;
		}
		return true;
	}

	public static double[] evaluate(String act_file, String pred_file) throws IOException {
		DependencyPipe2O pipe = new DependencyPipe2O ();

		pipe.setLabeledAndLemma(act_file);
		pipe.loadCoarseMap(act_file);

		BufferedReader in_g = new BufferedReader(new InputStreamReader(new FileInputStream(act_file),"UTF8"));
		BufferedReader in_p = new BufferedReader(new InputStreamReader(new FileInputStream(pred_file),"UTF8"));

		int total = 0; int corr = 0; int corrL = 0;
		int numsent = 0; int corrsent = 0; int corrsentL = 0;
		int noptotal = 0; int nopcorr = 0; int nopcorrL = 0;
		
		DependencyInstance gold = pipe.readConllSent(in_g, pipe.labeled);
		DependencyInstance pred = pipe.readConllSent(in_p, pipe.labeled);
		
		int cnt = 0;
		while (pred != null) {
			while (gold != null && !sameSent(gold, pred, pipe.posAlphabet)) {
				gold = pipe.readConllSent(in_g, pipe.labeled);
			}
			//if (!sameSent(gold, pred, pipe.posAlphabet)) {
			//	System.out.println("bad sent: " + cnt);
			//	for (int i = 1; i < gold.length; i++) {
			//		System.out.println(pipe.posName[pred.pos[i]] + "\t" + pipe.posName[gold.pos[i]]);
			//	}
			//}
			if (gold == null)
				break;
			
			if(gold.dep.length != pred.dep.length) 
				System.out.println("Lengths do not match");
			
			boolean whole = true;
			boolean wholeL = true;

			for(int i = 1; i < gold.dep.length; i++) {
				if(pred.dep[i] == gold.dep[i]) {
					corr++;
					if(pipe.labeled) {
						if(gold.labs[i] == pred.labs[i]) 
							corrL++;
						else 
							wholeL = false;
					}
				}
				else { 
					whole = false; wholeL = false; 
				}

				if (gold.specialPos[i] != SpecialPos.PNX) {
				//if (!gold.sentence[i].matches("[-!\"#%&'()*,./:;?@\\[\\]_{}、]+")) {
					noptotal++;
					if(pred.dep[i] == gold.dep[i]) {
						nopcorr++;
						if(pipe.labeled) {
							if(gold.labs[i] == pred.labs[i]) 
								nopcorrL++;
						}
					}
				}

			}
			total += gold.dep.length - 1;

			if(whole) corrsent++;
			if(wholeL) corrsentL++;
			numsent++;

			cnt++;
			if (cnt == DependencyParser.testSentences)
				break;

			gold = pipe.readConllSent(in_g, pipe.labeled);
    		pred = pipe.readConllSent(in_p, pipe.labeled);
		}

		in_g.close();
		in_p.close();

		System.out.println("Sent: " + cnt);
		System.out.println("Tokens: " + total);
		System.out.println("Correct: " + corr);
		System.out.println("Unlabeled Accuracy: " + ((double)corr/total));
		System.out.println("Unlabeled Complete Correct: " + ((double)corrsent/numsent));
		if(pipe.labeled) {
			System.out.println("Labeled Accuracy: " + ((double)corrL/total));
			System.out.println("Labeled Complete Correct: " + ((double)corrsentL/numsent));
		}

		System.out.println("No Punc Tokens: " + noptotal);
		System.out.println("No Punc Correct: " + nopcorr);
		System.out.println("No Punc Unlabeled Accuracy: " + ((double)nopcorr/noptotal));
		if(pipe.labeled) {
			System.out.println("No Punc Labeled Accuracy: " + ((double)nopcorrL/noptotal));
		}
		double[] ret = new double[2];
		if (act_file.indexOf("spmrl") == -1) {
			// without punc
			ret[0] = (double)nopcorr/noptotal;
			ret[1] = (double)nopcorrL/noptotal;
		}
		else {
			ret[0] = (double)corr/total;
			ret[1] = (double)corrL/total;
		}
		return ret;

	}

}
