package mstparser;

import java.io.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;
import java.util.zip.GZIPInputStream;

import mstparser.FeatureTemplate.TemplateType;

import gnu.trove.map.hash.TLongIntHashMap;

public class DependencyParser {

    public static String trainfile = null;
    public static String testfile = null;
    public static String devfile = null;		// development set
    public static String predfile = null;		// give the prediction and candidates
    
    public static boolean train = false;
    public static boolean eval = false;
    public static boolean test = false;
    public static boolean multithread = false;
    public static String modelName = "dep.model";
    public static String lossType = "punc";
    public static String decodeType = "proj";
    public static int numIters = 150;
    public static String outfile = "out.txt";
    public static String goldfile = null;
    public static int trainK = 1;
    public static int testK = 1;
    public static boolean secondOrder = true;
    public static boolean labeled = false;

	public static boolean outputWeight = false;
	
	public static int learningMode = 1;		// 0: sample rank perceptron; 1: sample rank mira;
											// 2: top-k sample rank; 3: top-k exact decoding mira update
											// 4: pseudo-max with mira
											// 5: exact mira
	
	public static boolean updateGold = true;
	public static int maxRerankK = 500;
	
	public static int testingMode = 0;		// 0: normal testing; 1: output the pruning space; 2: compare 1st np with mst certificate
											// 3: exact decode 1st order
											// 4: top-k exact decoding (don't need sampling)
	public static int maxHead = 30;
	public static int validationSize = -1;	// if > 0, do validation
	
	public static int trainSentences = 1000000;
	public static int maxLength = 700;
	public static double maxAvgHead = 200;
	public static int testSentences = 1000000;
	public double bestTestLabAcc = -1e30;
	public double bestTestUlabAcc = -1e30;
	public double bestLabAcc = -1e30;
	public double bestUlabAcc = -1e30;
	
	public static int devThread = 1;
	public static int devSplit = 1;
	public int devTimes = 0; 
	public int devStart = 1;
	public boolean isDevTesting = false;
	public boolean hasTest = false;
	public DevelopmentThread dt = null;
	
	public static int trainThread = 1;
	
	public static int pruningMode = 2;			//0: no pruning; 1: only have prediction; 2: have prediction and cand; 
												//3: have prediction, cand and score
												//4: prediction and cand in different file, need predHasCand and candExt
	public static boolean correctPos = false;
	public static boolean predHasCand = true;
	public static String candExt = "mst1o";
	
	public static boolean addPredAsFeature = false;
	
	// feature set
	public static boolean useSecondOrder = true;
	public static boolean useThirdOrder = true;
	public static boolean useGGPC = true;
	public static boolean useHB = true;
	public static boolean useAS = false;
	public static boolean usePSC = true;
	public static boolean useNonproj = true;
	public static boolean useHO = true;
	public static boolean useGlobal = true;
	// other
	public static boolean useMorphology = true;
	public static boolean useDep2 = false;
	public static boolean useTurbo = true;
	public static boolean useLargeTurbo = true;
	public static boolean usePOSFeature = true;
	public static boolean isSpmrl = false;
	
	public final static int maxFeatureNum = 7;
	
	// random seed
	public static int randomSeed = 0;
	
	// regularization
	public static double regC = 0.01;
	
	public static boolean outputScore = false;
	
    private DependencyPipe2O pipe;
    private DependencyDecoder2O decoder;
    private Parameters params;
    private Parameters devParams;
    private Parameters testParams;
    
    public static String firstOrderModel = null;
    private Parameters firstOrderParams;
    
    private boolean restart = false;
    private int restartIter = 0;

    public DependencyParser(DependencyPipe2O pipe) throws IOException {
    	this.pipe=pipe;
    	// Set up arrays
    	params = new Parameters(pipe.dataAlphabet.size());
    	decoder = new DependencyDecoder2O(pipe);
    	decoder.setProj();
    	decoder.setSeed(DependencyParser.randomSeed);
    }
    
    public Parameters getFirstOrderParams() {
    	return firstOrderParams;
    }
    
    public Parameters getDevParams() {
    	return devParams;
    }

    public Parameters getTestParams() {
    	return testParams;
    }

    public DependencyPipe2O getPipe() {
    	return pipe;
    }

    public void train(DependencyInstance[] il, String trainfile, String predfile) throws IOException {
		
    	System.out.println("About to train");
    	System.out.println("Num Feats: " + (pipe.dataAlphabet.size() - 1));

    	int i = 0;
    	devTimes = 0;
    	hasTest = false;
    	if (devThread > devSplit)
    		devThread = devSplit;
    	isDevTesting = false;
    	
    	restart = true;
    	restartIter = 0;
    	
    	for(i = 0; i < numIters; i++) {

    		System.out.println("========================");
    		System.out.println("Iteration: " + i);
    		System.out.println("========================");
    		System.out.print("Processed: ");

    		long start = System.currentTimeMillis();

    		trainingIter(il, trainfile, predfile, i+1);

    		long end = System.currentTimeMillis();
    		System.out.println("Training iter took: " + (end-start));

    	}

   		params.averageParams(decoder.updateTimes);
    	
    	// wait until dev finish
    	try {
    		while (isDevTesting()) {
    			Thread.sleep(3000);
    		}
    		
    		if (hasTest) {
    			System.out.println("should not be here");
    			System.exit(0);
        		// get last test accuracy
        		String testfile = DependencyParser.testfile;
        		String outfile = testfile + ".res";

        		System.out.println();
        		System.out.println("================================================");
        		System.out.println("Evaluate on test set");
        		double[] acc = DependencyEvaluator.evaluate(testfile, outfile);
        		bestTestUlabAcc = acc[0];
        		bestTestLabAcc = acc[1];
        	    System.out.println("dev acc: " + bestLabAcc + " " + bestUlabAcc);
        		System.out.println("================================================");
        		System.out.println();
        		
        		hasTest = false;
    		}
    		
    		if (devTimes > 0) {
    			// get last dev accuracy
        		String devfile = DependencyParser.devfile;
        		if (devfile == null) {
        			if (validationSize <= 0) {
        				// tuning on test
        				devfile = DependencyParser.testfile;
        			}
        			else {
        				devfile = DependencyParser.trainfile;
        			}
        		}
        		String outfile = devfile + ".res";
        		
        	    double[] acc = DependencyEvaluator.evaluate(devfile, outfile);
        	   // double[] acc = DependencyEvaluator.evaluateByScript(testfile, mergefile);
        	    if (acc[0] > bestUlabAcc
        	    		|| (acc[0] == bestUlabAcc && acc[1] > bestLabAcc)) {
        	    	bestLabAcc = acc[1];
        	    	bestUlabAcc = acc[0];
        	    	System.out.println("update best " + bestLabAcc + " " + bestUlabAcc);
        	    	saveModel(modelName + ".best", devParams);
        	    }
        	    System.out.println("best acc: " + bestLabAcc + " " + bestUlabAcc);
    		}
    	} catch (InterruptedException e) {
    		e.printStackTrace();
    	}
    }

    private void trainingIter(DependencyInstance[] il, String trainfile, String predfile, int iter) throws IOException {

		BufferedReader in_g = new BufferedReader(new InputStreamReader(new FileInputStream(trainfile),"UTF8"));
		BufferedReader in_p = null;
		if (learningMode == 0 || learningMode == 1 || learningMode == 4 || learningMode == 5)
			in_p = new BufferedReader(new InputStreamReader(new FileInputStream(predfile),"UTF8"));
		else if (learningMode == 2 || learningMode == 3)
			in_p = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(predfile + ".gz")),"UTF8"));
		
		ObjectInputStream in_s = null;
		if (DependencyParser.pruningMode == 3)
			in_s = new ObjectInputStream(new GZIPInputStream(new FileInputStream(predfile + ".score.gz")));
		
		BufferedReader in_c = null;
		if (DependencyParser.pruningMode == 4) {
			// candidate in a separate file
			String candfile = predfile.substring(0, predfile.lastIndexOf(".")) + "." + candExt;
			in_c = new BufferedReader(new InputStreamReader(new FileInputStream(candfile),"UTF8"));
		}

		//ObjectInputStream in_f = null;
		//if (learningMode == 2 || learningMode == 3)
		//	in_f = new ObjectInputStream(new GZIPInputStream(new FileInputStream(trainfile + ".forest")));
		
		int labCorr = 0;
		int ulabCorr = 0;
		int token = 0;

		if (trainThread <= 1) {
			// single thread

			for(int i = 0; i < il.length; i++) {
	    		if((i+1) % 500 == 0)
	    			System.out.print("  "+(i+1));

	    		DependencyInstance inst = il[i];
	    		int length = inst.length;
	    		
	    		DependencyInstance gold = pipe.readConllSent(in_g, pipe.labeled);
	    		DependencyInstance pred = pipe.readConllSent(in_p, pipe.labeled);
	    		pred.m = gold.m;
	    		pred.posA = gold.posA;
	    		pred.cm = gold.cm;
	    		pred.origWord = gold.origWord;
	    		if (correctPos) {
		    		for (int z = 1; z < pred.dep.length; z++) {
		    			if (pred.sentence[z] != gold.sentence[z]
		    					|| pred.goldPos[z] != gold.pos[z]) {
		    				System.out.println("train bug 1");
		    			}
		    		}
	    		}
	    		else {
		    		for (int z = 1; z < pred.dep.length; z++) {
		    			if (pred.sentence[z] != gold.sentence[z]
		    					|| pred.pos[z] != gold.pos[z]) {
		    				System.out.println("train bug 1");
		    			}
		    		}
	    		}
	    		
	    		if (gold.sentence.length != length) {
	    			System.out.println("train length mismatch");
	    		}
	    		
	    		pipe.copyPredDepLab(gold, pred.dep, pred.labs);
				pred.parserPredDep = gold.parserPredDep;
				pred.parserPredLab = gold.parserPredLab;

	    		if (learningMode == 0 || learningMode == 1) {
	    			// sample rank model

	    			LocalFeatureData lfd = new LocalFeatureData(length, pipe.types.length, true);
		    		if (DependencyParser.pruningMode == 0) {
		    			// currently the pred-file should still exist, lazy to change the code...
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    			in_p.readLine();
		    			for (int z = 1; z < gold.dep.length; z++)
		    				in_p.readLine();
		    			in_p.readLine();
		    		}
		    		else if (DependencyParser.pruningMode == 1) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    		}
		    		else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
			    		in_p.readLine();	// score
			    		lfd.loadCandidate(in_p, pipe, pred, gold, true);
			    		in_p.readLine();	// empty line
		    		}
		    		else if (DependencyParser.pruningMode == 4) {
		    			if (DependencyParser.predHasCand) {
			    			in_p.readLine();
			    			for (int z = 1; z < gold.dep.length; z++)
			    				in_p.readLine();
			    			in_p.readLine();
		    			}
		    			// load real candidate
		    			pipe.readConllSent(in_c, pipe.labeled);
		    			in_c.readLine();
		    			lfd.loadCandidate(in_c, pipe, pred, gold, true);
		    			in_c.readLine();
		    		}
		    		
		    		if (gold.sentence.length > DependencyParser.maxLength
		    				|| ((double)lfd.maxid / gold.sentence.length > DependencyParser.maxAvgHead) )
		    			continue;
		    		//in_g.readLine();	// empty line
		    		
		    		if (!pipe.isNotDev(i))
		    			continue;
	
					if (DependencyParser.pruningMode == 3) {
						gold.parserPredScore = new LocalFeatureData(gold.dep.length, pipe.possibleLab.size(), true);
						gold.parserPredScore.loadParserPredScore(in_s, true);
					}
					pred.parserPredScore = gold.parserPredScore;
		    		
		    		GlobalFeatureData gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
		    		
		    		if (correctPos)
		    			gold.fv = pipe.createFeatureVector(gold, lfd, -1);
		    		else
		    			gold.fv = pipe.createFeatureVector(gold, lfd, gfd, -1);
	    			
	    			// use il to store the sample result from last iteration
		    		// pos restart every time, lazy to change...
	    			//int restartIter = 20;
		    		//int restartIter = 50;
	    			//boolean restart = false;
	    			//if (iter == 1 || iter % restartIter == 1) {
	    			if (restart) {
	    				// first iteration, copy the parser prediction to il
	    				if (pruningMode == 0)
	    					decoder.initSample(pred, lfd);
	    				decoder.copyDepLab(il[i], pred.dep, pred.labs);
	    			}
	    			else {
	    				// after first iteration, copy sample result to pred
	    				decoder.copyDepLab(pred, il[i].dep, il[i].labs);
	    			}
	    			pipe.buildChild(pred);

	    			if (correctPos)
	    				decoder.sampleRankWithPos(gold, pred, lfd, gfd, params, iter);
	    			else {
	    				//decoder.sampleRank(gold, pred, lfd, gfd, params, iter);
	    				decoder.sampleRankRW(gold, pred, lfd, gfd, params, firstOrderParams, restart, iter - restartIter);
	    			}
	    				//decoder.pseudoMaxTrainingWithSample(gold, pred, lfd, gfd, params, iter);
	    			//copy sample result to il
	    			decoder.copyDepLab(il[i], pred.dep, pred.labs);
	    		}
	    		else if (learningMode == 2 || learningMode == 3) {
	    			// re-ranking model
	    			if (gold.sentence.length > DependencyParser.maxLength)
	    				continue;
	    			
	    			// compute feature vector for top-k list
	    			FeatureVector[] fvList = new FeatureVector[DependencyParser.maxRerankK];
	    			HashMap<Integer, String> labChange = pipe.processChangeStr(in_p.readLine());
	    			HashMap<Integer, String> depChange = pipe.processChangeStr(in_p.readLine());
	    			int[][] depList = new int[DependencyParser.maxRerankK][pred.dep.length];
	    			int[][] labList = new int[DependencyParser.maxRerankK][pred.labs.length];
	    			
	    			pred.parserRerankScore = 1.0;
	    			fvList[0] = pipe.createFeatureVector(pred);
	    			//fvList[0] = pipe.readFeatureVector(in_f);
	    			
	    			for (int z = 0; z < pred.dep.length; z++) {
	    				depList[0][z] = pred.dep[z];
	    				labList[0][z] = pred.labs[z];
	    			}
	    			
	    			double minErr = decoder.numError(gold.dep, gold.labs, pred.dep, pred.labs);
	    			int target = 0;

	    			int k = 1;
	    			String str = in_p.readLine();
	    			while (!str.isEmpty()) {
	    				if (k < DependencyParser.maxRerankK) {
		    				String[] data = str.split("\t");
		    				pred.parserRerankScore = Double.parseDouble(data[0]);
		    				for (int z = 1; z < data.length; z++) {
		    					int changeID = Integer.parseInt(data[z]);
		    					if (depChange.containsKey(changeID)) {
		    						String s = depChange.get(changeID);
		    						int p = s.indexOf(":");
		    						int mod = Integer.parseInt(s.substring(0, p));
		    						int head = Integer.parseInt(s.substring(p + 1));
		    						pred.dep[mod] = head;
		    					}
		    					else if (labeled && labChange.containsKey(z)) {
		    						String s = depChange.get(changeID);
		    						int p = s.indexOf(":");
		    						int mod = Integer.parseInt(s.substring(0, p));
		    						int lab = pipe.typeAlphabet.lookupIndex(s.substring(p + 1));
		    						pred.labs[mod] = lab;
		    					}
		    				}
		    				pipe.buildChild(pred);
		    				fvList[k] = pipe.createFeatureVector(pred);
		    				//fvList[k] = pipe.readFeatureVector(in_f);
		    				
			    			for (int z = 0; z < pred.dep.length; z++) {
			    				depList[k][z] = pred.dep[z];
			    				labList[k][z] = pred.labs[z];
			    			}
		    				
		    				double error = decoder.numError(gold.dep, gold.labs, pred.dep, pred.labs);
		    				if (error < minErr) {
		    					minErr = error;
		    					target = k;
		    				}
		    				
		    				// recover
		    				decoder.copyDepLab(pred, depList[0], labList[0]);
		    				pipe.buildChild(pred);
		    				
	    					k++;
	    				}
	    				str = in_p.readLine();
	    			}
	    			//int idx = in_f.readInt();
	    			//if (idx != -1) {
	    			//	System.out.println("read forest bug");
	    			//	System.exit(0);
	    			//}
	    			
		    		if (!pipe.isNotDev(i))
		    			continue;

		    		// begin to train
	    			if (learningMode == 2) {
	    				// sample rank
		    			int restartIter = 25;
		    			if (iter == 1 || iter % restartIter == 1) {
		    				// first iteration, initial sample is the first one
		    				if (il[i].dep == null)
		    					il[i].dep = new int[1];
		    				il[i].dep[0] = 0;
		    			}
		    			int sample = decoder.trainReranking(gold, pred, fvList, target, il[i].dep[0], depList, labList, 
		    					params, iter);
		    			il[i].dep[0] = sample;
	    			}
	    			else if (learningMode == 3) {
	    				decoder.trainReranking(gold, pred, fvList, target, 0, depList, labList, 
		    					params, iter);
	    			}
	    		}
	    		else if (learningMode == 4) {
	    			// pseudo-max
	    			
	    			LocalFeatureData lfd = new LocalFeatureData(length, pipe.types.length, true);
		    		if (DependencyParser.pruningMode == 0) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    			in_p.readLine();
		    			for (int z = 1; z < gold.dep.length; z++)
		    				in_p.readLine();
		    			in_p.readLine();
		    		}
		    		else if (DependencyParser.pruningMode == 1) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    		}
		    		else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
			    		in_p.readLine();	// score
			    		lfd.loadCandidate(in_p, pipe, pred, gold, true);
			    		in_p.readLine();	// empty line
		    		}
		    		else if (DependencyParser.pruningMode == 4) {
		    			if (DependencyParser.predHasCand) {
			    			in_p.readLine();
			    			for (int z = 1; z < gold.dep.length; z++)
			    				in_p.readLine();
			    			in_p.readLine();
		    			}
		    			// load real candidate
		    			pipe.readConllSent(in_c, pipe.labeled);
		    			in_c.readLine();
		    			lfd.loadCandidate(in_c, pipe, pred, gold, true);
		    			in_c.readLine();
		    		}
		    		
		    		if (gold.sentence.length > DependencyParser.maxLength
		    				|| ((double)lfd.maxid / gold.sentence.length > DependencyParser.maxAvgHead) )
		    			continue;
		    		//in_g.readLine();	// empty line
	
		    		if (!pipe.isNotDev(i))
		    			continue;

		    		if (DependencyParser.pruningMode == 3) {
						gold.parserPredScore = new LocalFeatureData(gold.dep.length, pipe.possibleLab.size(), true);
						gold.parserPredScore.loadParserPredScore(in_s, true);
					}
					pred.parserPredScore = gold.parserPredScore;
		    		
		    		GlobalFeatureData gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
		    		
		    		if (correctPos)
		    			gold.fv = pipe.createFeatureVector(gold, lfd, -1);
		    		else
		    			gold.fv = pipe.createFeatureVector(gold, lfd, gfd, -1);
	    			
    				decoder.pseudoMaxTraining(gold, pred, lfd, gfd, params, iter);

	    		}
	    		else if (learningMode == 5) {
	    			// first order MST
	    			LocalFeatureData lfd = new LocalFeatureData(length, pipe.types.length, true);
		    		if (DependencyParser.pruningMode == 0) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    			in_p.readLine();
		    			for (int z = 1; z < gold.dep.length; z++)
		    				in_p.readLine();
		    			in_p.readLine();
		    		}
		    		else if (DependencyParser.pruningMode == 1) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    		}
		    		else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
			    		in_p.readLine();	// score
			    		lfd.loadCandidate(in_p, pipe, pred, gold, true);
			    		in_p.readLine();	// empty line
		    		}
		    		else if (DependencyParser.pruningMode == 4) {
		    			if (DependencyParser.predHasCand) {
			    			in_p.readLine();
			    			for (int z = 1; z < gold.dep.length; z++)
			    				in_p.readLine();
			    			in_p.readLine();
		    			}
		    			// load real candidate
		    			pipe.readConllSent(in_c, pipe.labeled);
		    			in_c.readLine();
		    			lfd.loadCandidate(in_c, pipe, pred, gold, true);
		    			in_c.readLine();
		    		}
		    		
		    		if (gold.sentence.length > DependencyParser.maxLength
		    				|| ((double)lfd.maxid / gold.sentence.length > DependencyParser.maxAvgHead) )
		    			continue;
		    		//in_g.readLine();	// empty line
	
		    		if (!pipe.isNotDev(i))
		    			continue;

					if (DependencyParser.pruningMode == 3) {
						gold.parserPredScore = new LocalFeatureData(gold.dep.length, pipe.possibleLab.size(), true);
						gold.parserPredScore.loadParserPredScore(in_s, true);
					}
					pred.parserPredScore = gold.parserPredScore;
		    		
		    		GlobalFeatureData gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
		    		
	    			gold.fv = pipe.createFeatureVector(gold, lfd, gfd, -1);
	    			
	    			Object[][] d = decoder.decodeNonProjective(pred, gold, lfd, params, 1);
	    			String[] res = ((String)d[0][1]).split(" ");
	    			for (int j = 1; j < pred.sentence.length; j++) {
	    				String[] trip = res[j-1].split("[\\|:]");
	    				pred.dep[j] = Integer.parseInt(trip[0]);
	    				pred.labs[j] = Integer.parseInt(trip[2]);
	    			}
	    			FeatureVector newFV = gold.fv;
	    			double newScore = params.getScore(newFV);
	    			//FeatureVector oldFV = (FeatureVector)d[0][0];
	    			FeatureVector oldFV = pipe.createFeatureVector(pred, lfd, gfd, -1);
	    			double oldScore = params.getScore(oldFV);
	    			double comp = decoder.numError(gold.dep, gold.labs, pred.dep, pred.labs);
	    			double upd = (double)decoder.updateTimes;
/*
	    			if (i == 500) {
	    				System.out.println((String)d[0][1]);
	    				for (int j = 1; j < pred.dep.length; j++) {
	    					System.out.println(j + "\t" + pred.dep[j] + "\t" + gold.dep[j] + "\t" + pred.sentence[j] + "\t" + pred.pos[j] + "\t" + gold.sentence[j] + "\t" + gold.pos[j]);
	    				}
	    				System.out.println(newScore + "\t" + oldScore + "\t" + (Double)d[0][2]);
	    				
	    				double s_g = 0.0;
	    				for (int j = 1; j < gold.dep.length; j++)
	    					s_g += lfd.probs[gold.dep[j]][j];
	    				double s_p = 0;
	    				for (int j = 1; j < pred.dep.length; j++)
	    					s_p += lfd.probs[pred.dep[j]][j];
	    				System.out.println(s_g + "\t" + s_p);
	    				System.exit(0);
	    			}
	    			*/
	    			params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, comp, upd);
	    			decoder.updateTimes++;
	    		}

	    		// train accuracy
	    		token += length - 1;
	    		for (int j = 1; j < length; j++) {
	    			if (pred.dep[j] == gold.dep[j])
	    				ulabCorr++;
	    			if (pred.dep[j] == gold.dep[j] && pred.labs[j] == gold.labs[j])
	    				labCorr++;
	    		}
	    		
	    	}

			// set the restart
    		if ((double)ulabCorr / token >= 0.96) {
    			restart = true;
    			restartIter = iter;
    			System.out.println("restart mavkov chain!!!!!!!!!!");
    		}
    		else {
    			restart = false;
    		}
		}
		else {
			System.out.println("not support multi thread");
			System.exit(0);
			// multi-thread
			System.out.println("multi thread train: " + trainThread);
			Object readLock = new Object();
			Object paramsLock = new Object();
			Integer[] currentID = new Integer[1];
			currentID[0] = 0;
			Integer[] stats = new Integer[3];
			stats[0] = stats[1] = stats[2] = 0;
			
			SampleRankThread[] srt = new SampleRankThread[trainThread];
			// create train thread
			for (int i = 0; i < trainThread; i++) {
				srt[i] = new SampleRankThread(in_g, in_p, params, pipe, decoder, currentID, stats, il, iter, readLock, paramsLock, i);
			}
			for (int i = 0; i < trainThread; i++) {
				srt[i].join();
			}
			
			token = stats[0];
			ulabCorr = stats[1];
			labCorr = stats[2];
		}

		System.out.println("");

    	System.out.println("  "+il.length+" instances");
    	System.out.println((double)ulabCorr / token + " " + (double)labCorr / token);

    	//bw.close();
    	in_p.close();
    	in_g.close();
		if (DependencyParser.pruningMode == 3) {
			in_s.close();
		}
		
		if (DependencyParser.pruningMode == 4)
			in_c.close();
		
		//if (learningMode == 2 || learningMode == 3)
		//	in_f.close();

    	if (iter >= devStart)
    		checkDevStatus(iter);
    }
    
    public boolean isDevTesting() {
    	return isDevTesting; 
    }
    
/*    public void mergeOutput(String outfile, String mergefile) throws IOException {
    	BufferedWriter bw = new BufferedWriter(new FileWriter(mergefile));
    	BufferedReader[] br = new BufferedReader[devThread];
    	for (int i = 0; i < devThread; i++)
    		br[i] = new BufferedReader(new FileReader(outfile + i));
    	
    	int cnt = 0;
    	String str = br[0].readLine();
    	while (str != null) {
    		for (int i = 0; i < devThread; i++) {
    			ArrayList<String> data = new ArrayList<String>();
    			if (i == 0)
    				data.add(str);
    			
    			str = br[i].readLine();
    			while (!str.isEmpty()) {
    				data.add(str);
    				str = br[i].readLine();
    			}
    			
    			int j = cnt % devSplit < devThread ? cnt % devSplit : 0;
    			if (i == j) {
    				for (int k = 0; k < data.size(); k++) {
    					bw.write(data.get(k) + "\n");
    				}
    				bw.newLine();
    			}
    		}
    		str = br[0].readLine();
    		cnt++;
    	}
    	
    	bw.close();
    	for (int i = 0; i < devThread; i++)
    		br[i].close();
    }*/
    
    public void checkDevStatus(int iter) throws IOException {
    	if (isDevTesting()) {
    		System.out.print("processing sentences: ");
    		synchronized (dt.currID[1]) {
    			System.out.print(dt.currID[1][0] + "\t");
    		}
    		synchronized (dt.currID[0]) {
    			System.out.println(dt.currID[0][0]);
    		}
    	}
    	else {
    		System.out.println("dev and test done.");
    		//outputWeight("param.txt");
    		
    		if (hasTest) {
    			System.out.println("should not be here");
    			System.exit(0);
        		// get last test accuracy
        		String testfile = DependencyParser.testfile;
        		String outfile = testfile + ".res";

        		System.out.println();
        		System.out.println("================================================");
        		System.out.println("Evaluate on test set");
        		double[] acc = DependencyEvaluator.evaluate(testfile, outfile);
        		bestTestUlabAcc = acc[0];
        		bestTestLabAcc = acc[1];
        	    System.out.println("dev acc: " + bestLabAcc + " " + bestUlabAcc);
        		System.out.println("================================================");
        		System.out.println();
        		
        		hasTest = false;
    		}
    		
    		if (devTimes > 0) {
    			// get last dev accuracy
        		String devfile = DependencyParser.devfile;
        		if (devfile == null) {
        			if (validationSize <= 0) {
        				// tuning on test
        				devfile = DependencyParser.testfile;
        			}
        			else {
        				devfile = DependencyParser.trainfile;
        			}
        		}
        		String outfile = devfile + ".res";
        		
        		System.out.println("Evaluate on dev set");
        	    double[] acc = DependencyEvaluator.evaluate(devfile, outfile);
        	    if (acc[0] > bestUlabAcc
        	    		|| (acc[0] == bestUlabAcc && acc[1] > bestLabAcc)) {
        	    	bestLabAcc = acc[1];
        	    	bestUlabAcc = acc[0];
        	    	System.out.println("update best " + bestLabAcc + " " + bestUlabAcc);
        	    	saveModel(modelName + ".best", devParams);
        	    	if (DependencyParser.devfile != null) {
        	    		// try on testing set
        	    		hasTest = true;
        	    	}
        	    }
        	    System.out.println("best acc: " + bestLabAcc + " " + bestUlabAcc);
        	    System.out.println("best test acc: " + bestTestLabAcc + " " + bestTestUlabAcc);
    		}
    		
    		if (hasTest) {
    			System.out.println("should not be here");
    			System.exit(0);
    			// dev file should not be null
        		String devfile = DependencyParser.devfile;
        		String devoutfile = devfile + ".res";
        		String devpredfile = DependencyParser.predfile.replace("train2", "dev").replace("train", "dev");
        		String testfile = DependencyParser.testfile;
        		String testoutfile = testfile + ".res";
        		String testpredfile = DependencyParser.predfile.replace("train2", "test").replace("train", "test");
    			
    			// start new thread for both test and dev
    			testParams = devParams;
    			System.out.println("build dev params");
    			devParams = new Parameters(params);
       			devParams.averageParams((double)decoder.updateTimes);
       			
        		System.out.println("start new dev and test " + devTimes);
        		isDevTesting = true;
       			dt = new DevelopmentThread(this, this.pipe, this.decoder,
       					devfile, devpredfile, devoutfile, testfile, testpredfile, testoutfile, false);
        		devTimes++;
    		}
    		else {
    			// start new thread for dev
        		String devfile = DependencyParser.devfile;
        		String devpredfile = DependencyParser.predfile.replace("train2", "dev").replace("train", "dev");
        		if (devfile == null) {
        			if (validationSize <= 0) {
	        			// tuning on test
	        			devfile = DependencyParser.testfile;
	       				devpredfile = DependencyParser.predfile.replace("train2", "test").replace("train", "test");
	        			//devpredfile = DependencyParser.predfile;
        			}
        			else {
	        			devfile = DependencyParser.trainfile;
	       				devpredfile = DependencyParser.predfile;
        			}
        		}
        		String devoutfile = devfile + ".res";
        		//System.out.println(devfile);
        		//System.out.println(devpredfile);
        		
    			System.out.println("build dev params");
        		devParams = new Parameters(params);
        		if (devTimes == 0)
        			outputWeight("params.txt");
       			devParams.averageParams((double)decoder.updateTimes);
        		
        		System.out.println("start new dev " + devTimes);
        		isDevTesting = true;
        		dt = new DevelopmentThread(this, this.pipe, this.decoder,
        				devfile, devpredfile, devoutfile, null, null, null, false);
        		devTimes++;
    		}
    	}
    }

    ///////////////////////////////////////////////////////
    // Saving and loading models
    ///////////////////////////////////////////////////////
	public void outputWeight(BufferedWriter bw, TemplateType type, Parameters params) throws IOException {
		TLongIntHashMap intmap = pipe.dataAlphabet.getMap(type);
		for (long s : intmap.keys()) {
			int index = intmap.get(s);
			if (index > 0) {
				bw.write(s + "\t" + params.parameters[index]);
				if (params.total != null)
					bw.write("\t" + params.total[index]);
				bw.newLine();
			}
			else {
				System.out.println("can't find feature " + s);
			}
		}
	}

	public void outputWeight(String fStr) throws IOException {
		System.out.println("output feature weight to " + fStr);
		BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fStr),"UTF8"));
		for (Object s : pipe.dataAlphabet.strmap.keys()) {
			int index = pipe.dataAlphabet.lookupIndex((String)s);
			if (index > 0) {
				bw.write(s + "\t" + params.parameters[index]);
				if (params.total != null)
					bw.write("\t" + params.total[index]);
				bw.newLine();
			}
			else {
				System.out.println("can't find feature " + s.toString());
			}
		}
		
		outputWeight(bw, TemplateType.TArcMST, params);
		outputWeight(bw, TemplateType.TArcTurbo, params);
		
		bw.close();
	}

	public void saveModel(String file, Parameters params) throws IOException {
		ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file));
		out.writeObject(params.parameters);
		out.writeObject(pipe.dataAlphabet);
		out.writeObject(pipe.typeAlphabet);
		out.writeObject(pipe.morphologyIDAlphabet);
		out.writeObject(pipe.morphologyValueAlphabet);
		out.writeObject(pipe.cposAlphabet);
		out.writeObject(pipe.possibleLab);
		out.writeObject(pipe.posAlphabet);
		out.writeObject(pipe.lexAlphabet);
		out.close();
	}

	public void loadModel(String file) throws Exception {
		ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
		params.parameters = (double[])in.readObject();
		pipe.dataAlphabet = (FeatureAlphabet)in.readObject();
		pipe.typeAlphabet = (Alphabet)in.readObject();
		pipe.morphologyIDAlphabet = (Alphabet)in.readObject();
		pipe.morphologyValueAlphabet = (Alphabet)in.readObject();
		pipe.cposAlphabet = (Alphabet)in.readObject();
		pipe.possibleLab = (HashSet<String>)in.readObject();
		pipe.posAlphabet = (Alphabet)in.readObject();
		pipe.lexAlphabet = (Alphabet)in.readObject();
		in.close();
		pipe.closeAlphabets();
		
		pipe.setAndCheckOffset();
		
		// debug
		//for (int z = 0; z < 8; z++) {
		//	if (z == 4)
		//		continue;
		//	String nfile = file.replace("4", new Integer(z).toString());
		//	in = new ObjectInputStream(new FileInputStream(nfile));
		//	double[] tmpParams = (double[])in.readObject();
		//	in.close();
		//	for (int i = 0; i < params.parameters.length; i++) {
		//		params.parameters[i] = (params.parameters[i] + tmpParams[i]);
		//	}
		//}
	}
	
	public void loadFirstOrderModel(String file) throws Exception {
    	if (file != null) {
    		System.out.println("build first order params: " + pipe.dataAlphabet.size());
    		firstOrderParams = new Parameters(pipe.dataAlphabet.size());
    		ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
    		double[] oldParams = (double[])in.readObject();
    		FeatureAlphabet oldAlphabet = (FeatureAlphabet)in.readObject();

    		// map arc mst map
    		long[] keys = pipe.dataAlphabet.arcMSTMap.keys();
    		for (int i = 0; i < keys.length; i++) {
    			int oldID = oldAlphabet.lookupIndex(FeatureTemplate.TemplateType.TArcMST, keys[i]);
    			if (oldID > 0) {
    				int newID = pipe.dataAlphabet.lookupIndex(FeatureTemplate.TemplateType.TArcMST, keys[i]);
    				firstOrderParams.parameters[newID] = oldParams[oldID];
    			}
    		}

    		// map arc turbo map
    		keys = pipe.dataAlphabet.arcTurboMap.keys();
    		for (int i = 0; i < keys.length; i++) {
    			int oldID = oldAlphabet.lookupIndex(FeatureTemplate.TemplateType.TArcTurbo, keys[i]);
    			if (oldID > 0) {
    				int newID = pipe.dataAlphabet.lookupIndex(FeatureTemplate.TemplateType.TArcTurbo, keys[i]);
    				firstOrderParams.parameters[newID] = oldParams[oldID];
    			}
    		}

    		// map catib
    		keys = pipe.dataAlphabet.catibMap.keys();
    		for (int i = 0; i < keys.length; i++) {
    			int oldID = oldAlphabet.lookupIndex(FeatureTemplate.TemplateType.TCatib, keys[i]);
    			if (oldID > 0) {
    				int newID = pipe.dataAlphabet.lookupIndex(FeatureTemplate.TemplateType.TCatib, keys[i]);
    				firstOrderParams.parameters[newID] = oldParams[oldID];
    			}
    		}
    		in.close();
    	}		
	}

    //////////////////////////////////////////////////////
    // Get Best Parses ///////////////////////////////////
    //////////////////////////////////////////////////////
    
	public void outputConllSentence(BufferedWriter bw, DependencyInstance pred, String predStr) throws IOException {
		String[] res = predStr.split(" ");
		for (int i = 1; i < pred.sentence.length; i++) {
			String pos = pred.pos[i] <= 0 ? "UNK" : pipe.posName[pred.pos[i]];
			bw.write("" + i + "\t" + pred.origWord[i] + "\t" + pred.origWord[i] + "\t");
			if (correctPos) {
				if (pred.goldPos[i] <= DependencyPipe2O.NOTYPE) {
					System.out.println("output sentence bug: unseen gold pos");
				}
				boolean find = false;
				for (int j = 0; j < pred.candPos[i].length; j++) {
					if (pred.candPos[i][j] <= DependencyPipe2O.NOTYPE)
						System.out.println("output sentence bug 2");
					if (pred.candPos[i][j] == pred.pos[i]) {
						find = true;
						if (j != pred.currPosCandID[i]) {
							System.out.println("output sentence bug 3");
						}
					}
				}
				//if (pred.currPosCandID[i] != 0) {
				//	System.out.println("change pos at: " + i + "\t" + pred.origWord[i] + "\t" + pipe.posName[pred.pos[i]]);
				//}
				if (!find)
					System.out.println("output sentence bug 3");
				bw.write(pipe.posName[pred.goldPos[i]] + "\t" + pipe.posName[pred.pos[i]]);
				for (int j = 0; j < pred.candPos[i].length; j++) {
					if (j == pred.currPosCandID[i])
						continue;
					bw.write("||" + pipe.posName[pred.candPos[i][j]]);
				}
				bw.write("\t_\t");
			}
			else {
				bw.write(pos + "\t" + pos + "\t_\t");
			}
			String[] trip = res[i-1].split("[\\|:]");
			bw.write(trip[0] + "\t" + (pipe.labeled ? pipe.types[Integer.parseInt(trip[2])] : "_") + "\t_\t_\n");
		}
		bw.newLine();
		bw.flush();
	}
	
/*	public void outputConllSentence(BufferedWriter bw, DependencyInstance gold, DependencyInstance pred) throws IOException {
		for (int i = 1; i < gold.sentence.length; i++) {
			String pos = gold.pos[i] == -1 ? "UNK" : pipe.posName[gold.pos[i]];
			bw.write("" + i + "\t" + gold.origWord[i] + "\t" + gold.origWord[i] + "\t" + pos + "\t" + pos + "\t_\t");
			bw.write(pred.dep[i] + "\t" + (pipe.labeled ? pipe.types[pred.labs[i]] : "_") + "\t_\t_\n");
		}
		bw.newLine();
	}*/
	
	public void generateCandidate(BufferedWriter bw,
			DependencyInstance pred, DependencyInstance gold, String predStr,
			Parameters params, Integer[] stats) throws IOException {
		
		LocalFeatureData lfd = new LocalFeatureData(pred.dep.length, pipe.types.length, true);
		lfd.loadCandidate(null, pipe, pred, null, true);

		outputConllSentence(bw, pred, predStr);
		
		double thresh = Math.log(0.005);
		String scoreStr = "";
		for (int i = 1; i < pred.dep.length; i++) {
			double score = pipe.getLocalScore(pred, lfd, params, i);
			scoreStr += String.format("%.5f", score) + "\t";
		}
		bw.write(scoreStr.trim() + "\n");
		
		stats[0] += pred.dep.length - 1;
		for (int i = 1; i < pred.dep.length; i++) {
			if (correctPos && pred.pos[i] == gold.pos[i])
				stats[3]++;
			
			double[] score = new double[pred.dep.length];
			int[] hid = new int[pred.dep.length];
			
			double maxScore = -1e30;
			for (int h = 0; h < pred.dep.length; h++) {
				hid[h] = h;
				if (h == i)
					continue;
				//int oldH = pred.dep[i];
				pred.dep[i] = h;
				pred.labs[i] = 1;		// unlabeled parsing
				//pipe.updateChildList(pred, h, oldH, i);
				score[h] = pipe.getLocalScore(pred, lfd, params, i);
				if (score[h] > maxScore)
					maxScore = score[h];
			}
			
			// sort
			for (int h1 = 0; h1 < pred.dep.length; h1++)
				for (int h2 = h1 + 1; h2 < pred.dep.length; h2++)
					if (score[h1] < score[h2]) {
						double tmpd = score[h1];
						score[h1] = score[h2];
						score[h2] = tmpd;
						
						int tmpi = hid[h1];
						hid[h1] = hid[h2];
						hid[h2] = tmpi;
					}
			
			String candStr = "";
			int candNum = 0;
			for (int h = 0; h < pred.dep.length; h++) {
				if (hid[h] == i)
					continue;
				if (score[h] - maxScore > thresh) {
					stats[1]++;
					candNum++;
					if (hid[h] == gold.dep[i]) {
						stats[2]++;
					}
					candStr += hid[h] + "|#" + "\t";
				}
				if (candNum >= maxHead)
					break;
			}
			bw.write(candStr.trim() + "\n");
		}
		bw.newLine();
	}

	public void outputParses(String testfile, String predfile, String outfile, 
			Parameters params, boolean verbal)
	throws IOException {
		
		System.out.println("should not go here");
		System.exit(0);

		if (outputWeight) {
			String fStr = testfile + ".weight";
			outputWeight(fStr);
			return;
		}

		long start = System.currentTimeMillis();

		BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outfile),"UTF8"));
		//BufferedWriter pred = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(file + ".gz")),"UTF8"));

		BufferedReader in_g = new BufferedReader(new InputStreamReader(new FileInputStream(testfile),"UTF8"));
		BufferedReader in_p = null;
		if (testingMode <= 3)
			in_p = new BufferedReader(new InputStreamReader(new FileInputStream(predfile),"UTF8"));
		else if (testingMode == 4)
			in_p = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(predfile + ".gz")),"UTF8"));
		
		// score file
		ObjectInputStream in_s = null;
		if (DependencyParser.pruningMode == 3)
			in_s = new ObjectInputStream(new GZIPInputStream(new FileInputStream(predfile + ".score.gz")));

		BufferedReader in_c = null;
		if (DependencyParser.pruningMode == 4) {
			// candidate in a separate file
			String candfile = predfile.substring(0, predfile.lastIndexOf(".")) + "." + candExt;
			in_c = new BufferedReader(new InputStreamReader(new FileInputStream(candfile),"UTF8"));
		}

		//BufferedWriter bw = new BufferedWriter(new FileWriter("score.txt"));
		
		if (verbal)
			System.out.print("Processing Sentence: ");
		int cnt = 0;
		DependencyInstance gold = pipe.createInstance(in_g);
		
		Integer[] stats = new Integer[4];
		stats[0] = stats[1] = stats[2] = stats[3] = 0; // 0: token; 1: candNum; 2:oracleCorr; 3: pos acc
		
		int cert = 0;

		while(gold != null) {
			DependencyInstance pred = pipe.createInstance(in_p);
    		pred.m = gold.m;
    		pred.posA = gold.posA;
    		pred.cm = gold.cm;
    		pred.origWord = gold.origWord;
    		if (correctPos) {
	    		for (int z = 1; z < pred.dep.length; z++) {
	    			if (pred.sentence[z] != gold.sentence[z]
	    					|| pred.goldPos[z] != gold.pos[z]) {
	    				System.out.println("output parse bug 1");
	    			}
	    		}
    		}
    		else {
	    		for (int z = 1; z < pred.dep.length; z++) {
	    			if (pred.sentence[z] != gold.sentence[z]
	    					|| pred.pos[z] != gold.pos[z]) {
	    				System.out.println("output parse bug 1");
	    			}
	    		}
    		}
    		
			pipe.copyPredDepLab(gold, pred.dep, pred.labs);
			if (DependencyParser.pruningMode == 3) {
				gold.parserPredScore = new LocalFeatureData(gold.dep.length, pipe.possibleLab.size(), true);
				gold.parserPredScore.loadParserPredScore(in_s, true);
			}

			pred.parserPredScore = gold.parserPredScore;
			pred.parserPredDep = gold.parserPredDep;
			pred.parserPredLab = gold.parserPredLab;

			if (verbal) {
				if (testingMode == 0 || testingMode == 2 || correctPos)
					System.out.print(cnt+" ");
				else if (cnt % 100 == 0)
					System.out.print(cnt+" ");
			}

			int[] toks = gold.sentence;
			int length = toks.length;

			if (testingMode <= 3) {
				
				LocalFeatureData lfd = new LocalFeatureData(length, pipe.types.length, true);
	    		if (DependencyParser.pruningMode == 0) {
	    			// currently the pred-file should still exist, lazy to change the code...
	    			lfd.loadCandidate(null, pipe, pred, null, true);
	    			in_p.readLine();
	    			for (int z = 1; z < gold.dep.length; z++)
	    				in_p.readLine();
	    			in_p.readLine();
	    		}
	    		else if (DependencyParser.pruningMode == 1) {
					lfd.loadCandidate(null, pipe, pred, null, true);
				}
				else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3) {
					in_p.readLine();	// score
					lfd.loadCandidate(in_p, pipe, pred, null, true);
					in_p.readLine();	// empty line
				}
	    		else if (DependencyParser.pruningMode == 4) {
	    			if (DependencyParser.predHasCand) {
		    			in_p.readLine();
		    			for (int z = 1; z < gold.dep.length; z++)
		    				in_p.readLine();
		    			in_p.readLine();
	    			}
	    			// load real candidate
	    			pipe.readConllSent(in_c, pipe.labeled);
	    			in_c.readLine();
	    			lfd.loadCandidate(in_c, pipe, pred, gold, true);
	    			in_c.readLine();
	    		}
				//in_g.readLine();	// empty line
				
				GlobalFeatureData gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
				
				if (testingMode == 0 || testingMode == 2) {
					Object[][] d = null;
		
					if (correctPos)
						d = decoder.decodeSamplingWithPos(pred, lfd, gfd, params, true);
					else
						d = decoder.decodeSampling(pred, lfd, gfd, params, true);
					
					if (testingMode == 0) {
						outputConllSentence(out, pred, (String)d[0][1]);
						if (correctPos) {
							for (int i = 1; i < pred.pos.length; i++) {
								stats[0]++;
								if (pred.pos[i] == gold.pos[i])
									stats[1]++;
							}
						}
					}
					else if (testingMode == 2) {
						Object[][] mstd = null;
						if (DependencyParser.decodeType.equals("non-proj"))
							mstd = decoder.decodeNonProjective(pred, null, lfd, params, 1);
						else
							mstd = decoder.decodeProjective(pred, lfd, params, 1);
						double sampScore = (Double)d[0][2];
						double mstScore = params.getScore((FeatureVector)mstd[0][0]);
						out.write(sampScore + "\t" + mstScore + "\t" + (Math.abs(sampScore - mstScore) < 1e-4 ? 1 : 0) + "\n");
						if (Math.abs(sampScore - mstScore) < 1e-4)
								cert++;
						out.flush();
					}
				} 
				else if (testingMode == 1) {
					Object[][] d = null;
					if (DependencyParser.correctPos) {
						d = decoder.decodeSamplingWithPos(pred, lfd, null, params, true);
					}
					else {
						if (DependencyParser.decodeType.equals("non-proj"))
							d = decoder.decodeNonProjective(pred, null, lfd, params, 1);
						else
							d = decoder.decodeProjective(pred, lfd, params, 1);
					}
					generateCandidate(out, pred, gold, (String)d[0][1], params, stats);
				}
				else if (testingMode == 3) {
					Object[][] d = null;
					if (DependencyParser.decodeType.equals("non-proj"))
						d = decoder.decodeNonProjective(pred, null, lfd, params, 1);
					else
						d = decoder.decodeProjective(pred, lfd, params, 1);
					outputConllSentence(out, pred, (String)d[0][1]);
				}
			}
			else if (testingMode == 4) {
				// decode rerank
    			// compute feature vector for top-k list
    			FeatureVector[] fvList = new FeatureVector[DependencyParser.maxRerankK];
    			HashMap<Integer, String> labChange = pipe.processChangeStr(in_p.readLine());
    			HashMap<Integer, String> depChange = pipe.processChangeStr(in_p.readLine());
    			int[][] depList = new int[DependencyParser.maxRerankK][pred.dep.length];
    			int[][] labList = new int[DependencyParser.maxRerankK][pred.labs.length];
    			
    			pred.parserRerankScore = 1.0;
    			fvList[0] = pipe.createFeatureVector(pred);
    			for (int z = 0; z < pred.dep.length; z++) {
    				depList[0][z] = pred.dep[z];
    				labList[0][z] = pred.labs[z];
    			}

    			int k = 1;
    			String str = in_p.readLine();
    			while (!str.isEmpty()) {
    				if (k < DependencyParser.maxRerankK) {
	    				String[] data = str.split("\t");
	    				pred.parserRerankScore = Double.parseDouble(data[0]);
	    				for (int z = 1; z < data.length; z++) {
	    					int changeID = Integer.parseInt(data[z]);
	    					if (depChange.containsKey(changeID)) {
	    						String s = depChange.get(changeID);
	    						int p = s.indexOf(":");
	    						int mod = Integer.parseInt(s.substring(0, p));
	    						int head = Integer.parseInt(s.substring(p + 1));
	    						pred.dep[mod] = head;
	    					}
	    					else if (labeled && labChange.containsKey(z)) {
	    						String s = depChange.get(changeID);
	    						int p = s.indexOf(":");
	    						int mod = Integer.parseInt(s.substring(0, p));
	    						int lab = pipe.typeAlphabet.lookupIndex(s.substring(p + 1));
	    						pred.labs[mod] = lab;
	    					}
	    				}
	    				pipe.buildChild(pred);
	    				fvList[k] = pipe.createFeatureVector(pred);
		    			for (int z = 0; z < pred.dep.length; z++) {
		    				depList[k][z] = pred.dep[z];
		    				labList[k][z] = pred.labs[z];
		    			}
	    				
	    				// recover
	    				decoder.copyDepLab(pred, depList[0], labList[0]);
	    				pipe.buildChild(pred);

	    				k++;
    				}
    				str = in_p.readLine();
    			}
    			
    			// choose the one with best score
    			Object[][] d = decoder.decodeReranking(pred, fvList, depList, labList, params);
    			outputConllSentence(out, pred, (String)d[0][1]);
			}
			
			cnt++;
			if (cnt == testSentences)
				break;
			gold = pipe.createInstance(in_g);
		}
		if (verbal) {
			System.out.println();
			
			if (testingMode == 0 && correctPos) {
				System.out.println("Pos Acc: " + (double)stats[1] / stats[0]);
			}
			else if (testingMode == 1) {
				System.out.println("Oracle Acc: " + (double)stats[2] / stats[0]);
				System.out.println("Avg. head: " + (double)stats[1] / stats[0]);
				if (DependencyParser.correctPos) {
					System.out.println("POS Acc: " + (double)stats[3] / stats[0]);
				}
			}
			else if (testingMode == 2) {
				System.out.println("Cert: " + cert);
				System.out.println("Sent: " + cnt);
				System.out.println("Rate: " + (double)cert / cnt);
			}
		}

		out.close();
		in_g.close();
		in_p.close();
		if (DependencyParser.pruningMode == 3)
			in_s.close();
		
		if (DependencyParser.pruningMode == 4)
			in_c.close();
		//bw.close();

		long end = System.currentTimeMillis();
		if (verbal)
			System.out.println("Took: " + (end-start));

    }

    /////////////////////////////////////////////////////
    // RUNNING THE PARSER
    ////////////////////////////////////////////////////
	public static void hashMapTest() {
		TLongIntHashMap intmap = new TLongIntHashMap();
		Random r = new Random(0);
		int size = 10000 * 1000;
		int testCase = 10000 * 10000;
		for (int i = 0; i < size; i++)
			intmap.put(i, r.nextInt());
		long start = System.currentTimeMillis();

		for (int i = 0; i < testCase; i++) {
			intmap.get(Math.abs(r.nextInt()) % size);
		}
		
		long end = System.currentTimeMillis();
		System.out.println("Training iter took: " + (end-start));
		
	}
	
    public static void main (String[] args) throws FileNotFoundException, Exception
    {
		
	processArguments(args);
		
	if(train) {
		
	    DependencyPipe2O pipe = new DependencyPipe2O ();
			
	    pipe.setLabeledAndLemma(trainfile);
	    pipe.loadCoarseMap(trainfile);
	    
	    DependencyInstance[] trainingData = pipe.createInstances(trainfile, predfile);
		
	    //pipe.closeAlphabets();
		
	    DependencyParser dp = new DependencyParser(pipe);
	    dp.loadFirstOrderModel(firstOrderModel);

	    int numFeats = pipe.dataAlphabet.size() - 1;
	    int numTypes = pipe.typeAlphabet.size() - 1;
	    System.out.println("Num Feats: " + numFeats);	
	    System.out.println("Num Edge Labels: " + numTypes);

	    dp.train(trainingData, trainfile, predfile);
	
	    System.out.print("Saving model ... ");
	    dp.saveModel(modelName, dp.params);
	    System.out.println("done.");
			
	}
		
	if (test) {
	    DependencyPipe2O pipe = new DependencyPipe2O ();
	    pipe.setLabeledAndLemma(testfile);
	    pipe.loadCoarseMap(testfile);
	    
	    DependencyParser dp = new DependencyParser(pipe);

	    System.out.println("\nLoading model ... ");
	    dp.loadModel(modelName);
	    dp.loadFirstOrderModel(firstOrderModel);
	    System.out.println("done.");

	    //pipe.closeAlphabets();

		if (multithread) {
		    // run multi-thread to test
		    devThread = devSplit;
	    	dp.isDevTesting = true;
			dp.devParams = dp.params;
			dp.dt = new DevelopmentThread(dp, dp.pipe, dp.decoder, 
					testfile, predfile, outfile, null, null, null, true);
		    
		    // wait until all finishes
		    while (dp.isDevTesting()) {
		    	Thread.sleep(300);
		    }
	    }
	    else {
	    	dp.outputParses(testfile, predfile, outfile, dp.params, true);
	    }
	}
		
	if(eval) {
	    System.out.println("\nEVALUATION PERFORMANCE:");
	    DependencyEvaluator.evaluate(goldfile,outfile);
	    //DependencyEvaluator.evaluateByScript(goldfile, outfile);
	}
    }

    public static void processArguments(String[] args) {
    	for(int i = 0; i < args.length; i++) {
    		String[] pair = args[i].split(":");
    		if(pair[0].equals("train")) {
    			train = true;
    		}
    		if(pair[0].equals("eval")) {
    			eval = true;
    		}
    		if(pair[0].equals("test")) {
    			test = true;
    		}
    		if (pair[0].equals("mt")) {
    			multithread = true;
    		}
    		if(pair[0].equals("iters")) {
    			numIters = Integer.parseInt(pair[1]);
    		}
    		if(pair[0].equals("output-file")) {
    			outfile = pair[1];
    		}
    		if(pair[0].equals("gold-file")) {
    			goldfile = pair[1];
    		}
    		if(pair[0].equals("train-file")) {
    			trainfile = pair[1];
    		}
    		if(pair[0].equals("test-file")) {
    			testfile = pair[1];
    		}
    		if(pair[0].equals("dev-file")) {
    			devfile = pair[1];
    		}
    		if(pair[0].equals("pred-file")) {
    			predfile = pair[1];
    		}
    		if(pair[0].equals("model-name")) {
    			modelName = pair[1];
    		}
    		if(pair[0].equals("training-k")) {
    			trainK = Integer.parseInt(pair[1]);
    		}
    		if(pair[0].equals("test-k")) {
    			testK = Integer.parseInt(pair[1]);
    		}
    		if(pair[0].equals("loss-type")) {
    			lossType = pair[1];
    		}			
    		if(pair[0].equals("order") && pair[1].equals("2")) {
    			secondOrder = true;
    		}			
     		if(pair[0].equals("decode-type")) {
    			decodeType = pair[1];
    		}			
    		if (pair[0].equals("weight")) {
    			outputWeight = true;
    		}
    		if (pair[0].equals("seed")) {
    			randomSeed = Integer.parseInt(pair[1]);
    		}
    		if (pair[0].equals("devthread")) {
    			devSplit = Integer.parseInt(pair[1]);
    			devThread = devSplit;
    		}
    		if (pair[0].equals("trainthread")) {
    			trainThread = Integer.parseInt(pair[1]);
    		}
    		if (pair[0].equals("pruningmode")) {
    			pruningMode = Integer.parseInt(pair[1]);
    		}
    		if (pair[0].equals("turbo")) {
    			//useMorphology = false;
    			useGGPC = false;
    			usePSC = false;
    			useHO = false;
    			useGlobal = false;
    		}
    		if (pair[0].equals("useturbo")) {
    			useTurbo = true;
    		}
    		if (pair[0].equals("C")) {
    			regC = Double.parseDouble(pair[1]);
    		}
    		if (pair[0].equals("updategold")) {
    			updateGold = Boolean.parseBoolean(pair[1]);
    		}
    		if (pair[0].equals("outputscore")) {
    			outputScore = true;
    		}
    	}


		String file = trainfile;
		if (file == null)
			file = testfile;
    	
		if (file.indexOf("spmrl") != -1)
			isSpmrl = true;

		if (file.indexOf("arabic") != -1)
			isSpmrl = true;

    	if (pruningMode == 0) {
			// support label parsing for spmrl
			//if (file.indexOf("spmrl") == -1)
			labeled = false;
			
			useSecondOrder = false;
			useThirdOrder = false;
			useHO = false;
			useGlobal = false;
			useHB = false;
			useNonproj = false;
			useAS = false;
			usePSC = false;
			useGGPC = false;
			usePOSFeature = false;
		}
		
		if (testingMode == 1 || testingMode == 2) {
			eval = false;
		}
		
		// better with update gold: spanish, danish, portuguese, swedish, dutch
		// better without update gold: arabic, bulgarian, chinese, japanese
		// similar: slovene, turkish
		// current: update gold with small error, find a balance 
		//if (file.indexOf("japanese") != -1) {
		//	DependencyParser.updateGold = false;
		//}
		//if (file.indexOf("spanish") != -1) {
		//	updateGold = true;
		//}
		
		if (!correctPos) {
			usePOSFeature = false;
		}
		
		if (decodeType.equals("proj")) {
			useNonproj = false;
		}
		
		//firstOrderModel = "../data/1omodel/slovene.model";

    	System.out.println("------\nFLAGS\n------");
    	System.out.println("train-file: " + trainfile);
    	System.out.println("test-file: " + testfile);
    	System.out.println("pred-file: " + predfile);
    	System.out.println("gold-file: " + goldfile);
    	System.out.println("output-file: " + outfile);
    	System.out.println("model-name: " + modelName);
    	System.out.println("train: " + train);
    	System.out.println("test: " + test);
    	System.out.println("eval: " + eval);
    	System.out.println("loss-type: " + lossType);
    	System.out.println("second-order: " + secondOrder);
    	System.out.println("training-iterations: " + numIters);
    	System.out.println("training-k: " + trainK);
    	System.out.println("decode-type: " + decodeType);
    	System.out.println("seed: " + randomSeed);
    	System.out.println("use second: " + useSecondOrder);
    	System.out.println("use third: " + useThirdOrder);
    	System.out.println("use high order: " + useHO);
    	System.out.println("use global: " + useGlobal);
    	System.out.println("use head bigram: " + useHB);
    	System.out.println("use non proj: " + useNonproj);
    	System.out.println("use arbitrary sibling: " + useAS);
    	System.out.println("use psc: " + usePSC);
    	System.out.println("use ggpc: " + useGGPC);
    	System.out.println("use morphology: " + useMorphology);
    	System.out.println("use turbo: " + useTurbo);
    	System.out.println("use large turbo: " + useLargeTurbo);
    	System.out.println("use pos feature: " + usePOSFeature);
    	System.out.println("pred as features: " + addPredAsFeature);
    	System.out.println("learning mode: " + learningMode);
    	System.out.println("pruning mode: " + pruningMode);
    	System.out.println("testing mode: " + testingMode);
    	System.out.println("update gold: " + updateGold);
    	System.out.println("train thread: " + trainThread);
    	System.out.println("dev thread: " + devThread);
    	System.out.println("labeled: " + labeled);
    	System.out.println("correct pos: " + correctPos);
    	System.out.println("reg C: " + regC);
    	System.out.println("------\n");
    }
}
