package mstparser;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.OutputStreamWriter;
import java.util.zip.GZIPInputStream;

//import java.io.*;
//import java.util.*;

public class DevelopmentThread implements Runnable {
	// multi-thread to test on the development set
	DependencyParser dp = null;
	DependencyPipe2O pipe = null;
	DependencyDecoder2O decoder = null;
	String devfile = null;
	String devpredfile = null;
	String devoutfile = null;
	String testfile = null;
	String testpredfile = null;
	String testoutfile = null;
	boolean verbal = false;
	
	public Integer[][] currID;
	
	Thread t;
	public DevelopmentThread(DependencyParser dp, 
			DependencyPipe2O pipe,
			DependencyDecoder2O decoder,
			String devfile, String devpredfile, String devoutfile, 
			String testfile, String testpredfile, String testoutfile, boolean verbal) {
		this.dp = dp;
		this.pipe = pipe;
		this.decoder = decoder;
		this.devfile = devfile;
		this.devpredfile = devpredfile;
		this.devoutfile = devoutfile;
		this.testfile = testfile;
		this.testpredfile = testpredfile;
		this.testoutfile = testoutfile;
		this.verbal = verbal;
		this.currID = new Integer[2][1];
		
		t = new Thread(this, "dev");
		t.start();
	}
	
	private void work(String outfile, String goldfile, String predfile, Parameters params) throws Exception {
		BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outfile),"UTF8"));

		BufferedReader in_g = new BufferedReader(new InputStreamReader(new FileInputStream(goldfile),"UTF8"));
		BufferedReader in_p = null;
		if (DependencyParser.testingMode <= 3)
			in_p = new BufferedReader(new InputStreamReader(new FileInputStream(predfile),"UTF8"));
		else if (DependencyParser.testingMode == 4)
			in_p = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(predfile + ".gz")),"UTF8"));

		// score file
		ObjectInputStream in_s = null;
		if (DependencyParser.pruningMode == 3)
			in_s = new ObjectInputStream(new GZIPInputStream(new FileInputStream(predfile + ".score.gz")));

		BufferedReader in_c = null;
		if (DependencyParser.pruningMode == 4) {
			// candidate in a separate file
			String candfile = predfile.substring(0, predfile.lastIndexOf(".")) + "." + DependencyParser.candExt;
			in_c = new BufferedReader(new InputStreamReader(new FileInputStream(candfile),"UTF8"));
		}
		
		currID[0][0] = 0;
		currID[1][0] = 0;
		
		Integer[] stats = new Integer[4];
		stats[0] = stats[1] = stats[2] = stats[3] = 0; // 0: token; 1: candNum; 2:oracleCorr; 3: pos acc
														// 0: cert
		
		Double[] score = new Double[3];			// store score, time, token;
		score[0] = score[1] = score[2] = 0.0;
		
		OutputThread ot = null;
		if (DependencyParser.testingMode == 0 || DependencyParser.testingMode == 1 || 
				DependencyParser.testingMode == 2 || DependencyParser.testingMode == 4) {
			// build output thread
			ot = new OutputThread(out, currID[1], dp, params, stats);
		}
		
		DecodingThread[] dt = new DecodingThread[DependencyParser.devThread];
		for (int i = 0; i < dt.length; i++) {
			dt[i] = new DecodingThread(dp, pipe, decoder, params,
						in_g, in_p, in_s, in_c, out, stats, score, currID, ot, verbal);
		}
		for (int i = 0; i < dt.length; i++) {
			dt[i].join();
		}
		if (ot != null)
			ot.join();
		
		if (DependencyParser.testingMode == 0) {
			System.out.println();
			if (DependencyParser.correctPos) {
				System.out.println("Pos Acc: " + (double)stats[1] / stats[0]);
			}
			if (DependencyParser.outputScore) {
				System.out.println("No worse than gold: " + (double)stats[1] / stats[0]);
			}
			System.out.println("Total score: " + score[0]);
			System.out.println("Token per sec: " + "(" + score[2] + "/" + (score[1] / 1000) + "=)" + score[2] / (score[1] / 1000));
		}
		if (verbal) {
			System.out.println();
			
			if (DependencyParser.testingMode == 1) {
				System.out.println("Oracle Acc: " + (double)stats[2] / stats[0]);
				System.out.println("Avg. head: " + (double)stats[1] / stats[0]);
				if (DependencyParser.correctPos) {
					System.out.println("POS Acc: " + (double)stats[3] / stats[0]);
				}
			}
			else if (DependencyParser.testingMode == 2) {
				System.out.println("Cert: " + stats[1]);
				System.out.println("Sent: " + stats[0]);
				System.out.println("Rate: " + (double)stats[1] / stats[0]);
			}
		}
		
		if (DependencyParser.testingMode == 4) {
			System.out.println("Oracle Acc: " + (double)stats[2] / stats[0]);
		}

		out.close();
		in_g.close();
		in_p.close();
		if (DependencyParser.pruningMode == 3)
			in_s.close();
		
		if (DependencyParser.pruningMode == 4)
			in_c.close();
	}
	
	public void run() {
		try {
			long start = System.currentTimeMillis();
			
			if (testfile != null) {
	    		//dp.outputParses(testfile, testpredfile, testoutfile, params, DependencyParser.devSplit, id, verbal);
				if (verbal)
					System.out.print("Processing Test Sentence: ");
				Parameters params = dp.getTestParams();
				work(testoutfile, testfile, testpredfile, params);
			}
			if (devfile != null) {
	    		//dp.outputParses(devfile, devpredfile, devoutfile, params, DependencyParser.devSplit, id, verbal);
				if (verbal)
					System.out.print("Processing Dev Sentence: ");
				Parameters params = dp.getDevParams();
				work(devoutfile, devfile, devpredfile, params);
			}
    		long end = System.currentTimeMillis();
    		
    		if (verbal) {
    			System.out.println("Testing took: " + (end-start));
    		}

    		dp.isDevTesting = false;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
