package mstparser;

//import gnu.trove.TObjectIntHashMap;

import java.io.*;

import gnu.trove.map.hash.TLongIntHashMap;
import gnu.trove.map.hash.TObjectIntHashMap;

import mstparser.FeatureTemplate.*;


public class FeatureAlphabet implements Serializable
{
	// finally we will remove strmap
	TObjectIntHashMap<String> strmap;
	public TLongIntHashMap arcMSTMap;
	public TLongIntHashMap arcTurboMap;
	public TLongIntHashMap secondOrderMap;
	public TLongIntHashMap thirdOrderMap;
	public TLongIntHashMap highOrderMap;
	public TLongIntHashMap catibMap;

	int numEntries;
	boolean growthStopped = false;

	public FeatureAlphabet (int capacity)
	{
		this.strmap = new TObjectIntHashMap<String> (capacity);
		this.arcMSTMap = new TLongIntHashMap (capacity);
		this.arcTurboMap = new TLongIntHashMap (capacity);
		this.secondOrderMap = new TLongIntHashMap (capacity);
		this.thirdOrderMap = new TLongIntHashMap (capacity);
		this.highOrderMap = new TLongIntHashMap (capacity);
		this.catibMap = new TLongIntHashMap (capacity);
		// trove current verion start from 1
		numEntries = 1;
	}

	public FeatureAlphabet ()
	{
		this (10000);
	}

	/** Return -1 (in old trove version) or 0 (in trove current verion) if entry isn't present. */
	public int lookupIndex (String entry, boolean addIfNotPresent)
	{
		if (entry == null)
			throw new IllegalArgumentException ("Can't lookup \"null\" in an Alphabet.");
		int ret = strmap.get(entry);
		if (ret <= 0 && !growthStopped && addIfNotPresent) {
			ret = numEntries;
			numEntries++;
			strmap.put (entry, ret);
		}
		return ret;
	}
	
	public TLongIntHashMap getMap(TemplateType type) {
		TLongIntHashMap intmap = null;
		switch (type) {
		case TArcMST:
			intmap = arcMSTMap;
			break;
		case TArcTurbo:
			intmap = arcTurboMap;
			break;
		case TSecondOrder:
			intmap = secondOrderMap;
			break;
		case TThirdOrder:
			intmap = thirdOrderMap;
			break;
		case THighOrder:
			intmap = highOrderMap;
			break;
		case TCatib:
			intmap = catibMap;
			break;
		default:
			System.out.println("undefined template type");
			System.exit(0);
		}
		return intmap;
	}

	/** Return -1 (in old trove version) or 0 (in trove current verion) if entry isn't present. */
	public int lookupIndex (TemplateType type, long entry, boolean addIfNotPresent)
	{
		TLongIntHashMap intmap = getMap(type);
		
		int ret = intmap.get(entry);
		if (ret <= 0 && !growthStopped && addIfNotPresent) {
			ret = numEntries;
			numEntries++;
			intmap.put (entry, ret);
		}
		return ret;
	}

	public int lookupIndex (TLongIntHashMap intmap, long entry, boolean addIfNotPresent)
	{
		int ret = intmap.get(entry);
		if (ret <= 0 && !growthStopped && addIfNotPresent) {
			ret = numEntries;
			numEntries++;
			intmap.put (entry, ret);
		}
		return ret;
	}

	public int lookupIndex (String entry)
	{
		return lookupIndex (entry, true);
	}

	public int lookupIndex (TemplateType type, long entry)
	{
		return lookupIndex (type, entry, true);
	}

	public int size ()
	{
		return numEntries;
	}

	public void stopGrowth ()
	{
		growthStopped = true;
	}

	public void allowGrowth ()
	{
		growthStopped = false;
	}

	public boolean growthStopped ()
	{
		return growthStopped;
	}

	// Serialization 

	private static final long serialVersionUID = 1;
	private static final int CURRENT_SERIAL_VERSION = 0;

	private void writeObject (ObjectOutputStream out) throws IOException {
		out.writeInt (CURRENT_SERIAL_VERSION);
		out.writeInt (numEntries);
		out.writeObject(strmap);
		out.writeObject(arcMSTMap);
		out.writeObject(arcTurboMap);
		out.writeObject(secondOrderMap);
		out.writeObject(thirdOrderMap);
		out.writeObject(highOrderMap);
		out.writeObject(catibMap);
		out.writeBoolean (growthStopped);
	}

	private void readObject (ObjectInputStream in) throws IOException, ClassNotFoundException {
		int version = in.readInt ();
		numEntries = in.readInt();
		strmap = (TObjectIntHashMap<String>)in.readObject();
		arcMSTMap= (TLongIntHashMap)in.readObject();
		arcTurboMap= (TLongIntHashMap)in.readObject();
		secondOrderMap= (TLongIntHashMap)in.readObject();
		thirdOrderMap= (TLongIntHashMap)in.readObject();
		highOrderMap= (TLongIntHashMap)in.readObject();
		catibMap= (TLongIntHashMap)in.readObject();
		growthStopped = in.readBoolean();
	}

}
