package mstparser;

public class FeatureTemplate {
	
	/******************************
	 * template type
	 *****************************/
	
	public enum TemplateType {
		TArcMST, TArcTurbo, TSecondOrder, TThirdOrder, THighOrder, TCatib, COUNT,
	}
	
	/*******************************
	 *  offset
	 ******************************/
	
	public static int largeOff = 17;	// word, lemma
	public static int midOff = 9; 		// pos, cpos, type
	public static int moridOff = 3;		// morphology id
	public static int morvalOff = 3;		// morphology val
	public static int flagOff = 4;		// flag, children num, length diff etc.
	public static int tempOff = 7;		// template
	
	/**********************************
	 * notation
	 * P: pos; C: coarse pos; W: word; L: lemma
	 **********************************/
	public enum ArcMST {
		
		START,
	    
		/*************************************************
		 * First-order dependency feature from MST
		 * ***********************************************/
	    // feature posL posIn posR
	    LP_MP_RP,
	    LC_MC_RC,
	    
	    // feature posL-1 posL posR posR+1
	    pLP_LP_RP_nRP,
	    LP_RP_nRP,
	    pLP_RP_nRP,
	    pLP_LP_nRP,
	    pLP_LP_RP,
	    
	    pLC_LC_RC_nRC,
	    LC_RC_nRC,
	    pLC_RC_nRC,
	    pLC_LC_nRC,
	    pLC_LC_RC,
	    
	    // feature posL posL+1 posR-1 posR
	    LP_nLP_pRP_RP,
	    nLP_pRP_RP,
	    LP_pRP_RP,
	    LP_nLP_RP,
	    LP_nLP_pRP,

	    LC_nLC_pRC_RC,
	    nLC_pRC_RC,
	    LC_pRC_RC,
	    LC_nLC_RC,
	    LC_nLC_pRC,
	    
	    // feature posL-1 posL posR-1 posR
	    // feature posL posL+1 posR posR+1
	    pLP_LP_pRP_RP,
	    pLC_LC_pRC_RC,
	    LP_nLP_RP_nRP,
	    LC_nLC_RC_nRC,

	    // two obs (word, pos)
	    HP,
	    HW,
	    //MP,
	    //MW,
	    HW_HP,
	    //MW_MP,
	    HP_MP,
	    HP_MW,
	    HW_MP,
	    HW_MW,
	    HP_MP_MW,
	    HW_MP_MW,
	    HP_HW_MP,
	    HP_HW_MW,
	    HP_HW_MP_MW,
	    
	    // lemma pos
	    HL,
	    //ML,
	    HP_ML,
	    HL_MP,
	    HL_ML,
	    HP_HL,
	    //MP_ML,
	    HP_MP_ML,
	    HL_MP_ML,
	    HP_HL_MP,
	    HP_HL_ML,
	    HP_HL_MP_ML,
	    
	    // morphology, id, val, pos
	    FF_IDH_IDM_HV,
	    FF_IDH_IDM_MV,
	    FF_IDH_IDM_HV_MV,
	    FF_IDH_IDM_HP_MV,
	    FF_IDH_IDM_HV_MP,
	    FF_IDH_IDM_HV_HP,
	    FF_IDH_IDM_MV_MP,
	    FF_IDH_IDM_HP_MP_MV,
	    FF_IDH_IDM_HV_MP_MV,
	    FF_IDH_IDM_HP_HV_MP,
	    FF_IDH_IDM_HP_HV_MV,
	    FF_IDH_IDM_HP_HV_MP_MV,
	    
	    // label
	    LAB,
	    LAB_W_WP,
	    LAB_WP,
	    LAB_pWP_WP,
	    LAB_WP_nWP,
	    LAB_pWP_WP_nWP,
	    LAB_W,
	    LAB_L_WP,
	    LAB_L,
	    LAB_HP_MP,
	    LAB_HL_MP,
	    LAB_HP_ML,
	    LAB_HL_ML,
	    
	    // pred, score
	    PRED,
	    PRED_HC_MC,
	    PREDL,
	    PREDL_LAB_HC_MC,
	    SCORE1,
	    SCORE2,
	    
	    // contextual
	    pHW,
	    nHW,
	    pMW,
	    MW,
	    nMW,
	    
	    COUNT;
	}

	public enum ArcTurbo {
		
		START,
	    
		/*************************************************
		 * First-order dependency feature from Turbo parser
		 * ***********************************************/
		pHC_HL_MC,
		HL_nHC_MC,
		HL_pMC_MC,
		HL_MC_nMC,
		pHC_HL_MC_nMC,
		HL_nHC_MC_nMC,
		pHC_HL_pMC_MC,
		HL_nHC_pMC_MC,
		
		pHC_HC_ML,
		HC_nHC_ML,
		HC_pMC_ML,
		HC_ML_nMC,
		pHC_HC_ML_nMC,
		HC_nHC_ML_nMC,
		pHC_HC_pMC_ML,
		HC_nHC_pMC_ML,
		
		pHC_HL_ML,
		HL_nHC_ML,
		HL_pMC_ML,
		HL_ML_nMC,
		pHC_HL_ML_nMC,
		HL_nHC_ML_nMC,
		pHC_HL_pMC_ML,
		HL_nHC_pMC_ML,
		
	    
	    // flag
	    HC_MC_FLAG,
	    HL_ML_FLAG,

	    COUNT,
	}
	
	public enum SecondOrder {
		START,
		
		HP_SP_MP,
		HC_SC_MC,
		
		pHC_HC_SC_MC,
		HC_nHC_SC_MC,
		HC_pSC_SC_MC,
		HC_SC_nSC_MC,
		HC_SC_pMC_MC,
		HC_SC_MC_nMC,
		
		pHC_HL_SC_MC,
		HL_nHC_SC_MC,
		HL_pSC_SC_MC,
		HL_SC_nSC_MC,
		HL_SC_pMC_MC,
		HL_SC_MC_nMC,
		
		pHC_HC_SL_MC,
		HC_nHC_SL_MC,
		HC_pSC_SL_MC,
		HC_SL_nSC_MC,
		HC_SL_pMC_MC,
		HC_SL_MC_nMC,
		
		pHC_HC_SC_ML,
		HC_nHC_SC_ML,
		HC_pSC_SC_ML,
		HC_SC_nSC_ML,
		HC_SC_pMC_ML,
		HC_SC_ML_nMC,
		
	    HC_MC_SC_pHC_pMC, 			
	    HC_MC_SC_pHC_pSC, 			
	    HC_MC_SC_pMC_pSC, 			
	    HC_MC_SC_nHC_nMC, 			
	    HC_MC_SC_nHC_nSC, 			
	    HC_MC_SC_nMC_nSC, 		
	    HC_MC_SC_pHC_nMC, 			
	    HC_MC_SC_pHC_nSC, 	
	    HC_MC_SC_pMC_nSC, 			
	    HC_MC_SC_nHC_pMC, 			
	    HC_MC_SC_nHC_pSC, 			
	    HC_MC_SC_nMC_pSC, 			

	    SP_MP,
		SW_MW,
		SW_MP,
		SP_MW,
		SC_MC,
		SL_ML,
		SL_MC,
		SC_ML,
		
		SCORE3,
		SCORE4,
		
		// head bigram
		H1P_H2P_M1P_M2P,
		H1P_H2P_M1P_M2P_DIR,
		H1C_H2C_M1C_M2C,
		H1C_H2C_M1C_M2C_DIR,
		
		// gp-p-c
		GP_HP_MP,
		GC_HC_MC,
		GL_HC_MC,
		GC_HL_MC,
		GC_HC_ML,
		
		pGC_GC_HC_MC,
		GC_nGC_HC_MC,
		GC_pHC_HC_MC,
		GC_HC_nHC_MC,
		GC_HC_pMC_MC,
		GC_HC_MC_nMC,
		
		pGC_GL_HC_MC,
		GL_nGC_HC_MC,
		GL_pHC_HC_MC,
		GL_HC_nHC_MC,
		GL_HC_pMC_MC,
		GL_HC_MC_nMC,
		
		pGC_GC_HL_MC,
		GC_nGC_HL_MC,
		GC_pHC_HL_MC,
		GC_HL_nHC_MC,
		GC_HL_pMC_MC,
		GC_HL_MC_nMC,
		
		pGC_GC_HC_ML,
		GC_nGC_HC_ML,
		GC_pHC_HC_ML,
		GC_HC_nHC_ML,
		GC_HC_pMC_ML,
		GC_HC_ML_nMC,
		
	    GC_HC_MC_pGC_pHC, 			
	    GC_HC_MC_pGC_pMC, 			
	    GC_HC_MC_pHC_pMC, 			
	    GC_HC_MC_nGC_nHC, 			
	    GC_HC_MC_nGC_nMC, 			
	    GC_HC_MC_nHC_nMC, 		
	    GC_HC_MC_pGC_nHC, 		
	    GC_HC_MC_pGC_nMC, 			
	    GC_HC_MC_pHC_nMC, 			
	    GC_HC_MC_nGC_pHC, 			
	    GC_HC_MC_nGC_pMC, 			
	    GC_HC_MC_nHC_pMC, 			

	    COUNT,
	}
	
	public enum ThirdOrder {
		START,
		
		// move some gpc features here...
	    GL_HC_MC_pGC, 				
	    GL_HC_MC_nGC, 			
	    GL_HC_MC_pHC, 			
	    GL_HC_MC_nHC, 				
	    GL_HC_MC_pMC, 			
	    GL_HC_MC_nMC, 		
	    GC_HL_MC_pGC, 			
	    GC_HL_MC_nGC, 			
	    GC_HL_MC_pHC, 			
	    GC_HL_MC_nHC, 				
	    GC_HL_MC_pMC, 				
	    GC_HL_MC_nMC, 				
	    GC_HC_ML_pGC, 			
	    GC_HC_ML_nGC, 				
	    GC_HC_ML_pHC, 	
	    GC_HC_ML_nHC, 			
	    GC_HC_ML_pMC, 		
	    GC_HC_ML_nMC, 		
		
	    GL_HL_MC,				
	    GL_HC_ML,				
	    GC_HL_ML,				
	    GL_HL_ML,			

	    GC_HC,				
	    GL_HC,			
	    GC_HL,			
	    GL_HL,	
	    
	    // only cross with dir flag
	    GC_MC,			
	    GL_MC,				
	    GC_ML,				
	    GL_ML,				
	    HC_MC,				
	    HL_MC,				
	    HC_ML,				
	    HL_ML,				

		// ggpc
		GGC_GC_HC_MC,
		GGL_GC_HC_MC,
		GGC_GL_HC_MC,
		GGC_GC_HL_MC,
		GGC_GC_HC_ML,
		
		GGC_HC_MC,
		GGL_HC_MC,
		GGC_HL_MC,
		GGC_HC_ML,
		GGC_GC_MC,
		GGL_GC_MC,
		GGC_GL_MC,
		GGC_GC_ML,
		GGC_MC,
		GGL_MC,
		GGC_ML,
		GGL_ML,
		
		HC_MC_CC_SC,
		HL_MC_CC_SC,
		HC_ML_CC_SC,
		HC_MC_CL_SC,
		HC_MC_CC_SL,
		
		HC_CC_SC,
		HL_CC_SC,
		HC_CL_SC,
		HC_CC_SL,
		
		// gp sibling
		GC_HC_MC_SC,
		GL_HC_MC_SC,
		GC_HL_MC_SC,
		GC_HC_ML_SC,
		GC_HC_MC_SL,
		
		// tri-sibling
		HC_PC_MC_NC,
		HL_PC_MC_NC,
		HC_PL_MC_NC,
		HC_PC_ML_NC,
		HC_PC_MC_NL,
		
	    HC_PC_NC,			
	    PC_MC_NC,		
	    HL_PC_NC,		
	    HC_PL_NC,		
	    HC_PC_NL,		
	    PL_MC_NC,		
	    PC_ML_NC,			
	    PC_MC_NL,			
		
	    PC_NC,
	    PL_NC,
	    PC_NL,
	    
		COUNT,
	}
	
	public enum HighOrder {
		START,
		
		// pp attachment
		PP_HC_MC,
		PP_HL_MC,
		PP_HC_ML,
		PP_HL_ML,
		
		PP_PL_HC_MC,
		PP_PL_HL_MC,
		PP_PL_HC_ML,
		PP_PL_HL_ML,
		
		// conjunction
		CC_CP_LP_RP,
		CC_CP_LC_RC,
		CC_CW_LP_RP,
		CC_CW_LC_RC,
		
		CC_LC_RC_FID,
		
		CC_CP_HC_AC,
		CC_CP_HL_AL,
		CC_CW_HC_AC,
		CC_CW_HL_AL,
		
		CC_LP_RP_LENDIFF,
		CC_LC_RC_LENDIFF,
		CC_LENDIFF,
		
		CC_LP_RP_CHILDF,
		CC_LC_RC_CHILDF,
		
		// PNX
		PNX_MW,
		PNX_HP_MW,
		
		// right branch
		RB,
		
		// child num
		CN_HP_NUM,
		CN_HP_LNUM_RNUM,
		
		// heavy
		HV_HP,
		HV_HC,
		
		// neighbor
		NB_HP_LC_RC,
		NB_HC_LC_RC,
		NB_HL_LC_RC,
		
		// non-proj
		NP,
		NP_MC,
		NP_HC,
		NP_HL,
		NP_ML,
		NP_HC_MC,
		NP_HL_MC,
		NP_HC_ML,
		NP_HL_ML,
		
	    // pos tagging features
	    ppP_P,
	    pP_P,
	    P_nP,		// duplicated...
	    P_nnP,

	    ppP_pP_P,
	    pP_P_nP,
	    P_nP_nnP,
	    ppP_P_nP,
	    pP_P_nnP,
	    ppP_P_nnP,
	    
	    ppP_pP_P_nP,
	    ppP_pP_P_nnP,
	    pP_P_nP_nnP,
	    ppP_P_nP_nnP,
	    
	    ppP_pP_P_nP_nnP,

	    ppL_P,
	    pL_P,
	    L_P,
	    P_nL,
	    P_nnL,
	    pP_L_P,
	    L_P_nP,
	    pP_L_P_nP,
	    ppP_pP_L_P,
	    L_P_nP_nnP,
	    
		COUNT,
	}
	
	public enum CatibFeature {
		START,

		// first order
		HD_BD_MD,

		pHD_HD_MD_nMD,
		HD_MD_nMD,
		pHD_MD_nMD,
		pHD_HD_nMD,
		pHD_HD_MD,

		HD_nHD_pMD_MD,
		nHD_pMD_MD,
		HD_pMD_MD,
		HD_nHD_MD,
		HD_nHD_pMD,

		pHD_HD_pMD_MD,
		HD_nHD_MD_nMD,

		HD,
		HD_MD,

		NA_HGEN_HNUM_HRAT_MGEN_MNUM,
		NA_AGREE,

		LAB_WD,
		LAB_pWD_WD,
		LAB_WD_nWD,
		LAB_pWD_WD_nWD,
		LAB_HD_MD,

		SBJ_VS1,
		SBJ_VS_AGREE,
		SBJ_SV_AGREE,
		SBJ_HGEN_HNUM_MRAT_MGEN_MNUM,

		// second order
		HD_SD_MD,
		SD_MD,
		GD_HD_MD,

		// global
		ADJ_DT_AGREE,
		IDF_AGREE,
		IDF_DIR,
		IDF_NUM,

		// temporary for arbitrary sibling
		HC_SC_MC,
		
		pHC_HC_SC_MC,
		HC_nHC_SC_MC,
		HC_pSC_SC_MC,
		HC_SC_nSC_MC,
		HC_SC_pMC_MC,
		HC_SC_MC_nMC,
		
		pHC_HL_SC_MC,
		HL_nHC_SC_MC,
		HL_pSC_SC_MC,
		HL_SC_nSC_MC,
		HL_SC_pMC_MC,
		HL_SC_MC_nMC,
		
		pHC_HC_SL_MC,
		HC_nHC_SL_MC,
		HC_pSC_SL_MC,
		HC_SL_nSC_MC,
		HC_SL_pMC_MC,
		HC_SL_MC_nMC,
		
		pHC_HC_SC_ML,
		HC_nHC_SC_ML,
		HC_pSC_SC_ML,
		HC_SC_nSC_ML,
		HC_SC_pMC_ML,
		HC_SC_ML_nMC,
		
	    HC_MC_SC_pHC_pMC, 			
	    HC_MC_SC_pHC_pSC, 			
	    HC_MC_SC_pMC_pSC, 			
	    HC_MC_SC_nHC_nMC, 			
	    HC_MC_SC_nHC_nSC, 			
	    HC_MC_SC_nMC_nSC, 		
	    HC_MC_SC_pHC_nMC, 			
	    HC_MC_SC_pHC_nSC, 	
	    HC_MC_SC_pMC_nSC, 			
	    HC_MC_SC_nHC_pMC, 			
	    HC_MC_SC_nHC_pSC, 			
	    HC_MC_SC_nMC_pSC, 			

		COUNT
	}
	
	/*********************************
     * code generator
     * generally flag will be added lately, because code without flag is also needed
     */
	
    public static int getBits(long x) {
        long y = 1;
        int i = 0;
        while (y < x) {
                y = y << 1;
                ++i;
        }
        return i;
    }
    
    public static long genCodePF(long temp, long p1) {
    	return ((p1 << flagOff) << tempOff) | temp;
    }

    public static long genCodePPF(long temp, long p1, long p2) {
    	return ((((p1 << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    public static long genCodePP(long temp, long p1, long p2) {
    	return (((p1 << midOff) | p2) << tempOff) | temp;
    }

    public static long genCodePPPF(long temp, long p1, long p2, long p3) {
    	return ((((((p1 << midOff) | p2) << midOff) | p3) << flagOff) << tempOff) | temp;
    }
	
    public static long genCodePPP(long temp, long p1, long p2, long p3) {
    	return (((((p1 << midOff) | p2) << midOff) | p3) << tempOff) | temp;
    }
	
    public static long genCodePPW(long temp, long p1, long p2, long w1) {
    	return (((((w1 << midOff) | p1) << midOff) | p2) << tempOff) | temp;
    }
	
    public static long genCodePPPPF(long temp, long p1, long p2, long p3, long p4) {
    	return ((((((((p1 << midOff) | p2) << midOff) | p3) << midOff) | p4) << flagOff) << tempOff) | temp;
    }
	
    public static long genCodePPPPPF(long temp, long p1, long p2, long p3, long p4, long p5) {
    	return ((((((((((p1 << midOff) | p2) << midOff) | p3) << midOff) | p4) << midOff) << p5) << flagOff) << tempOff) | temp;
    }
	
    public static long genCodeWF(long temp, long w1) {
    	return ((w1 << flagOff) << tempOff) | temp;
    }

    public static long genCodePWF(long temp, long p1, long w1) {
    	return ((((w1 << midOff) | p1) << flagOff) << tempOff) | temp;
    }

    public static long genCodeWWF(long temp, long w1, long w2) {
    	return ((((w1 << largeOff) | w2) << flagOff) << tempOff) | temp;
    }

    public static long genCodeWWW(long temp, long w1, long w2, long w3) {
    	return (((((w1 << largeOff) | w2) << largeOff) | w3) << tempOff) | temp;
    }

    public static long genCodePPWF(long temp, long p1, long p2, long w1) {
    	return ((((((w1 << midOff) | p1) << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    public static long genCodePPPWF(long temp, long p1, long p2, int p3, long w1) {
    	return ((((((((w1 << midOff) | p1) << midOff) | p2) << midOff) | p3) << flagOff) << tempOff) | temp;
    }

    public static long genCodePWWF(long temp, long p1, long w1, long w2) {
    	return ((((((w1 << largeOff) | w2) << midOff) | p1) << flagOff) << tempOff) | temp;
    }

    public static long genCodePPWWF(long temp, long p1, long p2, long w1, long w2) {
    	return ((((((((w1 << largeOff) | w2) << midOff) | p1) << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIPPF(long temp, long i1, long p1, long p2) {
    	return ((((((i1 << midOff) | p1) << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVF(long temp, long i1, long i2, long v1) {
    	return ((((((i1 << moridOff) | i2) << morvalOff) | v1) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVVF(long temp, long i1, long i2, long v1, long v2) {
    	return ((((((((i1 << moridOff) | i2) << morvalOff) | v1) << morvalOff) | v2) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVVPF(long temp, long i1, long i2, long v1, long v2, long p1) {
    	return ((((((((((i1 << moridOff) | i2) << morvalOff) | v1) << morvalOff) | v2) << midOff) | p1) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVPF(long temp, long i1, long i2, long v1, long p1) {
    	return ((((((((i1 << moridOff) | i2) << morvalOff) | v1) << midOff) | p1) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVPPF(long temp, long i1, long i2, long v1, long p1, long p2) {
    	return ((((((((((i1 << moridOff) | i2) << morvalOff) | v1) << midOff) | p1) << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    public static long genCodeIIVVPPF(long temp, long i1, long i2, long v1, long v2, long p1, long p2) {
    	return ((((((((((((i1 << moridOff) | i2) << morvalOff) | v1) << morvalOff) | v2)
    			<< midOff) | p1) << midOff) | p2) << flagOff) << tempOff) | temp;
    }

    // more than one flag, add directly
    public static long genCodePFF(long temp, long p1, long f1, long f2) {
    	return (((((p1 << flagOff) | f1) << flagOff) | f2) << tempOff) | temp;
    }

    public static long genCodeV(long temp, long v1) {
    	return (v1 << tempOff) | temp;
    }
  
    public static long genCodeVVVVV(long temp, long v1, long v2, long v3, long v4, long v5) {
    	return (((((((((v1 << morvalOff) | v2) << morvalOff) | v3) << morvalOff) | v4)
    				<< morvalOff) | v5) << tempOff) | temp;
    }

}
