package mstparser;

import java.io.*;
import java.util.ArrayList;
import static mstparser.DependencyPipe2O.NOTYPE;;

public class LocalFeatureData {
    public FeatureVector[][] fvs;
    public double[][] probs;
    public boolean[][] cand;
    
    // not memory efficient here
    
    public FeatureVector[][][] fvs_trips;		
    public double[][][] probs_trips;
    public boolean[][][] cand_trips;
    
    public FeatureVector[][][] fvs_sibs;		// could be more efficient
    public double[][][] probs_sibs;
    public boolean[][][] cand_sibs;
    
    public FeatureVector[][][][] fvs_lab;
    public double[][][][] probs_lab;
    
    public FeatureVector[][][] fvs_dep;			// could be more efficient
    public double[][][] probs_dep;
    public boolean[][][] cand_dep;
   
    public FeatureVector[][][] fvs_gpc;			// could be more efficient
    public double[][][] probs_gpc;
    
    public double[][][] probs_ppcc1;		// pp attachment, punc head and part of conjunction
    public int[] puncPair;					// punc counterparts. -1 if not exist
    public FeatureVector[][][] fvs_ppcc1;	
    
    public double[][][] probs_cc2;		// arg, head, left/right
    public FeatureVector[][][] fvs_cc2;
    
    // try to memory efficient here

    public int[][] dep2id;					// map a possible dep to an id
    public int maxid;
    public double[][][] probs_gpsib;		// grandparent-parent-child-sibling, gp-p is mapped to id [maxid][len][len]
    public FeatureVector[][][] fvs_gpsib;	
    
    public double[][][] probs_trisib;		// parent-sibling-child-sibling [maxid][len][len]
    public FeatureVector[][][] fvs_trisib;
    
    public double[][][] probs_dep2;		// [maxid][lab][lab], cand_dep+cand_lab
    public FeatureVector[][][] fvs_dep2;

    public double[][] probs_arbsib;		//[maxid][len], cand
    public FeatureVector[][] fvs_arbsib;
    
    public double[][] probs_headbi;		// [maxid][len], cand
    public FeatureVector[][] fvs_headbi;
    
    public double[][] probs_ggpc;		// [maxid][maxid], cand+cand_gpc
    public FeatureVector[][] fvs_ggpc;
    
    public double[][] probs_psc;		// parent-sib-mod-child, [maxid][maxid] cand+cand_gpc
    public FeatureVector[][] fvs_psc;
    
    public double[][] probs_nonproj;	// nonproj arc, [maxid][nonproj binned num]
    public FeatureVector[][] fvs_nonproj;
    
    public int len;
    public int type;
    
    public LocalFeatureData(int len, int type, boolean createScoreArray) {
    	this.len = len;
    	this.type = type;
    	
    	cand = new boolean[len][len];
    	cand_trips = new boolean[len][len][len];
    	cand_sibs = new boolean[len][len][2];
    	cand_dep = new boolean[len][len][type];
    	dep2id = new int[len][len];
    	
    	puncPair = new int[len];
    	
    	if (createScoreArray) {
        	fvs = new FeatureVector[len][len];
        	fvs_lab = new FeatureVector[len][type][2][2];
        	fvs_dep = new FeatureVector[len][len][type];
        	
        	probs = new double[len][len];
        	probs_lab = new double[len][type][2][2];
        	probs_dep = new double[len][len][type];

        	if (DependencyParser.useSecondOrder) {
        		fvs_trips = new FeatureVector[len][len][len];
        		fvs_sibs = new FeatureVector[len][len][2];
        		fvs_gpc = new FeatureVector[len][len][len];

        		probs_trips = new double[len][len][len];
            	probs_sibs = new double[len][len][2];
            	probs_gpc = new double[len][len][len];
        	}
        	
        	if (DependencyParser.useHO) {
	        	fvs_ppcc1 = new FeatureVector[len][len][len];
	        	fvs_cc2 = new FeatureVector[len][len][len];
	    		
	        	probs_ppcc1 = new double[len][len][len];
	        	probs_cc2 = new double[len][len][len];
        	}        	
        	// memory efficient list can only be built latter
    	}
    }
    
    public void loadCandidate(BufferedReader br, DependencyPipe2O pipe, 
    		DependencyInstance pred, DependencyInstance gold, boolean createScoreArray) throws IOException {
    	if (DependencyParser.pruningMode != 0) {
	    	int[] dep = pred.dep;
	    	int[] lab = pred.labs;
	    	
	    	// also add parser prediction into the candidate?
	    	for (int i = 1; i < len; i++) {
	    		int type = lab[i];
	    		if (type <= 0) {
	    			//System.out.println("unseen type");
	    			type = NOTYPE;
	    		}
	    		cand[dep[i]][i] = true;
	    		cand_dep[dep[i]][i][type] = true;
	    		
	    		// add gold if available
	    		if (DependencyParser.learningMode == 4 && gold != null) {
		    		cand[gold.dep[i]][i] = true;
		    		type = gold.labs[i];
		    		if (type <= 0)
		    			type = NOTYPE;
		    		cand_dep[gold.dep[i]][i][type] = true;
	    		}
	    		
	    		if (br != null) {
		    		String[] data = br.readLine().split("\t");
		    		for (int j = 0; j < data.length; j++) {
		    			String[] tmp = data[j].split("\\|");
		    			int h = Integer.parseInt(tmp[0]);
		    			cand[h][i] = true;
		    			if (pipe.labeled) {
			    			for (int k = 1; k < tmp.length; k++) {
			    				if (tmp[k].equals("<no-type>"))
			    					continue;
			    				int l = pipe.typeAlphabet.lookupIndex(tmp[k]);
			    				if (l > 0) {
			    					cand_dep[h][i][l] = true;
			    				}
			    			}
		    			}
		    			else {
		    				cand_dep[h][i][type] = true;
		    			}
		    		}
	    		}
	    		else {
	    			// generate candidate
	    			ArrayList<Integer> candList = generateCandidate(pred.dep, i, 4, -1);
	    			for (int z = 0; z < candList.size(); z++) {
	    				int h = candList.get(z);
	    				cand[h][i] = true;
	    				cand_dep[h][i][type] = true;
	    			}
	    		}
	    	}
    	}
    	else {
    		for (int i = 1; i < len; i++) {
    			for (int h = 0; h < len; h++) {
        			if (h != i) {
        				cand[h][i] = true;
        				if (DependencyParser.labeled) {
        					for (int t = NOTYPE + 1; t < type; t++) {
        						cand_dep[h][i][t] = true;
        					}
        				}
        				else {
        					cand_dep[h][i][NOTYPE] = true;
        				}
        			}
    			}
    		}
    	}
    	
    	// build id
    	int id = 0;
    	for (int i = 0; i < len; i++)
    		for (int j = 0; j < len; j++)
    			if (cand[i][j]) {
    				dep2id[i][j] = id;
    				id++;
    			}
    	maxid = id;
    	
    	// build second order
    	if (DependencyParser.useSecondOrder) {
	    	for (int i = 0; i < len; i++) {
	    		for (int prev = i; prev < len; prev++) {
	    			if (prev != i && !cand[i][prev])
	    				continue;
	        		for(int j = prev + 1; j < len; j++) {
	        			if(cand[i][j]) {
	        				cand_trips[i][prev][j] = true;
	        				cand_sibs[prev][j][prev == i ? 0 : 1] = true;
	        			}
	        		}
	    		}
	    		for (int prev = i; prev >= 0; prev--) {
	    			if (prev != i && !cand[i][prev])
	    				continue;
	        		for(int j = prev - 1; j >= 0; j--) {
	        			if(cand[i][j]) {
	        				cand_trips[i][prev][j] = true;
	        				cand_sibs[prev][j][prev == i ? 0 : 1] = true;
	        			}
	        		}
	    		}
	    	}
    	}
    	
    	if (createScoreArray) {
    		if (DependencyParser.useThirdOrder) {
		    	// build gpsib
		    	probs_gpsib = new double[maxid][len][len];
		    	fvs_gpsib = new FeatureVector[maxid][len][len];
		    	// build tripsib
		    	probs_trisib = new double[maxid][len][len];
		    	fvs_trisib = new FeatureVector[maxid][len][len];
    		}
    		if (DependencyParser.useDep2) {
		    	// build dep-lab-lab
		        probs_dep2 = new double[maxid][type][type];
		        fvs_dep2 = new FeatureVector[maxid][type][type];
    		}
	        // build arbitrary sibling
    		if (DependencyParser.useAS) {
	        	probs_arbsib = new double[maxid][len];		
	        	fvs_arbsib = new FeatureVector[maxid][len];
    		}
	        // build head bigram
    		if (DependencyParser.useHB) {
    			probs_headbi = new double[maxid][len];		
    			fvs_headbi = new FeatureVector[maxid][len];
    		}
    		if (DependencyParser.useGGPC) {
		        // build a->b->c->d
		        probs_ggpc = new double[maxid][maxid];		
		        fvs_ggpc = new FeatureVector[maxid][maxid];
    		}
    		if (DependencyParser.usePSC) {
		        // build parent-sibling-mod-child
		        probs_psc = new double[maxid][maxid];		
		        fvs_psc = new FeatureVector[maxid][maxid];
    		}
    		if (DependencyParser.useNonproj) {
    			// build nonproj
    			probs_nonproj = new double[maxid][8];
    			fvs_nonproj = new FeatureVector[maxid][8];
    		}
    	}
    	
    	if (DependencyParser.useHO) {
	    	for (int i = 0; i < len; i++) {
				if (pred.specialPos[i] == SpecialPos.PNX) {
					// head-punc-punc
					int j = pipe.findPuncCounterpart(pred.sentence, i);
					puncPair[i] = j;
				}
	    	}
    	}
    }
    
    public void loadParserPredScore(ObjectInputStream in, boolean hasInt) throws IOException {
    	if (hasInt)
    		in.readInt();
		for (int i = 0; i < len; i++)
			for (int j = i + 1; j < len; j++)
				for (int k = 0; k < 2; k++) {
					int head = k == 0 ? i : j;
					int child = k == 0 ? j : i;
					probs[head][child] = in.readDouble();
				}
		for (int i = 0; i < len; i++)
			for (int j = 0; j < type; j++)
				for (int k = 0; k < 2; k++)
					for (int l = 0; l < 2; l++)
						probs_lab[i][j][k][l] = in.readDouble();
		for (int i = 0; i < len; i++) {
			for (int j = i; j < len; j++)
				for (int k = j + 1; k < len; k++)
					probs_trips[i][j][k] = in.readDouble();
			for (int j = i; j >= 0; j--)
				for (int k = j - 1; k >= 0; k--)
					probs_trips[i][j][k] = in.readDouble();
		}
		for (int i = 0; i < len; i++)
			for (int j = 0; j < len; j++)
				for (int k = 0; k < 2; k++)
					if (i != j)
						probs_sibs[i][j][k] = in.readDouble();
		int next = in.readInt();
		if (next != -1) {
			System.out.println("load pred score bug");
			System.exit(0);
		}
    }
    
	private ArrayList<Integer> generateCandidate(int[] dep, int arg, int maxAncDist, int minDesDist) {
		int[] dist = getDistance(dep, arg);
		ArrayList<Integer> cand = new ArrayList<Integer>();
		for (int i = 0; i < dist.length; i++)
			if (dist[i] > 0 && dist[i] <= maxAncDist)
				cand.add(i);
			else if (dist[i] < 0 && dist[i] >= minDesDist)
				cand.add(i);
		return cand;
	}
	
	private int[] getDistance(int[] dep, int arg) {
		boolean[] v = new boolean[dep.length];
		v[arg] = true;
		int[] dist = new int[dep.length];
		getAncestorDistance(dep, dist, v, arg);
		getDescendantDistance(dep, dist, v, arg);
		for (int i = 0; i < dep.length; i++)
			if (!v[i])
				dist[i] = 10000;
		return dist;
	}

	public void getAncestorDistance(int[] dep, int[] dist, boolean[] v, int arg) {
		// parent
		if (dep[arg] != -1 && !v[dep[arg]]) {
			v[dep[arg]] = true;
			dist[dep[arg]] = dist[arg] + 1;
			getAncestorDistance(dep, dist, v, dep[arg]);
		}
		// sibling
		if (dist[arg] != 0) {
			for (int i = 0; i < dep.length; i++) {
				if (v[i] || dep[i] != arg)
					continue;
				v[i] = true;
				dist[i] = Math.min(10000, dist[arg] + 1);
				getAncestorDistance(dep, dist, v, i);
			}
		}
	}
	
	public void getDescendantDistance(int[] dep, int[] dist, boolean[] v, int arg) {
		// child
		for (int i = 0; i < dep.length; i++) {
			if (v[i] || dep[i] != arg)
				continue;
			v[i] = true;
			dist[i] = dist[arg] - 1;
			getDescendantDistance(dep, dist, v, i);
		}
	}
	
}
