package mstparser;
import java.io.BufferedReader;
import java.io.ObjectInputStream;
import java.util.*;

public class SampleRankThread implements Runnable  {
	private Object readLock;
	private Object paramsLock;
	
	private BufferedReader in_g;
	private BufferedReader in_p;
	
	private Parameters params;
	private DependencyPipe2O pipe;
	private DependencyDecoder2O decoder;
	private Integer[] currentID;
	private Integer[] stats;
	private DependencyInstance[] il;
	private int iter;
	//private double upd;
	
	private Thread t;
	
	public SampleRankThread(BufferedReader in_g, BufferedReader in_p,
			Parameters params, DependencyPipe2O pipe, DependencyDecoder2O decoder, 
			Integer[] currentID, Integer[] stats, DependencyInstance[] il, int iter,
			Object readLock, Object paramsLock, int id) {
		this.in_g = in_g;
		this.in_p = in_p;
		this.params = params;
		this.pipe = pipe;
		this.decoder = decoder;
		this.currentID = currentID;
		this.stats = stats;
		this.il = il;
		this.iter = iter;
		this.readLock = readLock;
		this.paramsLock = paramsLock;
		
		t = new Thread(this, "sample rank thread " + id);
		t.start();
	}
	
	public void join() {
		try {
			if (t.isAlive()) {
				t.join();
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
    public void gibbsSampling(DependencyInstance pred, int i, 
    		ArrayList<Integer> dep, ArrayList<Integer> lab, ArrayList<FeatureVector> fvList, 
    		LocalFeatureData lfd, GlobalFeatureData gfd, Parameters params, double temp) {
    	int len = pred.dep.length;
    	
		int oldDep = pred.dep[i];
		int oldLab = pred.labs[i];

		// build score
		for (int candH = 0; candH < len; candH++) {
			if (!lfd.cand[candH][i])
				continue;
			if (decoder.isDescendant(candH, i, pred.dep))		// loop
				continue;
			if (decoder.proj && !pipe.isProj(candH, i, pred.dep))
				continue;
			int oldH = pred.dep[i];
			pred.dep[i] = candH;
			pipe.updateChildList(pred, candH, oldH, i);
			for (int candT = 0; candT < lfd.type; candT++) {
				if (!lfd.cand_dep[candH][i][candT])
					continue;
				dep.add(candH);
				lab.add(candT);
				pred.labs[i] = candT;
				
				FeatureVector fv = pipe.createFeatureVector(pred, lfd, gfd, i);
				fvList.add(fv);
			}
		}
		
		int oldH = pred.dep[i];
		pred.dep[i] = oldDep;
		pred.labs[i] = oldLab;
		pipe.updateChildList(pred, oldDep, oldH, i);
    }

    public void sampleRank(DependencyInstance gold, DependencyInstance pred, 
			LocalFeatureData lfd, GlobalFeatureData gfd,
			Parameters params) {
		// sample rank training
		int len = pred.sentence.length;
		int maxIter = 1;
		double T = 1.0;

		for (int iter = 0; iter < maxIter; iter++) {
			for (int i = 1; i < len; i++) {
				ArrayList<Integer> dep = new ArrayList<Integer>();
				ArrayList<Integer> lab = new ArrayList<Integer>();
				ArrayList<FeatureVector> fvList = new ArrayList<FeatureVector>(); 
				ArrayList<Double> score = new ArrayList<Double>();

				gibbsSampling(pred, i, dep, lab, fvList, lfd, gfd, params, T);

				synchronized(paramsLock) {
					// sample and update params
					for (int z = 0; z < fvList.size(); z++) {
						double s = params.getScore(fvList.get(z));
						score.add(s);
					}
					int sample = decoder.samplePoint(score, T, decoder.r);
					if (sample != -1) {
						//decoder.updateTimes++;
						int oldDep = pred.dep[i];
						int oldLab = pred.labs[i];

						int newDep = dep.get(sample);
						int newLab = lab.get(sample);

						int oldSample = decoder.findOldSample(oldDep, oldLab, dep, lab);
						
						FeatureVector oldFV = fvList.get(oldSample);
						double oldScore = score.get(oldSample);
						FeatureVector newFV = fvList.get(sample);
						double newScore = score.get(sample);
						
						if (oldDep != newDep || oldLab != newLab) {

							int comp = decoder.compare(gold.dep[i], gold.labs[i], oldDep, oldLab, newDep, newLab);

							if (comp > 0 && newScore - oldScore < comp) {
								// update toward new tree
								double upd = (double)decoder.updateTimes;

								if (DependencyParser.learningMode == 0) {
									// perceptron
									params.updatePerceptron(newFV, oldFV, 1.0, upd);
								}
								else if (DependencyParser.learningMode == 1) {
									// mira
									params.updateParamsMIRA(newFV, oldFV, newScore, oldScore, comp, upd);
								}
								//updateTimes++;
							}
							if (comp < 0 && oldScore - newScore < -comp) {
								// update toward old tree
								double upd = (double)decoder.updateTimes;

								if (DependencyParser.learningMode == 0) {
									// perceptron
									params.updatePerceptron(oldFV, newFV, 1.0, upd);
								}
								else if (DependencyParser.learningMode == 1) {
									// mira
									params.updateParamsMIRA(oldFV, newFV, oldScore, newScore, -comp, upd);
								}
								//updateTimes++;
							}
							decoder.updateTimes++;

							// always accept
							int oldH = pred.dep[i];
							pred.dep[i] = newDep;
							pred.labs[i] = newLab;
							pipe.updateChildList(pred, newDep, oldH, i);
						}
					}
				}
				
			}
		}
	}

	@Override
	public void run() {
		try {
			while (true) {
				DependencyInstance gold = null;
				DependencyInstance pred = null;
				LocalFeatureData lfd = null;
				GlobalFeatureData gfd = null;
				int curid = -1;
				
				synchronized(readLock) {
					curid = currentID[0].intValue();
					if (curid >= il.length) {
						break;
					}
					else {
						currentID[0]++;		// id for next instance
					}
	
					//upd = decoder.updateTimes;
					//decoder.updateTimes++;
					
					DependencyInstance inst = il[curid];
		    		int length = inst.length;
		    		
		    		gold = pipe.readConllSent(in_g, pipe.labeled);
		    		pred = pipe.readConllSent(in_p, pipe.labeled);
		    		pred.m = gold.m;
		    		pred.posA = gold.posA;
	
		    		if (gold.sentence.length != length) {
		    			System.out.println("train length mismatch");
		    		}
		    		
		    		lfd = new LocalFeatureData(length, pipe.types.length, true);
		    		if (DependencyParser.pruningMode == 0 || DependencyParser.pruningMode == 1) {
		    			lfd.loadCandidate(null, pipe, pred, gold, true);
		    		}
		    		else if (DependencyParser.pruningMode == 2 || DependencyParser.pruningMode == 3){
			    		in_p.readLine();	// score
			    		lfd.loadCandidate(in_p, pipe, pred, gold, true);
			    		in_p.readLine();	// empty line
		    		}
		    		if (!(gold.sentence.length > DependencyParser.maxLength
		    				|| ((double)lfd.maxid / gold.sentence.length > DependencyParser.maxAvgHead)) ) {
			    		gfd = new GlobalFeatureData(lfd, pipe.cposAlphabet.size());
		    		}
		
				}
				
	    		if((curid+1) % 500 == 0)
	    			System.out.print("  "+(curid+1));
	
	    		if (gold.sentence.length > DependencyParser.maxLength
	    				|| ((double)lfd.maxid / gold.sentence.length > DependencyParser.maxAvgHead) )
	    			continue;
	
	  			// new upd
	   			//double upd = (double)decoder.updateTimes;
		    			
				int restartIter = 25;
	   			if (iter == 1 || iter % restartIter == 1) {
			    	// first iteration, copy the parser prediction to il
			 		decoder.copyDepLab(il[curid], pred.dep, pred.labs);
			 	}
			    else {
			    	// after first iteration, copy sample result to pred
			    	decoder.copyDepLab(pred, il[curid].dep, il[curid].labs);
			    }
			    pipe.buildChild(pred);
			    			
			    // sample rank
			    sampleRank(gold, pred, lfd, gfd, params);
			    
			    //copy sample result to il
			    decoder.copyDepLab(il[curid], pred.dep, pred.labs);
	
			    synchronized(stats) {
			   		stats[0] += gold.sentence.length - 1;
			   		for (int j = 1; j < gold.sentence.length; j++) {
			   			if (pred.dep[j] == gold.dep[j])
			   				stats[1]++;
			   			if (pred.dep[j] == gold.dep[j] && pred.labs[j] == gold.labs[j])
			   				stats[2]++;
			   		}
			    }
			}
			
			System.out.println(t.getName() + " done.");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
